/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.email

import java.nio.charset.StandardCharsets

import cats.*
import cats.syntax.all.*
import com.comcast.ip4s.*

import scala.util.matching.Regex

/** The transport strategies for communication with the email server.
  */
enum SmtpTransport {

    /** Use plaintext communication with the server (NOT RECOMMENDED!).
      */
    case PLAIN

    /** Encapsulate the entire communication within TLS.
      */
    case SMTPS

    /** Initialise via plaintext SMTP and issue a mandatory STARTTLS command first.
      */
    case TLS
}

opaque type EmailServerPassword = Array[Byte]
object EmailServerPassword {

    /** Create an instance of EmailServerPassword from the given Array[Byte] type.
      *
      * @param source
      *   An instance of type Array[Byte] which will be returned as a EmailServerPassword.
      * @return
      *   The appropriate instance of EmailServerPassword.
      */
    def apply(source: Array[Byte]): EmailServerPassword = source

    /** Try to create an instance of EmailServerPassword from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a EmailServerPassword.
      * @return
      *   An option to the successfully converted EmailServerPassword.
      */
    def from(source: String): Option[EmailServerPassword] = Option(source).map(_.getBytes(StandardCharsets.UTF_8))
}

extension (password: EmailServerPassword) {
    def toArray: Array[Byte] = password
}

opaque type EmailServerUsername = String
object EmailServerUsername {

    /** Create an instance of EmailServerUsername from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a EmailServerUsername.
      * @return
      *   The appropriate instance of EmailServerUsername.
      */
    def apply(source: String): EmailServerUsername = source

    /** Try to create an instance of EmailServerUsername from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a EmailServerUsername.
      * @return
      *   An option to the successfully converted EmailServerUsername.
      */
    def from(source: String): Option[EmailServerUsername] = Option(source)

}

/** An email address must fulfil several format requirements which in detail should be looked up in the implementation.
  */
opaque type EmailAddress = String
object EmailAddress {
    given Eq[EmailAddress] = Eq.fromUniversalEquals

    val validateString: Regex =
        """^([a-zA-Z0-9.!#$%&’'*+/=?^_`{|}~-]+)@([a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*)$""".r

    /** Create an instance of EmailAddress from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a EmailAddress.
      * @return
      *   The appropriate instance of EmailAddress.
      */
    def apply(source: String): EmailAddress = source

    /** Try to create an instance of EmailAddress from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a EmailAddress.
      * @return
      *   An option to the successfully converted EmailAddress.
      */
    def from(source: String): Option[EmailAddress] = {
        // Must obviously not be null or empty.
        val notEmpty = Option(source)
        // Must have at least one '@' sign.
        val hasAtChar = Option(source).filter(_.exists(_ === '@'))
        // Must have at least one '.' character.
        val hasDotChar = Option(source).filter(_.exists(_ === '.'))
        // The maximum allowed length is 128 characters.
        val belowMaxLength = Option(source).filter(_.length <= 128)
        // The last part of the email address (usually the top level domain) must at least be 2 characters long.
        val lastPartValid = Option(source) match {
            case None => None
            case Some(string) =>
                val parts = string.split('.')
                if (parts.lastOption.map(_.length >= 2).getOrElse(false))
                    Option(string)
                else
                    None
        }
        // Check against a regular expression. TODO: Maybe we can skip the other tests alltogether?
        val passesRegularExpression = Option(source).filter(validateString.matches)
        // Validate pre-conditions
        (notEmpty, hasAtChar, hasDotChar, belowMaxLength, lastPartValid, passesRegularExpression).mapN {
            case (email, _, _, _, _, _) => email
        }
    }

    extension (email: EmailAddress) {

        /** Convert the email into a [[de.smederee.email.FromAddress]] to be useable within the email middleware.
          *
          * @return
          *   The email address to be used in a `From` header.
          */
        def toFromAddress: FromAddress = FromAddress(email.toString)

        /** Convert the email into a [[de.smederee.email.ToAddress]] to be useable within the email middleware.
          *
          * @return
          *   The email address to be used in a `To` header.
          */
        def toToAddress: ToAddress = ToAddress(email.toString)
    }
}

opaque type FromAddress = String
object FromAddress {
    given Eq[FromAddress] = Eq.fromUniversalEquals

    val validateString: Regex =
        """^([a-zA-Z0-9.!#$%&’'*+/=?^_`{|}~-]+)@([a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*)$""".r

    /** Create an instance of FromAddress from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a FromAddress.
      * @return
      *   The appropriate instance of FromAddress.
      */
    def apply(source: String): FromAddress = source

    /** Try to create an instance of FromAddress from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a FromAddress.
      * @return
      *   An option to the successfully converted FromAddress.
      */
    def from(source: String): Option[FromAddress] = {
        // Must obviously not be null or empty.
        val notEmpty = Option(source)
        // Must have at least one '@' sign.
        val hasAtChar = Option(source).filter(_.exists(_ === '@'))
        // Must have at least one '.' character.
        val hasDotChar = Option(source).filter(_.exists(_ === '.'))
        // The maximum allowed length is 128 characters.
        val belowMaxLength = Option(source).filter(_.length <= 128)
        // The last part of the email address (usually the top level domain) must at least be 2 characters long.
        val lastPartValid = Option(source) match {
            case None => None
            case Some(string) =>
                val parts = string.split('.')
                if (parts.lastOption.map(_.length >= 2).getOrElse(false))
                    Option(string)
                else
                    None
        }
        // Check against a regular expression. TODO: Maybe we can skip the other tests alltogether?
        val passesRegularExpression = Option(source).filter(validateString.matches)
        // Validate pre-conditions
        (notEmpty, hasAtChar, hasDotChar, belowMaxLength, lastPartValid, passesRegularExpression).mapN {
            case (email, _, _, _, _, _) => email
        }
    }
}

opaque type SubjectLine = String
object SubjectLine {
    val MaxLength: Int = 78

    /** Create an instance of SubjectLine from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a SubjectLine.
      * @return
      *   The appropriate instance of SubjectLine.
      */
    def apply(source: String): SubjectLine = source

    /** Try to create an instance of SubjectLine from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a SubjectLine.
      * @return
      *   An option to the successfully converted SubjectLine.
      */
    def from(source: String): Option[SubjectLine] = Option(source).filter(_.length <= MaxLength)
}

opaque type TextBody = String
object TextBody {

    /** Create an instance of TextBody from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a TextBody.
      * @return
      *   The appropriate instance of TextBody.
      */
    def apply(source: String): TextBody = source

    /** Try to create an instance of TextBody from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a TextBody.
      * @return
      *   An option to the successfully converted TextBody.
      */
    def from(source: String): Option[TextBody] = Option(source).filter(_.nonEmpty)
}

opaque type ToAddress = String
object ToAddress {
    given Eq[ToAddress] = Eq.fromUniversalEquals

    val validateString: Regex =
        """^([a-zA-Z0-9.!#$%&’'*+/=?^_`{|}~-]+)@([a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*)$""".r

    /** Create an instance of ToAddress from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a ToAddress.
      * @return
      *   The appropriate instance of ToAddress.
      */
    def apply(source: String): ToAddress = source

    /** Try to create an instance of ToAddress from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a ToAddress.
      * @return
      *   An option to the successfully converted ToAddress.
      */
    def from(source: String): Option[ToAddress] = {
        // Must obviously not be null or empty.
        val notEmpty = Option(source)
        // Must have at least one '@' sign.
        val hasAtChar = Option(source).filter(_.exists(_ === '@'))
        // Must have at least one '.' character.
        val hasDotChar = Option(source).filter(_.exists(_ === '.'))
        // The maximum allowed length is 128 characters.
        val belowMaxLength = Option(source).filter(_.length <= 128)
        // The last part of the email address (usually the top level domain) must at least be 2 characters long.
        val lastPartValid = Option(source) match {
            case None => None
            case Some(string) =>
                val parts = string.split('.')
                if (parts.lastOption.map(_.length >= 2).getOrElse(false))
                    Option(string)
                else
                    None
        }
        // Check against a regular expression. TODO: Maybe we can skip the other tests alltogether?
        val passesRegularExpression = Option(source).filter(validateString.matches)
        // Validate pre-conditions
        (notEmpty, hasAtChar, hasDotChar, belowMaxLength, lastPartValid, passesRegularExpression).mapN {
            case (email, _, _, _, _, _) => email
        }
    }
}

/** Configuration for the email middleware.
  *
  * @param host
  *   The hostname of the email server (SMTP) to connect to.
  * @param port
  *   The port number to be used for the connection.
  * @param transport
  *   Specify the transport method (security) to be used for the connection (should either be SMTPS or TLS).
  * @param username
  *   An optional username if authentication is required.
  * @param password
  *   An optional password if authentication is required.
  */
final case class EmailMiddlewareConfiguration(
    host: Host,
    port: Port,
    transport: SmtpTransport,
    username: Option[EmailServerUsername],
    password: Option[EmailServerPassword]
)

object EmailMiddlewareConfiguration {
    // A default configuration for sending email via a local SMTP server without authentication.
    val LocalSmtpServer: EmailMiddlewareConfiguration = EmailMiddlewareConfiguration(
        host = host"localhost",
        port = port"25",
        transport = SmtpTransport.PLAIN,
        username = None,
        password = None
    )
}

/** An abstraction layer for functionality regarding the handling of email to ease testing.
  *
  * @tparam F
  *   A higher kinded type which usually must provide effect capabilities.
  */
trait EmailMiddleware[F[_]] {

    /** Send the given message via the email protocol. Details depend on the configuration used for the underlying
      * implementation.
      *
      * @param email
      *   An email message containing all necessary information (headers and message body) to be sent.
      * @return
      *   Either an error message or Unit.
      */
    def send(email: EmailMessage): F[Either[String, Unit]]

}
