/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.email

import java.nio.charset.StandardCharsets

import cats.effect.*
import cats.syntax.all.*
import org.simplejavamail.api.email.Email
import org.simplejavamail.api.mailer.Mailer
import org.simplejavamail.api.mailer.config.TransportStrategy
import org.simplejavamail.email.EmailBuilder
import org.simplejavamail.mailer.MailerBuilder

/** Provides an [[EmailMiddleware]] using the [Simple Java Mail](https://www.simplejavamail.org/) library and the IO
  * monad from cats effect.
  *
  * @param configuration
  *   The configuration containing all needed information to connect to the email server and send emails.
  */
final class SimpleJavaMailMiddleware(configuration: EmailMiddlewareConfiguration) extends EmailMiddleware[IO] {
    import SimpleJavaMailMiddleware.helpers.*

    private val cachedMailer: IO[Mailer] = for {
        builder <- IO.delay {
            val mailerBuilder = (configuration.username, configuration.password)
                .mapN { case (username, password) =>
                    MailerBuilder.withSMTPServer(
                        configuration.host.toString,
                        configuration.port.toString.toInt,
                        username.toString,
                        new String(password.toArray, StandardCharsets.UTF_8)
                    )
                }
                .getOrElse(
                    MailerBuilder.withSMTPServer(configuration.host.toString, configuration.port.toString.toInt)
                )
            // The mailer is a mutable construct from the Java library therefore we can omit the return value here.
            val _ = configuration.transport match {
                case SmtpTransport.PLAIN => mailerBuilder.withTransportStrategy(TransportStrategy.SMTP)
                case SmtpTransport.SMTPS => mailerBuilder.withTransportStrategy(TransportStrategy.SMTPS)
                case SmtpTransport.TLS   => mailerBuilder.withTransportStrategy(TransportStrategy.SMTP_TLS)
            }
            mailerBuilder
        }
    } yield builder.async().buildMailer()

    override def send(email: EmailMessage): IO[Either[String, Unit]] = {
        val sending = for {
            mailer  <- cachedMailer
            message <- IO.delay(email.toEmail)
            // A bit more involving because we use the asynchronous mailer.
            _ <- IO.fromCompletableFuture(IO.delay(mailer.sendMail(message)))
        } yield ()
        sending.attempt.map(_.leftMap(_.getMessage))
    }
}

object SimpleJavaMailMiddleware {
    object helpers {
        extension (email: EmailMessage) {

            /** Convert our custom email message into an email instance which is understood by our used library.
              *
              * @return
              *   An email useable by the Simple Java Mail library.
              */
            def toEmail: Email = {
                val builder = EmailBuilder
                    .startingBlank()
                    .from(email.from.toString)
                    .withSubject(email.subject.toString)
                    .withPlainText(email.textBody.toString)
                email.to.toList.foreach(address => builder.to(address.toString))
                email.cc.foreach(address => builder.cc(address.toString))
                email.bcc.foreach(address => builder.bcc(address.toString))
                builder.buildEmail()
            }
        }
    }
}
