/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.email

import java.nio.charset.StandardCharsets

import de.smederee.email.Generators.given

import munit.*

import org.scalacheck.Prop.*

final class EmailMiddlewareTest extends ScalaCheckSuite {
    property("EmailServerPassword.apply and .toArray must be reversible") {
        forAll { (randomBytes: Array[Byte]) =>
            val password = EmailServerPassword(randomBytes)
            assert(java.util.Arrays.equals(password.toArray, randomBytes))
        }
    }

    property("EmailServerPassword.from must create correct instances") {
        forAll { (randomString: String) =>
            EmailServerPassword.from(randomString) match {
                case None => fail("Could not create EmailServerPassword!")
                case Some(password) =>
                    assert(java.util.Arrays.equals(password.toArray, randomString.getBytes(StandardCharsets.UTF_8)))
            }
        }
    }

    property("FromAddress.from must not accept invalid input") {
        forAll { (randomString: String) =>
            assert(FromAddress.from(randomString).isEmpty)
        }
    }

    property("FromAddress.from must accept valid input") {
        forAll { (validFrom: FromAddress) =>
            FromAddress.from(validFrom.toString) match {
                case None          => fail("Could not create FromAddress!")
                case Some(address) => assertEquals(address, validFrom)
            }
        }
    }

    property("SubjectLine.from must only accept invalid input") {
        forAll { (randomString: String) =>
            if (randomString.length <= SubjectLine.MaxLength)
                assert(SubjectLine.from(randomString).nonEmpty)
            else
                assert(SubjectLine.from(randomString).isEmpty)
        }
    }

    test("TextBody.from must not accept empty input") {
        assert(TextBody.from("").isEmpty)
    }

    property("ToAddress.from must not accept invalid input") {
        forAll { (randomString: String) =>
            assert(ToAddress.from(randomString).isEmpty)
        }
    }

    property("ToAddress.from must accept valid input") {
        forAll { (validTo: ToAddress) =>
            ToAddress.from(validTo.toString) match {
                case None          => fail("Could not create ToAddress!")
                case Some(address) => assertEquals(address, validTo)
            }
        }
    }

}
