/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.html

import java.util.Locale

import cats.syntax.all.*
import org.commonmark.ext.gfm.strikethrough.StrikethroughExtension
import org.commonmark.ext.gfm.tables.TablesExtension
import org.commonmark.ext.heading.anchor.HeadingAnchorExtension
import org.commonmark.ext.task.list.items.TaskListItemsExtension
import org.commonmark.node.*
import org.commonmark.parser.Parser
import org.commonmark.renderer.NodeRenderer
import org.commonmark.renderer.html.*
import org.http4s.Uri
import org.slf4j.Logger
import org.slf4j.LoggerFactory

import scala.annotation.nowarn
import scala.jdk.CollectionConverters.*
import scala.util.matching.Regex

object MarkdownRenderer {
    private val log = LoggerFactory.getLogger(getClass())

    private val MarkdownExtensions =
        List(
            HeadingAnchorExtension.create(),
            StrikethroughExtension.create(),
            TablesExtension.create(),
            TaskListItemsExtension.create()
        )

    /** Render the given markdown content into HTML.
      *
      * @param markdownSource
      *   Markdown source code that shall be rendered.
      * @return
      *   A string containing the rendered markdown (HTML).
      */
    def render(markdownSource: String): String = {
        val parser   = Parser.builder().extensions(MarkdownExtensions.asJava).build()
        val markdown = parser.parse(markdownSource)
        val renderer = HtmlRenderer
            .builder()
            .escapeHtml(true)
            .extensions(MarkdownExtensions.asJava)
            .sanitizeUrls(true)
            .build()
        renderer.render(markdown)
    }

    /** Render the given markdown sources and adjust all relative links by prefixing them with the path for the
      * repostiory file browsing (`repo-name/files`).
      *
      * @param repositoryName
      *   The name of the repository which contains the markdown sources. If the option is empty then no URI path prefix
      *   will be set.
      * @param markdownSource
      *   A string containing the markdown sources to be rendered (usually the content of the README.md file in the
      *   repository root).
      * @return
      *   A string containing the rendered markdown (HTML).
      */
    def renderRepositoryMarkdownFile(repositoryName: Option[String])(markdownSource: String): String = {
        val parser   = Parser.builder().extensions(MarkdownExtensions.asJava).build()
        val markdown = parser.parse(markdownSource)
        val renderer = HtmlRenderer
            .builder()
            .attributeProviderFactory(new AttributeProviderFactory {
                override def create(context: AttributeProviderContext): AttributeProvider =
                    new LinkHrefCorrector(repositoryName)
            })
            .escapeHtml(true)
            .extensions(MarkdownExtensions.asJava)
            .nodeRendererFactory(new HtmlNodeRendererFactory {
                override def create(context: HtmlNodeRendererContext): NodeRenderer = new ToDoTextRenderer(context)
            })
            .sanitizeUrls(true)
            .build()
        renderer.render(markdown)
    }

    /** A helper class used by the `renderRepositoryOverviewReadme` function to adjust the `href` attribute of links
      * that are not absolute.
      *
      * @param repositoryName
      *   The name of the repository which contains the markdown sources. If the option is empty then no URI path prefix
      *   will be set.
      */
    @SuppressWarnings(
        Array("scalafix:DisableSyntax.isInstanceOf")
    ) // TODO: Find a way to get rid of the isInstanceOf below.
    @nowarn("msg=discarded non-Unit value.*")
    class LinkHrefCorrector(private val repositoryName: Option[String]) extends AttributeProvider {
        override def setAttributes(node: Node, tagName: String, attributes: java.util.Map[String, String]): Unit =
            if (node.isInstanceOf[Link]) {
                (repositoryName, attributes.asScala.get("href").flatMap(href => Uri.fromString(href).toOption)).mapN {
                    case (repositoryName, uri) =>
                        if (uri.scheme.isEmpty) {
                            val pathPrefix =
                                Uri.Path(Vector(Uri.Path.Segment(repositoryName), Uri.Path.Segment("files")))
                            val correctedUri =
                                if (uri.path.startsWith(pathPrefix))
                                    uri
                                else
                                    uri.copy(path = pathPrefix |+| uri.path)
                            log.debug(s"Corrected URI for repository overview README rendering: $uri -> $correctedUri")
                            val _ = attributes.put("href", correctedUri.toString)
                        }
                }
            }
    }

    /** A custom text node renderer which is supposed to highlight several words which are considered "todo items". This
      * is currently a very limited approach with a fixed list of matching words.
      *
      * @param context
      *   A context for an html node renderer that is needed to extract the html writer from it.
      */
    class ToDoTextRenderer(context: HtmlNodeRendererContext) extends NodeRenderer {
        import ToDoTextCssMapping.*

        private final val htmlWriter: HtmlWriter = context.getWriter()
        private final val log: Logger            = LoggerFactory.getLogger(getClass())

        override def getNodeTypes(): java.util.Set[Class[? <: Node]] = Set(classOf[Text]).asJava

        @SuppressWarnings(Array("scalafix:DisableSyntax.null", "scalafix:DisableSyntax.asInstanceOf"))
        override def render(node: Node): Unit = {
            val text = node.asInstanceOf[Text] // We only receive text nodes (see `getNodeTypes`).
            isToDoItem.findFirstMatchIn(text.getLiteral()) match {
                case Some(matchedItem) =>
                    log.debug(s"Matched TODO item: ${text.getLiteral()} (${matchedItem.groupCount})")
                    if (matchedItem.group(4) === null) {
                        val prefix   = matchedItem.group(1)
                        val item     = matchedItem.group(2)
                        val suffix   = matchedItem.group(3)
                        val cssClass = todoCssClasses(item.toLowerCase(Locale.ROOT))
                        htmlWriter.text(prefix)
                        htmlWriter.tag("span", Map("class" -> cssClass).asJava)
                        htmlWriter.text(item + ":")
                        htmlWriter.tag("/span")
                        htmlWriter.text(suffix)
                    } else {
                        val item     = matchedItem.group(4)
                        val cssClass = todoCssClasses(item.toLowerCase(Locale.ROOT))
                        htmlWriter.tag("span", Map("class" -> cssClass).asJava)
                        htmlWriter.text(item)
                        htmlWriter.tag("/span")
                    }
                case _ =>
                    htmlWriter.text(text.getLiteral())
            }
        }

    }
}

object ToDoTextCssMapping {
    /* We want to match either on specific words followed by a colon (`:`) and also on complete text nodes
     * which can be created if markdown rendering is involved.
     * For example a `**WORD** some text` will produce two text nodes (`WORD` and ` some text`)
     * while `WORD: some text` will only produce one.
     */
    val isToDoItem: Regex = "(?i)(.*)(DEBUG|FIXME|HACK|TODO):(.*)|^(DEBUG|FIXME|HACK|TODO)$".r

    // Mapping of todo item (words) to css classes for highlighting.
    val todoCssClasses: Map[String, String] = Map(
        "todo"  -> "todo-default",
        "fixme" -> "todo-error",
        "debug" -> "todo-info",
        "hack"  -> "todo-warning"
    ).withDefaultValue("todo-default")

}
