/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.html

enum MetaReferrerPolicy(val tag: String) {

    /** The Referer header will be omitted: sent requests do not include any referrer information.
      */
    case NoReferrer extends MetaReferrerPolicy("no-referrer")

    /** Send the origin, path, and querystring in Referer when the protocol security level stays the same or improves
      * (HTTP -> HTTP, HTTP -> HTTPS, HTTPS -> HTTPS). Don't send the Referer header for requests to less secure
      * destinations (HTTPS -> HTTP, HTTPS -> file).
      */
    case NoReferrerWhenDowngrade extends MetaReferrerPolicy("no-referrer-when-downgrade")

    /** Send only the origin in the Referer header. For example, a document at https://example.com/page.html will send
      * the referrer https://example.com/.
      */
    case Origin extends MetaReferrerPolicy("origin")

    /** When performing a same-origin request to the same protocol level (HTTP -> HTTP, HTTPS -> HTTPS), send the
      * origin, path, and query string. Send only the origin for cross origin requests and requests to less secure
      * destinations (HTTPS -> HTTP).
      */
    case OriginWhenCrossOrigin extends MetaReferrerPolicy("origin-when-cross-origin")

    /** Send the origin, path, and query string for same-origin requests. Don't send the Referer header for cross-origin
      * requests.
      */
    case SameOrigin extends MetaReferrerPolicy("same-origin")

    /** Send only the origin when the protocol security level stays the same (HTTPS -> HTTPS). Don't send the Referer
      * header to less secure destinations (HTTPS -> HTTP).
      */
    case StrictOrigin extends MetaReferrerPolicy("strict-origin")

    /** Send the origin, path, and querystring when performing a same-origin request. For cross-origin requests send the
      * origin (only) when the protocol security level stays same (HTTPS -> HTTPS). Don't send the Referer header to
      * less secure destinations (HTTPS -> HTTP).
      */
    case StrictOriginWhenCrossOrigin extends MetaReferrerPolicy("strict-origin-when-cross-origin")

    /** Send the origin, path, and query string when performing any request, regardless of security.
      */
    case UnsafeUrl extends MetaReferrerPolicy("unsafe-url")
}

enum MetaRobotsDirective(val tag: String) {

    /** Equivalent to index, follow. Used by Google
      */
    case All extends MetaRobotsDirective("all")

    /** Allows the robot to follow the links on the page (default). Used by all.
      */
    case Follow extends MetaRobotsDirective("follow")

    /** Allows the robot to index the page (default). Used by all.
      */
    case Index extends MetaRobotsDirective("index")

    /** Requests the search engine not to cache the page content. Used by Bing, Google, Yahoo.
      */
    case NoArchive extends MetaRobotsDirective("noarchive")

    /** Synonym of noarchive. Used by Bing.
      */
    case NoCache extends MetaRobotsDirective("nocache")

    /** Requests the robot to not follow the links on the page. Used by all.
      */
    case NoFollow extends MetaRobotsDirective("nofollow")

    /** Requests this page not to appear as the referring page of an indexed image. Used by Google.
      */
    case NoImageIndex extends MetaRobotsDirective("noimageindex")

    /** Requests the robot to not index the page. Used by all.
      */
    case NoIndex extends MetaRobotsDirective("noindex")

    /** Prevents displaying any description of the page in search engine results. Used by Bing, Google.
      */
    case NoSnippet extends MetaRobotsDirective("nosnippet")

    /** Equivalent to noindex, nofollow. Used by Google.
      */
    case None extends MetaRobotsDirective("none")
}

opaque type MetaDescription = String
object MetaDescription {

    /** Create an instance of MetaDescription from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a MetaDescription.
      * @return
      *   The appropriate instance of MetaDescription.
      */
    def apply(source: String): MetaDescription = source

    /** Try to create an instance of MetaDescription from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a MetaDescription.
      * @return
      *   An option to the successfully converted MetaDescription.
      */
    def from(source: String): Option[MetaDescription] = Option(source)

}

opaque type MetaKeyWords = List[String]
object MetaKeyWords {

    /** Create an instance of MetaKeyWords from the given List[String] type.
      *
      * @param source
      *   An instance of type List[String] which will be returned as a MetaKeyWords.
      * @return
      *   The appropriate instance of MetaKeyWords.
      */
    def apply(source: List[String]): MetaKeyWords = source

    /** Return an empty instance of MetaKeyWords.
      *
      * @return
      *   An empty list.
      */
    def empty: MetaKeyWords = List.empty

    /** Try to create an instance of MetaKeyWords from the given List[String].
      *
      * @param source
      *   A List[String] that should fulfil the requirements to be converted into a MetaKeyWords.
      * @return
      *   An option to the successfully converted MetaKeyWords.
      */
    def from(source: List[String]): Option[MetaKeyWords] =
        source.flatMap(string => Option(string)) match {
            case Nil      => None
            case keywords => Option(keywords)
        }

    extension (keywords: MetaKeyWords) {
        def isEmpty: Boolean  = keywords.isEmpty
        def mkString: String  = keywords.toList.mkString(", ")
        def nonEmpty: Boolean = keywords.nonEmpty
    }

}

/** HTML meta attributes which can be written into the header part of an HTML page.
  *
  * @param description
  *   An optional description for the related meta tag which should not be too long (max. 160/200 characters).
  * @param keywords
  *   A list of keywords which can be empty and should not be too long.
  * @param referrer
  *   The Referrer-Policy HTTP header controls how much referrer information (sent with the Referer header) should be
  *   included with requests.
  * @param robots
  *   A set of directives for the meta tag "robots" to affect search engine / crawler indexing behaviour.
  */
final case class MetaTags(
    description: Option[MetaDescription],
    keywords: MetaKeyWords,
    referrer: Option[MetaReferrerPolicy],
    robots: Set[MetaRobotsDirective]
)

object MetaTags {

    /** Return a default instance of meta tags which is empty and has the nofollow and noindex flags for robots set.
      *
      * @return
      *   An instance of meta tags containing only nofollow and noindex flags for robots.
      */
    def default: MetaTags =
        MetaTags(
            description = None,
            keywords = MetaKeyWords.empty,
            referrer = None,
            robots = Set(MetaRobotsDirective.NoFollow, MetaRobotsDirective.NoIndex)
        )

    /** Return an empty meta tags instance.
      *
      * @return
      *   An instance of meta tags containing no values, resulting in no tags being rendered.
      */
    def empty: MetaTags =
        MetaTags(description = None, keywords = MetaKeyWords.empty, referrer = None, robots = Set.empty)
}
