CREATE TABLE tickets.users
(
  uid              UUID                     NOT NULL,
  name             CHARACTER VARYING(32)    NOT NULL,
  email            CHARACTER VARYING(128)   NOT NULL,
  created_at       TIMESTAMP WITH TIME ZONE NOT NULL,
  updated_at       TIMESTAMP WITH TIME ZONE NOT NULL,
  CONSTRAINT users_pk           PRIMARY KEY (uid),
  CONSTRAINT users_unique_name  UNIQUE (name),
  CONSTRAINT users_unique_email UNIQUE (email)
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE tickets.users IS 'All users for the ticket system live within this table.';
COMMENT ON COLUMN tickets.users.uid IS 'A globally unique ID for the related user account. It must match the user ID from the hub account.';
COMMENT ON COLUMN tickets.users.name IS 'A username between 2 and 32 characters which must be globally unique, contain only lowercase alphanumeric characters and start with a character.';
COMMENT ON COLUMN tickets.users.email IS 'A globally unique email address associated with the account.';
COMMENT ON COLUMN tickets.users.created_at IS 'The timestamp of when the account was created.';
COMMENT ON COLUMN tickets.users.updated_at IS 'A timestamp when the account was last changed.';

CREATE TABLE tickets.sessions
(
  id         VARCHAR(32)              NOT NULL,
  uid        UUID                     NOT NULL,
  created_at TIMESTAMP WITH TIME ZONE NOT NULL,
  updated_at TIMESTAMP WITH TIME ZONE NOT NULL,
  CONSTRAINT sessions_pk     PRIMARY KEY (id),
  CONSTRAINT sessions_fk_uid FOREIGN KEY (uid)
    REFERENCES tickets.users (uid) ON UPDATE CASCADE ON DELETE CASCADE
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE tickets.sessions IS 'Keeps the sessions of users.';
COMMENT ON COLUMN tickets.sessions.id IS 'A globally unique session ID.';
COMMENT ON COLUMN tickets.sessions.uid IS 'The unique ID of the user account to whom the session belongs.';
COMMENT ON COLUMN tickets.sessions.created_at IS 'The timestamp of when the session was created.';
COMMENT ON COLUMN tickets.sessions.updated_at IS 'The session ID should be re-generated in regular intervals resulting in a copy of the old session entry with a new ID and the corresponding timestamp in this column.';

CREATE TABLE tickets.projects
(
  id                 BIGINT                   GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
  name               CHARACTER VARYING(64)    NOT NULL,
  owner              UUID                     NOT NULL,
  is_private         BOOLEAN                  NOT NULL DEFAULT FALSE,
  description        CHARACTER VARYING(254),
  created_at         TIMESTAMP WITH TIME ZONE NOT NULL,
  updated_at         TIMESTAMP WITH TIME ZONE NOT NULL,
  next_ticket_number INTEGER                  NOT NULL DEFAULT 1,
  CONSTRAINT projects_unique_owner_name UNIQUE (owner, name),
  CONSTRAINT projects_fk_uid FOREIGN KEY (owner)
    REFERENCES tickets.users (uid) ON UPDATE CASCADE ON DELETE CASCADE
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE tickets.projects IS 'All projects which are basically mirrored repositories from the hub are stored within this table.';
COMMENT ON COLUMN tickets.projects.id IS 'An auto generated primary key.';
COMMENT ON COLUMN tickets.projects.name IS 'The name of the project. A project name must start with a letter or number and must contain only alphanumeric ASCII characters as well as minus or underscore signs. It must be between 2 and 64 characters long.';
COMMENT ON COLUMN tickets.projects.owner IS 'The unique ID of the user account that owns the project.';
COMMENT ON COLUMN tickets.projects.is_private IS 'A flag indicating if this project is private i.e. only visible / accessible for users with appropriate permissions.';
COMMENT ON COLUMN tickets.projects.description IS 'An optional short text description of the project.';
COMMENT ON COLUMN tickets.projects.created_at IS 'The timestamp of when the project was created.';
COMMENT ON COLUMN tickets.projects.updated_at IS 'A timestamp when the project was last changed.';
COMMENT ON COLUMN tickets.projects.next_ticket_number IS 'Tickets are numbered ascending per project and this field holds the next logical ticket number to be used and must be incremented upon creation of a new ticket.';

CREATE TABLE tickets.labels
(
  id          BIGINT                 GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
  project     BIGINT                 NOT NULL,
  name        CHARACTER VARYING(40)  NOT NULL,
  description CHARACTER VARYING(254) DEFAULT NULL,
  colour      CHARACTER VARYING(7)   NOT NULL,
  CONSTRAINT labels_unique_project_label UNIQUE (project, name),
  CONSTRAINT labels_fk_project FOREIGN KEY (project)
    REFERENCES tickets.projects (id) ON UPDATE CASCADE ON DELETE CASCADE
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE tickets.labels IS 'Labels used to add information to tickets.';
COMMENT ON COLUMN tickets.labels.id IS 'An auto generated primary key.';
COMMENT ON COLUMN tickets.labels.project IS 'The project to which this label belongs.'; 
COMMENT ON COLUMN tickets.labels.name IS 'A short descriptive name for the label which is supposed to be unique in a project context.';
COMMENT ON COLUMN tickets.labels.description IS 'An optional description if needed.';
COMMENT ON COLUMN tickets.labels.colour IS 'A hexadecimal HTML colour code which can be used to mark the label on a rendered website.';

CREATE TABLE tickets.milestones
(
  id          BIGINT                 GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
  project     BIGINT                 NOT NULL,
  title       CHARACTER VARYING(64)  NOT NULL,
  due_date    DATE                   DEFAULT NULL,
  description TEXT                   DEFAULT NULL,
  CONSTRAINT milestones_unique_project_title UNIQUE (project, title),
  CONSTRAINT milestones_fk_project FOREIGN KEY (project)
    REFERENCES tickets.projects (id) ON UPDATE CASCADE ON DELETE CASCADE
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE tickets.milestones IS 'Milestones used to organise tickets';
COMMENT ON COLUMN tickets.milestones.project IS 'The project to which this milestone belongs.';
COMMENT ON COLUMN tickets.milestones.title IS 'A title for the milestone, usually a version number, a word or a short phrase that is supposed to be unique within a project context.';
COMMENT ON COLUMN tickets.milestones.due_date IS 'An optional date on which the milestone is supposed to be reached.';
COMMENT ON COLUMN tickets.milestones.description IS 'An optional longer description of the milestone.';

CREATE TABLE tickets.tickets
(
  id         BIGINT                   GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
  project    BIGINT                   NOT NULL,
  number     INT                      NOT NULL,
  title      CHARACTER VARYING(72)    NOT NULL,
  content    TEXT                     DEFAULT NULL,
  status     CHARACTER VARYING(16)    NOT NULL,
  submitter  UUID                     DEFAULT NULL,
  created_at TIMESTAMP WITH TIME ZONE NOT NULL,
  updated_at TIMESTAMP WITH TIME ZONE NOT NULL,
  CONSTRAINT tickets_unique_project_ticket UNIQUE (project, number),
  CONSTRAINT tickets_fk_project FOREIGN KEY (project)
    REFERENCES tickets.projects (id) ON UPDATE CASCADE ON DELETE CASCADE,
  CONSTRAINT tickets_fk_submitter FOREIGN KEY (submitter)
    REFERENCES tickets.users (uid) ON UPDATE CASCADE ON DELETE SET NULL
)
WITH (
  OIDS=FALSE
);

CREATE INDEX tickets_status ON tickets.tickets (status);

COMMENT ON TABLE tickets.tickets IS 'Information about tickets for projects.';
COMMENT ON COLUMN tickets.tickets.id IS 'An auto generated primary key.';
COMMENT ON COLUMN tickets.tickets.project IS 'The unique ID of the project which is associated with the ticket.';
COMMENT ON COLUMN tickets.tickets.number IS 'The number of the ticket which must be unique within the scope of the project.';
COMMENT ON COLUMN tickets.tickets.title IS 'A concise and short description of the ticket which should not exceed 72 characters.';
COMMENT ON COLUMN tickets.tickets.content IS 'An optional field to describe the ticket in great detail if needed.';
COMMENT ON COLUMN tickets.tickets.status IS 'The current status of the ticket describing its life cycle.';
COMMENT ON COLUMN tickets.tickets.submitter IS 'The person who submitted (created) this ticket which is optional because of possible account deletion or other reasons.';
COMMENT ON COLUMN tickets.tickets.created_at IS 'The timestamp when the ticket was created / submitted.';
COMMENT ON COLUMN tickets.tickets.updated_at IS 'The timestamp when the ticket was last updated. Upon creation the update time equals the creation time.';

CREATE TABLE tickets.milestone_tickets
(
  milestone BIGINT NOT NULL,
  ticket    BIGINT NOT NULL,
  CONSTRAINT milestone_tickets_pk PRIMARY KEY (milestone, ticket),
  CONSTRAINT milestone_tickets_fk_milestone FOREIGN KEY (milestone)
    REFERENCES tickets.milestones (id) ON UPDATE CASCADE ON DELETE CASCADE,
  CONSTRAINT milestone_tickets_fk_ticket FOREIGN KEY (ticket)
    REFERENCES tickets.tickets (id) ON UPDATE CASCADE ON DELETE CASCADE
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE tickets.milestone_tickets IS 'This table stores the relation between milestones and their tickets.';
COMMENT ON COLUMN tickets.milestone_tickets.milestone IS 'The unique ID of the milestone.';
COMMENT ON COLUMN tickets.milestone_tickets.ticket IS 'The unique ID of the ticket that is attached to the milestone.';

CREATE TABLE tickets.ticket_assignees
(
  ticket     BIGINT NOT NULL,
  assignee   UUID NOT NULL,
  CONSTRAINT ticket_assignees_pk PRIMARY KEY (ticket, assignee),
  CONSTRAINT ticket_assignees_fk_ticket FOREIGN KEY (ticket)
    REFERENCES tickets.tickets (id) ON UPDATE CASCADE ON DELETE CASCADE,
  CONSTRAINT ticket_assignees_fk_assignee FOREIGN KEY (assignee)
    REFERENCES tickets.users (uid) ON UPDATE CASCADE ON DELETE CASCADE
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE tickets.ticket_assignees IS 'This table stores the relation between tickets and their assignees.';
COMMENT ON COLUMN tickets.ticket_assignees.ticket IS 'The unqiue ID of the ticket.';
COMMENT ON COLUMN tickets.ticket_assignees.assignee IS 'The unique ID of the user account that is assigned to the ticket.';

CREATE TABLE tickets.ticket_labels
(
  ticket BIGINT NOT NULL,
  label  BIGINT NOT NULL,
  CONSTRAINT ticket_labels_pk PRIMARY KEY (ticket, label),
  CONSTRAINT ticket_labels_fk_ticket FOREIGN KEY (ticket)
    REFERENCES tickets.tickets (id) ON UPDATE CASCADE ON DELETE CASCADE,
  CONSTRAINT ticket_labels_fk_label FOREIGN KEY (label)
    REFERENCES tickets.labels (id) ON UPDATE CASCADE ON DELETE CASCADE
)
WITH (
  OIDS=FALSE
);

COMMENT ON TABLE tickets.ticket_labels IS 'This table stores the relation between tickets and their labels.';
COMMENT ON COLUMN tickets.ticket_labels.ticket IS 'The unqiue ID of the ticket.';
COMMENT ON COLUMN tickets.ticket_labels.label IS 'The unique ID of the label that is attached to the ticket.';
