/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.nio.charset.StandardCharsets

import cats.*
import cats.syntax.all.*
import de.smederee.email.EmailAddress
import de.smederee.i18n.LanguageCode
import de.smederee.security.*
import org.springframework.security.crypto.argon2.Argon2PasswordEncoder

import scala.util.matching.Regex

/** Extension methods for a password instance.
  *
  * TODO: Refactor this into the security package!
  *
  * @param p
  *   A password.
  */
extension (p: Password) {

    /** Encode the password using the argon2 algorithm.
      *
      * @return
      *   The hash value for the password generated by the underlying Argon2 implementation.
      */
    def encode: PasswordHash = {
        val encoder = PasswordEncoder.Argon2
        PasswordHash(encoder.encode(new String(p.toArray, StandardCharsets.UTF_8)))
    }

    /** Verify if the password matches the given `PasswordHash`.
      *
      * @param hash
      *   A password hash generated by the Argon2 algorithm.
      * @return
      *   `true` if the hash was generated from the given password and `false` otherwise.
      */
    def matches(hash: PasswordHash): Boolean = {
        val encoder = PasswordEncoder.Argon2
        encoder.matches(new String(p.toArray, StandardCharsets.UTF_8), hash.toString)
    }
}

/** Initialises and holds our preferred password encoding algorithm.
  */
object PasswordEncoder {
    // Initialisation parameters for the argon2 algorithm (see whitepaper for details).
    private val saltLength  = 16    // 128 bits
    private val hashLength  = 32    // 256 bits
    private val parallelism = 1
    private val memoryUsage = 12288 // 12 MB
    private val iterations  = 10

    def Argon2 = new Argon2PasswordEncoder(saltLength, hashLength, parallelism, memoryUsage, iterations)

}

opaque type ResetToken = String
object ResetToken {
    val Format: Regex = "^[a-zA-z0-9]+".r
    val Length: Int   = 128

    given Eq[ResetToken] = Eq.fromUniversalEquals

    /** Create an instance of ResetToken from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a ResetToken.
      * @return
      *   The appropriate instance of ResetToken.
      */
    def apply(source: String): ResetToken = source

    /** Try to create an instance of ResetToken from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a ResetToken.
      * @return
      *   An option to the successfully converted ResetToken.
      */
    def from(source: String): Option[ResetToken] = Option(source).filter(s => s.length === Length && Format.matches(s))

    /** Generate a new reset token.
      *
      * @return
      *   A randomly generated reset token.
      */
    def generate: ResetToken = scala.util.Random.alphanumeric.take(Length).mkString
}

/** Extractor to retrieve a ResetToken from a path parameter.
  */
object ResetTokenPathParameter {
    def unapply(str: String): Option[ResetToken] = Option(str).flatMap(ResetToken.from)
}

opaque type UnlockToken = String
object UnlockToken {
    val Format: Regex = "^[a-zA-z0-9]+".r
    val Length: Int   = 128

    given Eq[UnlockToken] = Eq.fromUniversalEquals

    /** Create an instance of UnlockToken from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a UnlockToken.
      * @return
      *   The appropriate instance of UnlockToken.
      */
    def apply(source: String): UnlockToken = source

    /** Try to create an instance of UnlockToken from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a UnlockToken.
      * @return
      *   An option to the successfully converted UnlockToken.
      */
    def from(source: String): Option[UnlockToken] = Option(source).filter(s => s.length === Length && Format.matches(s))

    /** Generate a new unlock token.
      *
      * @return
      *   A randomly generated unlock token.
      */
    def generate: UnlockToken = scala.util.Random.alphanumeric.take(Length).mkString

}

/** Extractor to retrieve an UnlockToken from a path parameter.
  */
object UnlockTokenPathParameter {
    def unapply(str: String): Option[UnlockToken] = UnlockToken.from(str)
}

/** Extractor to retrieve an Username from a path parameter.
  */
object UsernamePathParameter {
    def unapply(str: String): Option[Username] =
        Option(str).flatMap { string =>
            if (string.startsWith("~"))
                Username.from(string.drop(1))
            else
                None
        }
}

opaque type ValidationToken = String
object ValidationToken {
    val Format: Regex = "^[a-zA-z0-9]+".r
    val Length: Int   = 64

    given Eq[ValidationToken] = Eq.fromUniversalEquals

    /** Create an instance of ValidationToken from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a ValidationToken.
      * @return
      *   The appropriate instance of ValidationToken.
      */
    def apply(source: String): ValidationToken = source

    /** Try to create an instance of ValidationToken from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a ValidationToken.
      * @return
      *   An option to the successfully converted ValidationToken.
      */
    def from(source: String): Option[ValidationToken] =
        Option(source).filter(s => s.length === Length && Format.matches(s))

    /** Generate a new unlock token.
      *
      * @return
      *   A randomly generated unlock token.
      */
    def generate: ValidationToken = scala.util.Random.alphanumeric.take(Length).mkString

}

/** Extractor to retrieve a validation token from a path parameter.
  */
object ValidationTokenPathParameter {
    def unapply(str: String): Option[ValidationToken] = ValidationToken.from(str)
}

/** A user account.
  *
  * @param uid
  *   The unique ID of the account.
  * @param name
  *   A unique name which can be used for login and to identify the user.
  * @param email
  *   The email address of the user.
  * @param fullName
  *   The optional human readable full name of an account which must be non empty and is limited to 128 characters.
  * @param validatedEmail
  *   This flag indicates if the email address of the user has been validated via a validation email.
  * @param language
  *   The language code of the users preferred language.
  */
final case class Account(
    uid: UserId,
    name: Username,
    email: EmailAddress,
    fullName: Option[FullName],
    validatedEmail: Boolean,
    language: Option[LanguageCode]
)

object Account {
    given Eq[Account] =
        Eq.instance { (a, b) =>
            a.uid === b.uid && a.name === b.name && a.email === b.email
        }

    extension (account: Account) {

        /** Create vcs repository owner metadata from the account.
          *
          * @return
          *   Descriptive information about the owner of a vcs repository based on the account.
          */
        def toVcsRepositoryOwner: VcsRepositoryOwner = VcsRepositoryOwner(account.uid, account.name, account.email)
    }
}
