/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.OffsetDateTime
import java.time.ZoneOffset

import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.html.ExternalUrlConfiguration
import de.smederee.hub.*
import de.smederee.hub.RequestHelpers.instances.given_RequestHelpers_Request
import de.smederee.hub.SessionHelpers.instances.toAuthenticationCookie
import de.smederee.hub.config.*
import de.smederee.hub.forms.types.FormErrors
import de.smederee.hub.forms.types.FormFieldError
import de.smederee.security.*
import org.http4s.*
import org.http4s.dsl.Http4sDsl
import org.http4s.headers.Location
import org.http4s.implicits.*
import org.http4s.twirl.TwirlInstances.*
import org.slf4j.LoggerFactory

import scala.concurrent.duration.*

/** Enumeration of possible kinds of authentication failures.
  */
enum AuthenticationFailure {

    /** The account was not found in the database.
      */
    case AccountNotFound extends AuthenticationFailure

    /** The account is locked.
      */
    case AccountLocked extends AuthenticationFailure

    /** The provided password was wrong.
      */
    case WrongPassword extends AuthenticationFailure
}

/** The routes for handling the authentication related tasks like login, logout, password reset, unlocking etc.
  *
  * @param clock
  *   A clock to generate nounces for cryptograpic operations.
  * @param authenticationConfig
  *   The configuration for the authentication module.
  * @param external
  *   Settings affecting how the service will communicate several information to the "outside world" e.g. if it runs
  *   behind a reverse proxy.
  * @param repo
  *   The database repository providing needed functionality.
  * @param signAndValidate
  *   A class providing functions to handle session token signing and validation.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class AuthenticationRoutes[F[_]: Async](
    clock: java.time.Clock,
    authenticationConfig: AuthenticationConfiguration,
    external: ExternalUrlConfiguration,
    repo: AuthenticationRepository[F],
    signAndValidate: SignAndValidate
) extends Http4sDsl[F] {
    private val log = LoggerFactory.getLogger(getClass)

    private val loginPath = uri"/login"
    private val resetPath = uri"/forgot-password"

    private val parseLoginForm: HttpRoutes[F] = HttpRoutes.of[F] { case request @ POST -> Root / "login" =>
        request.decodeStrict[F, UrlForm] { urlForm =>
            for {
                csrf <- Sync[F].delay(request.getCsrfToken)
                formData <- Sync[F].delay {
                    urlForm.values.map { t =>
                        val (key, values) = t
                        (
                            key,
                            values.headOption.getOrElse("")
                        ) // Pick the first value (a field might get submitted multiple times)!
                    }
                }
                form <- Sync[F].delay(LoginForm.validate(formData))
                response <- form match {
                    case Validated.Invalid(es) =>
                        BadRequest(
                            views.html
                                .login()(loginPath, csrf, resetPath, title = "Smederee - Login to your account".some)(
                                    formData,
                                    FormErrors.fromNec(es)
                                )
                        )
                    case Validated.Valid(loginForm) =>
                        for {
                            user <- repo.findAccountByName(loginForm.name)
                            auth <- user.map(_.uid).traverse(repo.findPasswordHashAndAttempts).map(_.getOrElse(None))
                            check = auth.map { tuple =>
                                val (hash, failedAttempts) = tuple
                                (
                                    loginForm.password.matches(hash),
                                    failedAttempts <= authenticationConfig.lockAfter.toInt,
                                    failedAttempts
                                )
                            }
                            source <- Sync[F].delay(request.from.map(_.toString).getOrElse("UNKNOWN_ADDRESS"))
                            login <- check match {
                                case None =>
                                    // No account was found!
                                    for {
                                        delay <- Sync[F]
                                            .delay(
                                                scala.util.Random.nextInt(2) + 1
                                            ) // Prevent fast response to avoid account guessing.
                                        _ <- Sync[F].sleep(FiniteDuration(delay, SECONDS))
                                    } yield Left(AuthenticationFailure.AccountNotFound)
                                case Some((true, true, _)) =>
                                    // The password is valid and the attempts are below or equal to the limit.
                                    Sync[F].delay(user.toRight(AuthenticationFailure.AccountNotFound))
                                case Some((false, true, attempts)) =>
                                    // The password is wrong and the attempts are below or equal to the limit.
                                    for {
                                        _ <- user.map(_.uid).traverse(repo.incrementFailedAttempts)
                                        _ <- user.traverse(user =>
                                            Sync[F].delay(log.warn(s"Login failure for ${user.name} from $source."))
                                        )
                                        // We increase the delay for processing failed login attempts to discourage brute forcing.
                                        delay <- Sync[F].delay(
                                            (attempts - 1) * 5
                                        ) // TODO: Extract the base delay into the configuration.
                                        _ <- Sync[F].sleep(FiniteDuration(delay, SECONDS))
                                    } yield Left(AuthenticationFailure.WrongPassword)
                                case Some((_, false, _)) =>
                                    // The login attempts are above the limit.
                                    for {
                                        _ <- user
                                            .map(_.uid)
                                            .traverse(uid => repo.lockAccount(uid)(UnlockToken.generate.some))
                                        _ <- user.traverse(user =>
                                            Sync[F].delay(
                                                log.warn(
                                                    s"Locked account ${user.name} (too many authentication failures), latest request from $source!"
                                                )
                                            )
                                        )
                                    } yield Left(AuthenticationFailure.AccountLocked)
                            }
                            response <- login match {
                                case Left(_) =>
                                    BadRequest(
                                        views.html.login()(
                                            loginPath,
                                            csrf,
                                            resetPath,
                                            title = "Smederee - Login to your account".some
                                        )(
                                            formData,
                                            Map(LoginForm.fieldGlobal -> List(FormFieldError("Invalid credentials!")))
                                        )
                                    )
                                case Right(account) =>
                                    for {
                                        session <- Sync[F].delay(
                                            Session.create(
                                                account.uid,
                                                OffsetDateTime.now(ZoneOffset.UTC),
                                                OffsetDateTime.now(ZoneOffset.UTC)
                                            )
                                        )
                                        _ <- repo.createUserSession(session)
                                        _ <- repo.resetFailedAttempts(account.uid)
                                        token <- Sync[F].delay(
                                            signAndValidate.signToken(session.id.toString)(clock.millis.toString)
                                        )
                                        response <- SeeOther(Location(Uri(path = Uri.Path.Root)))
                                            .map(
                                                _.addCookie(
                                                    token.toAuthenticationCookie(external.host.toString.some)(None)(
                                                        secure = true
                                                    )
                                                )
                                            )
                                    } yield response
                            }
                        } yield response
                }
            } yield response
        }
    }

    private val parseLoginFormForLoggedInUsers: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case _ @POST -> Root / "login" as _ =>
            SeeOther(Location(Uri(path = Uri.Path.Root))) // Redirect already logged in users.
    }

    private val logout: AuthedRoutes[Account, F] = AuthedRoutes.of { case ar @ POST -> Root / "logout" as _ =>
        for {
            authToken <- Sync[F].delay(ar.req.getAuthenticationToken)
            sessionId <- Sync[F].delay(authToken.flatMap(signAndValidate.validate).flatMap(SessionId.from))
            _         <- sessionId.traverse(repo.deleteUserSession)
            response  <- SeeOther(Location(loginPath)).map(_.removeCookie(Constants.authenticationCookieName.toString))
        } yield response
    }

    private val showLoginForm: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root / "login" =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            response <- Ok(
                views.html.login()(loginPath, csrf, resetPath, title = "Smederee - Login to your account".some)()
            )
        } yield response
    }

    private val showLoginFormForLoggedInUsers: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case _ @GET -> Root / "login" as _ =>
            SeeOther(Location(Uri(path = Uri.Path.Root))) // Redirect already logged in users.
    }

    private val unlockAccount: HttpRoutes[F] = HttpRoutes.of {
        case request @ GET -> Root / "unlock" / UsernamePathParameter(name) / UnlockTokenPathParameter(token) =>
            for {
                csrf     <- Sync[F].delay(request.getCsrfToken)
                user     <- repo.findLockedAccount(name)(token.some)
                _        <- user.map(_.uid).traverse(repo.unlockAccount)
                _        <- user.map(_.uid).traverse(repo.resetFailedAttempts)
                _        <- user.traverse(user => Sync[F].delay(log.info(s"Unlocked account for ${user.name}.")))
                response <- SeeOther(Location(loginPath))
            } yield response
    }

    val protectedRoutes =
        if (authenticationConfig.enabled)
            logout <+> showLoginFormForLoggedInUsers <+> parseLoginFormForLoggedInUsers
        else
            AuthedRoutes.empty[Account, F]

    val routes =
        if (authenticationConfig.enabled)
            showLoginForm <+> parseLoginForm <+> unlockAccount
        else
            HttpRoutes.empty[F]

}
