/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.data.*
import cats.syntax.all.*
import de.smederee.hub.forms.*
import de.smederee.hub.forms.types.*
import de.smederee.security.*

/** Data container for changine an account password in the reset password process.
  *
  * @param name
  *   The username which must match the name of the account with the appropriate reset token.
  * @param password
  *   The password that shall be set.
  * @param passwordConfirmation
  *   A password confirmation field for safety reasons to reduce errors due to typos.
  * @param token
  *   The reset password token for the current request.
  */
final case class ChangePasswordForm(
    name: Username,
    password: Password,
    passwordConfirmation: Password,
    token: ResetToken
)

object ChangePasswordForm extends FormValidator[ChangePasswordForm] {
    val fieldName: FormField                 = FormField("name")
    val fieldPassword: FormField             = FormField("password")
    val fieldPasswordConfirmation: FormField = FormField("password_confirmation")
    val fieldResetToken: FormField           = FormField("token")

    override def validate(data: Map[String, String]): ValidatedNec[FormErrors, ChangePasswordForm] = {
        val name: ValidatedNec[FormErrors, Username] = data
            .get(fieldName)
            .fold(FormFieldError("No username given!").invalidNec)(s =>
                Username.from(s).fold(FormFieldError("Invalid username!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldName -> es.toList)))
        val password: ValidatedNec[FormErrors, Password] = data
            .get(fieldPassword)
            .fold("No password given!".invalidNec)(Password.validate)
            .leftMap(es => NonEmptyChain.of(Map(fieldPassword -> es.toList.map(FormFieldError.apply))))
        val passwordConfirmation: ValidatedNec[FormErrors, Password] = data
            .get(fieldPasswordConfirmation)
            .fold("No password given!".invalidNec)(Password.validate)
            .leftMap(es => NonEmptyChain.of(Map(fieldPasswordConfirmation -> es.toList.map(FormFieldError.apply))))
        val passwordsMatching: ValidatedNec[FormErrors, Password] = (password.toOption, passwordConfirmation.toOption)
            .mapN { case (pw, pwc) =>
                (pw, pwc)
            }
            .filter(tuple => tuple._1.matches(tuple._2.encode)) match {
                case None =>
                    Map(fieldPasswordConfirmation -> List(FormFieldError("Passwords do not match!"))).invalidNec
                case Some((pw, _)) => pw.validNec
            }
        val token: ValidatedNec[FormErrors, ResetToken] = data
            .get(fieldResetToken)
            .fold(FormFieldError("No reset token!").invalidNec)(s =>
                ResetToken.from(s).fold(FormFieldError("Invalid reset token!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldResetToken -> es.toList)))
        (name, password, passwordConfirmation, passwordsMatching, token).mapN {
            case (validName, validPassword, validPasswordConfirmation, _, validResetToken) =>
                ChangePasswordForm(validName, validPassword, validPasswordConfirmation, validResetToken)
        }
    }
}
