/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.util.UUID

import cats.effect.*
import de.smederee.email.EmailAddress
import de.smederee.i18n.LanguageCode
import de.smederee.security.*
import de.smederee.ssh.*
import doobie.*
import doobie.implicits.*
import doobie.postgres.implicits.*
import fs2.Stream

final class DoobieAccountManagementRepository[F[_]: Sync](tx: Transactor[F]) extends AccountManagementRepository[F] {
    given Meta[EmailAddress]    = Meta[String].timap(EmailAddress.apply)(_.toString)
    given Meta[EncodedKeyBytes] = Meta[String].timap(EncodedKeyBytes.unsafeFrom)(_.toString)
    given Meta[FullName]        = Meta[String].timap(FullName.apply)(_.toString)
    given Meta[KeyComment]      = Meta[String].timap(KeyComment.apply)(_.toString)
    given Meta[KeyFingerprint]  = Meta[String].timap(KeyFingerprint.apply)(_.toString)
    given Meta[LanguageCode]    = Meta[String].timap(LanguageCode.apply)(_.toString)
    given Meta[PasswordHash]    = Meta[String].timap(PasswordHash.apply)(_.toString)
    given Meta[SshKeyType]      = Meta[String].timap(id => SshKeyType.Mappings(id))(_.identifier)
    given Meta[UserId]          = Meta[UUID].timap(UserId.apply)(_.toUUID)
    given Meta[Username]        = Meta[String].timap(Username.apply)(_.toString)
    given Meta[ValidationToken] = Meta[String].timap(ValidationToken.apply)(_.toString)

    private val selectAccountColumns =
        fr"""SELECT uid, name, email, full_name, validated_email, language FROM "hub"."accounts""""

    override def addSshKey(key: PublicSshKey): F[Int] =
        sql"""INSERT INTO "hub"."ssh_keys" (id, uid, key_type, key, fingerprint, comment, created_at)
      VALUES(${key.id}, ${key.ownerId}, ${key.keyType}, ${key.keyBytes}, ${key.fingerprint}, ${key.comment}, NOW())""".update.run
            .transact(tx)

    override def deleteAccount(uid: UserId): F[Int] =
        sql"""DELETE FROM "hub"."accounts" WHERE uid = $uid""".update.run.transact(tx)

    override def deleteSshKey(keyId: UUID, ownerId: UserId): F[Int] =
        sql"""DELETE FROM "hub"."ssh_keys" WHERE id = $keyId AND uid = $ownerId""".update.run.transact(tx)

    override def findByValidationToken(token: ValidationToken): F[Option[Account]] = {
        val query = selectAccountColumns ++ fr"""WHERE validation_token = $token""" ++ fr"""LIMIT 1"""
        query.query[Account].option.transact(tx)
    }

    override def findPasswordHash(uid: UserId): F[Option[PasswordHash]] =
        sql"""SELECT password FROM "hub"."accounts" WHERE uid = $uid LIMIT 1""".query[PasswordHash].option.transact(tx)

    override def listSshKeys(uid: UserId): Stream[F, PublicSshKey] =
        sql"""SELECT id, uid, key_type, key, fingerprint, comment, created_at, last_used_at FROM "hub"."ssh_keys" WHERE uid = $uid"""
            .query[PublicSshKey]
            .stream
            .transact(tx)

    override def markAsValidated(uid: UserId): F[Int] =
        sql"""UPDATE "hub"."accounts" SET validated_email = TRUE, validation_token = NULL WHERE uid = $uid""".update.run
            .transact(tx)

    override def setLanguage(uid: UserId, language: Option[LanguageCode]): F[Int] =
        sql"""UPDATE "hub"."accounts" SET language = $language WHERE uid = $uid""".update.run.transact(tx)

    override def setValidationToken(uid: UserId, token: ValidationToken): F[Int] =
        sql"""UPDATE "hub"."accounts" SET validation_token = $token WHERE uid = $uid""".update.run.transact(tx)

}
