/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.util.UUID

import cats.effect.*
import de.smederee.email.EmailAddress
import de.smederee.i18n.LanguageCode
import de.smederee.security.*
import doobie.*
import doobie.Fragments.*
import doobie.implicits.*
import doobie.postgres.implicits.*
import fs2.Stream
import org.http4s.Uri

final class DoobieOrganisationRepository[F[_]: Sync](tx: Transactor[F]) extends OrganisationRepository[F] {
    given Meta[EmailAddress]            = Meta[String].timap(EmailAddress.apply)(_.toString)
    given Meta[FullName]                = Meta[String].timap(FullName.apply)(_.toString)
    given Meta[LanguageCode]            = Meta[String].timap(LanguageCode.apply)(_.toString)
    given Meta[OrganisationDescription] = Meta[String].timap(OrganisationDescription.apply)(_.toString)
    given Meta[OrganisationId]          = Meta[UUID].timap(OrganisationId.apply)(_.toUUID)
    given Meta[PasswordHash]            = Meta[String].timap(PasswordHash.apply)(_.toString)
    given Meta[SessionId]               = Meta[String].timap(SessionId.apply)(_.toString)
    given Meta[UnlockToken]             = Meta[String].timap(UnlockToken.apply)(_.toString)
    given Meta[Uri]                     = Meta[String].timap(Uri.unsafeFromString)(_.toString)
    given Meta[UserId]                  = Meta[UUID].timap(UserId.apply)(_.toUUID)
    given Meta[Username]                = Meta[String].timap(Username.apply)(_.toString)

    private val selectOrganisationColumns =
        fr"""SELECT id, name, owner, full_name, description, is_private, website FROM hub.organisations AS organisations"""

    override def addAdministrator(organisationId: OrganisationId)(user: UserId): F[Int] =
        sql"""INSERT INTO hub.organisation_admins (
                organisation,
                admin
              ) VALUES (
                $organisationId,
                $user
              ) ON CONFLICT (organisation, admin) DO NOTHING""".update.run
            .transact(tx)

    override def allByOwner(owner: UserId): Stream[F, Organisation] = {
        val ownerFilter = fr"""owner = $owner"""
        val sqlQuery    = selectOrganisationColumns ++ whereAnd(ownerFilter)
        sqlQuery.query[Organisation].stream.transact(tx)
    }

    override def create(org: Organisation): F[Int] =
        sql"""INSERT INTO hub.organisations (
                id,
                name,
                owner,
                full_name,
                description,
                is_private,
                website,
                created_at,
                updated_at
              ) VALUES (
                ${org.oid},
                ${org.name},
                ${org.owner},
                ${org.fullName},
                ${org.description},
                ${org.isPrivate},
                ${org.website},
                NOW(),
                NOW()
              )""".update.run.transact(tx)

    override def delete(organisationId: OrganisationId): F[Int] =
        sql"""DELETE FROM hub.organisations WHERE id = $organisationId""".update.run.transact(tx)

    override def find(organisationId: OrganisationId): F[Option[Organisation]] = {
        val idFilter = fr"""id = $organisationId"""
        val sqlQuery = selectOrganisationColumns ++ whereAnd(idFilter)
        sqlQuery.query[Organisation].option.transact(tx)
    }

    override def findByName(name: Username): F[Option[Organisation]] = {
        val nameFilter = fr"""organisations.name = $name"""
        val sqlQuery   = selectOrganisationColumns ++ whereAnd(nameFilter)
        sqlQuery.query[Organisation].option.transact(tx)
    }

    override def findOwner(organisationId: OrganisationId): F[Option[Account]] = {
        val organisationFilter = fr"""organisations.id = $organisationId"""
        val sqlQuery =
            fr"""SELECT
                   accounts.uid,
                   accounts.name,
                   accounts.email,
                   accounts.full_name,
                   accounts.validated_email,
                   accounts.language
                 FROM hub.accounts AS accounts
                 JOIN hub.organisations AS organisations
                 ON accounts.uid = organisations.owner""" ++ whereAnd(organisationFilter)
        sqlQuery.query[Account].option.transact(tx)
    }

    override def getAdministrators(organisationId: OrganisationId): Stream[F, Account] = {
        val organisationFilter = fr"""admins.organisation = $organisationId"""
        val sqlQuery =
            fr"""SELECT
                   accounts.uid,
                   accounts.name,
                   accounts.email,
                   accounts.full_name,
                   accounts.validated_email,
                   accounts.language
                 FROM hub.accounts AS accounts
                 JOIN hub.organisation_admins AS admins
                 ON accounts.uid = admins.admin""" ++ whereAnd(organisationFilter)
        sqlQuery.query[Account].stream.transact(tx)
    }

    override def removeAdministrator(organisationId: OrganisationId)(user: UserId): F[Int] =
        sql"""DELETE FROM hub.organisation_admins WHERE organisation = $organisationId AND admin = $user""".update.run
            .transact(tx)

    override def update(org: Organisation): F[Int] =
        sql"""UPDATE hub.organisations SET
                name = ${org.name},
                owner = ${org.owner},
                full_name = ${org.fullName},
                description = ${org.description},
                is_private = ${org.isPrivate},
                website = ${org.website},
                updated_at = NOW()
              WHERE id = ${org.oid}""".update.run.transact(tx)
}
