/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.OffsetDateTime
import java.util.UUID

import cats.effect.*
import de.smederee.email.EmailAddress
import de.smederee.i18n.LanguageCode
import de.smederee.security.*
import doobie.*
import doobie.Fragments.*
import doobie.implicits.*
import doobie.postgres.implicits.*

final class DoobieResetPasswordRepository[F[_]: Sync](tx: Transactor[F]) extends ResetPasswordRepository[F] {
    given Meta[EmailAddress] = Meta[String].timap(EmailAddress.apply)(_.toString)
    given Meta[FullName]     = Meta[String].timap(FullName.apply)(_.toString)
    given Meta[LanguageCode] = Meta[String].timap(LanguageCode.apply)(_.toString)
    given Meta[PasswordHash] = Meta[String].timap(PasswordHash.apply)(_.toString)
    given Meta[ResetToken]   = Meta[String].timap(ResetToken.apply)(_.toString)
    given Meta[SessionId]    = Meta[String].timap(SessionId.apply)(_.toString)
    given Meta[UnlockToken]  = Meta[String].timap(UnlockToken.apply)(_.toString)
    given Meta[UserId]       = Meta[UUID].timap(UserId.apply)(_.toUUID)
    given Meta[Username]     = Meta[String].timap(Username.apply)(_.toString)

    private val notLockedFilter              = fr"""locked_at IS NULL"""
    private val resetTokenExpiryNotSetFilter = fr"""reset_expiry IS NULL"""
    private val resetTokenExpirySetFilter    = fr"""reset_expiry IS NOT NULL"""
    private val resetTokenNotExpiredFilter   = fr"""reset_expiry > NOW()"""
    private val selectAccountColumns =
        fr"""SELECT uid, name, email, full_name, validated_email, language FROM hub.accounts"""

    override def findByNameAndResetPasswordToken(name: Username, token: ResetToken): F[Option[Account]] = {
        val nameFilter       = fr"""name = $name"""
        val resetTokenFilter = fr"""reset_token = $token"""
        val query = selectAccountColumns ++ whereAnd(
            notLockedFilter,
            nameFilter,
            resetTokenExpiryNotSetFilter,
            resetTokenFilter
        ) ++ fr"""LIMIT 1"""
        query.query[Account].option.transact(tx)
    }

    override def findByResetPasswordToken(token: ResetToken): F[Option[Account]] = {
        val resetTokenFilter = fr"""reset_token = $token"""
        val query = selectAccountColumns ++ whereAnd(
            notLockedFilter,
            resetTokenExpirySetFilter,
            resetTokenNotExpiredFilter,
            resetTokenFilter
        ) ++ fr"""LIMIT 1"""
        query.query[Account].option.transact(tx)
    }

    override def removeResetPasswordExpirationDate(uid: UserId): F[Int] =
        sql"""UPDATE hub.accounts SET reset_expiry = NULL WHERE uid = $uid""".update.run.transact(tx)

    override def removeResetPasswordToken(uid: UserId): F[Int] =
        sql"""UPDATE hub.accounts SET reset_expiry = NULL, reset_token = NULL WHERE uid = $uid""".update.run
            .transact(tx)

    override def setPassword(uid: UserId)(hash: PasswordHash): F[Int] =
        sql"""UPDATE hub.accounts SET password = $hash WHERE uid = $uid""".update.run.transact(tx)

    override def setResetPasswordToken(uid: UserId)(token: ResetToken, tokenExpiration: OffsetDateTime): F[Int] =
        sql"""UPDATE hub.accounts SET  reset_expiry = $tokenExpiration, reset_token = $token WHERE uid = $uid""".update.run
            .transact(tx)
}
