/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.util.UUID

import cats.effect.*
import cats.syntax.all.*
import de.smederee.email.EmailAddress
import de.smederee.hub.VcsMetadataRepositoriesOrdering.*
import de.smederee.security.UserId
import de.smederee.security.Username
import doobie.*
import doobie.Fragments.*
import doobie.implicits.*
import doobie.postgres.implicits.*
import fs2.Stream
import org.http4s.Uri

final class DoobieVcsMetadataRepository[F[_]: Sync](tx: Transactor[F]) extends VcsMetadataRepository[F] {
    given Meta[EmailAddress]             = Meta[String].timap(EmailAddress.apply)(_.toString)
    given Meta[Uri]                      = Meta[String].timap(Uri.unsafeFromString)(_.toString)
    given Meta[UserId]                   = Meta[UUID].timap(UserId.apply)(_.toUUID)
    given Meta[Username]                 = Meta[String].timap(Username.apply)(_.toString)
    given Meta[VcsRepositoryDescription] = Meta[String].timap(VcsRepositoryDescription.apply)(_.toString)
    given Meta[VcsRepositoryId]          = Meta[Long].timap(VcsRepositoryId.apply)(_.toLong)
    given Meta[VcsRepositoryName]        = Meta[String].timap(VcsRepositoryName.apply)(_.toString)
    given Meta[VcsType]                  = Meta[String].timap(VcsType.valueOf)(_.toString)

    private val selectRepositoryColumns =
        fr"""SELECT
          repos.name AS name,
          accounts.uid AS owner_id,
          accounts.name AS owner_name,
          accounts.email AS owner_email,
          repos.is_private AS is_private,
          repos.description AS description,
          repos.tickets_enabled AS tickets_enabled,
          repos.vcs_type AS vcs_type,
          repos.website AS website
        FROM hub.repositories AS repos
        JOIN hub.accounts AS accounts
        ON repos.owner = accounts.uid"""

    override def createFork(source: VcsRepositoryId, target: VcsRepositoryId): F[Int] =
        sql"""INSERT INTO hub.forks (
            original_repo,
            forked_repo
          ) VALUES (
            $source,
            $target
          )""".update.run.transact(tx)

    override def createVcsRepository(repository: VcsRepository): F[Int] =
        sql"""INSERT INTO hub.repositories (
            name,
            owner,
            is_private,
            description,
            tickets_enabled,
            vcs_type,
            website,
            created_at,
            updated_at
          ) VALUES (
            ${repository.name},
            ${repository.owner.uid},
            ${repository.isPrivate},
            ${repository.description},
            ${repository.ticketsEnabled},
            ${repository.vcsType},
            ${repository.website},
            NOW(),
            NOW()
          )""".update.run.transact(tx)

    override def deleteVcsRepository(repository: VcsRepository): F[Int] =
        sql"""DELETE FROM hub.repositories
          WHERE owner = ${repository.owner.uid}
          AND name = ${repository.name}""".update.run.transact(tx)

    override def findVcsRepository(
        owner: VcsRepositoryOwner,
        name: VcsRepositoryName
    ): F[Option[VcsRepository]] = {
        val nameFilter  = fr"""repos.name = $name"""
        val ownerFilter = fr"""repos.owner = ${owner.uid}"""
        val query       = selectRepositoryColumns ++ whereAnd(ownerFilter, nameFilter) ++ fr"""LIMIT 1"""
        query.query[VcsRepository].option.transact(tx)
    }

    override def findVcsRepositoryBranches(
        originalRepositoryId: VcsRepositoryId
    ): Stream[F, (Username, VcsRepositoryName)] = {
        val query = sql"""SELECT
                        accounts.name AS owner_name,
                        repos.name AS repository_name
                      FROM hub.forks AS forks
                      JOIN hub.repositories AS repos
                      ON forks.forked_repo = repos.id
                      JOIN hub.accounts AS accounts
                      ON repos.owner = accounts.uid
                      WHERE forks.original_repo = $originalRepositoryId"""
        query.query[(Username, VcsRepositoryName)].stream.transact(tx)
    }

    override def findVcsRepositoryId(owner: VcsRepositoryOwner, name: VcsRepositoryName): F[Option[VcsRepositoryId]] = {
        val nameFilter  = fr"""name = $name"""
        val ownerFilter = fr"""owner = ${owner.uid}"""
        val query = fr"""SELECT id FROM hub.repositories""" ++ whereAnd(ownerFilter, nameFilter) ++ fr"""LIMIT 1"""
        query.query[VcsRepositoryId].option.transact(tx)
    }

    override def findVcsRepositoryOwner(name: Username): F[Option[VcsRepositoryOwner]] =
        sql"""SELECT uid, name, email FROM hub.accounts WHERE name = $name LIMIT 1"""
            .query[VcsRepositoryOwner]
            .option
            .transact(tx)

    override def findVcsRepositoryParentFork(
        owner: VcsRepositoryOwner,
        name: VcsRepositoryName
    ): F[Option[VcsRepository]] = {
        val query =
            selectRepositoryColumns ++ fr"""WHERE repos.id = (SELECT original_repo FROM hub.forks WHERE forked_repo = (SELECT id FROM hub.repositories WHERE name = $name AND owner = ${owner.uid}))"""
        query.query[VcsRepository].option.transact(tx)
    }

    override def listAllRepositories(
        requester: Option[Account]
    )(ordering: VcsMetadataRepositoriesOrdering): Stream[F, VcsRepository] = {
        val orderClause = ordering match {
            case NameAscending  => fr"""ORDER BY name ASC"""
            case NameDescending => fr"""ORDER BY name DESC"""
        }
        // We use a SQL UNION here which first queries all others repos and then (potentially) the repos of the requester.
        val query = requester.fold(selectRepositoryColumns ++ fr"""WHERE is_private = FALSE""")(account =>
            selectRepositoryColumns ++ fr"""WHERE is_private = FALSE AND owner != ${account.uid}""" ++ fr"""UNION""" ++ selectRepositoryColumns ++ fr"""WHERE owner = ${account.uid}"""
        ) ++ orderClause
        query.query[VcsRepository].stream.transact(tx)
    }

    override def listRepositories(
        requester: Option[Account]
    )(owner: VcsRepositoryOwner): Stream[F, VcsRepository] = {
        val ownerFilter = fr"""repos.owner = ${owner.uid}"""
        val whereClause = requester match {
            case None => whereAnd(ownerFilter, fr"""repos.is_private = FALSE""") // Guest only see public repos.
            case Some(account) =>
                if (account.uid === owner.uid)
                    whereAnd(ownerFilter) // The user asks for their own repositories.
                else
                    whereAnd(ownerFilter, fr"""repos.is_private = FALSE""") // TODO: More logic later (groups, perms).
        }
        val query = selectRepositoryColumns ++ whereClause ++ fr"""ORDER BY name ASC"""
        query.query[VcsRepository].stream.transact(tx)
    }

    override def updateVcsRepository(repository: VcsRepository): F[Int] =
        sql"""UPDATE hub.repositories
            SET is_private = ${repository.isPrivate},
            description = ${repository.description},
            tickets_enabled = ${repository.ticketsEnabled},
            website = ${repository.website},
            updated_at = NOW()
          WHERE owner = ${repository.owner.uid}
          AND name = ${repository.name}""".update.run.transact(tx)
}
