/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.data.*
import cats.syntax.all.*
import de.smederee.hub.forms.*
import de.smederee.hub.forms.types.*
import org.http4s.Uri

/** Data container for the form to edit / customise a VCS repository.
  *
  * @param name
  *   The name of the repository. A repository name must start with a letter or number and must contain only
  *   alphanumeric ASCII characters as well as minus or underscore signs. It must be between 2 and 64 characters long.
  * @param isPrivate
  *   A flag indicating if this repository is private i.e. only visible / accessible for accounts with appropriate
  *   permissions.
  * @param description
  *   An optional short text description of the repository.
  * @param ticketsEnabled
  *   A flag indicating if ticket tracking support via the ticket service is enabled for this repository.
  * @param website
  *   An optional uri pointing to a website related to the repository / project.
  */
final case class EditVcsRepositoryForm(
    name: VcsRepositoryName,
    isPrivate: Boolean,
    description: Option[VcsRepositoryDescription],
    ticketsEnabled: Boolean,
    website: Option[Uri]
)

object EditVcsRepositoryForm extends FormValidator[EditVcsRepositoryForm] {
    val fieldDescription: FormField    = FormField("description")
    val fieldIsPrivate: FormField      = FormField("is_private")
    val fieldName: FormField           = FormField("name")
    val fieldTicketsEnabled: FormField = FormField("tickets_enabled")
    val fieldWebsite: FormField        = FormField("website")

    /** Create a form for editing a vcs repository filled with the data from the given repository.
      *
      * @param repo
      *   A VCS repository object containing the data that shall be used to fill the form.
      * @return
      *   A edit vcs repository form filled with the data from the repository.
      */
    def fromVcsRepository(repo: VcsRepository): EditVcsRepositoryForm =
        EditVcsRepositoryForm(repo.name, repo.isPrivate, repo.description, repo.ticketsEnabled, repo.website)

    override def validate(data: Map[String, String]): ValidatedNec[FormErrors, EditVcsRepositoryForm] = {
        val name = data
            .get(fieldName)
            .fold(FormFieldError("No repository name given!").invalidNec)(s =>
                VcsRepositoryName.from(s).fold(FormFieldError("Invalid repository name!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldName -> es.toList)))
        val privateFlag: ValidatedNec[FormErrors, Boolean] =
            data.get(fieldIsPrivate).fold(false.validNec)(s => s.matches("true").validNec)
        val description = data
            .get(fieldDescription)
            .fold(Option.empty[VcsRepositoryDescription].validNec) { s =>
                if (s.trim.isEmpty)
                    Option.empty[VcsRepositoryDescription].validNec // Sometimes "empty" strings are sent.
                else
                    VcsRepositoryDescription
                        .from(s)
                        .fold(FormFieldError("Invalid repository description!").invalidNec)(descr =>
                            Option(descr).validNec
                        )
            }
            .leftMap(es => NonEmptyChain.of(Map(fieldDescription -> es.toList)))
        val ticketsEnabledFlag: ValidatedNec[FormErrors, Boolean] =
            data.get(fieldTicketsEnabled).fold(false.validNec)(s => s.matches("true").validNec)
        val website = data
            .get(fieldWebsite)
            .fold(Option.empty[Uri].validNec) { s =>
                if (s.trim.isEmpty)
                    Option.empty[Uri].validNec // Sometimes "empty" strings are sent.
                else
                    Uri
                        .fromString(s)
                        .toOption
                        .fold(FormFieldError("Invalid website URI!").invalidNec) { uri =>
                            uri.scheme match {
                                case Some(Uri.Scheme.http) | Some(Uri.Scheme.https) => Option(uri).validNec
                                case _ => FormFieldError("Invalid website URI!").invalidNec
                            }
                        }
            }
            .leftMap(es => NonEmptyChain.of(Map(fieldWebsite -> es.toList)))
        (name, privateFlag, description, ticketsEnabledFlag, website).mapN {
            case (validName, isPrivate, validDescription, ticketsEnabled, validWebsite) =>
                EditVcsRepositoryForm(validName, isPrivate, validDescription, ticketsEnabled, validWebsite)
        }
    }

    extension (form: EditVcsRepositoryForm) {

        /** Convert the form class into a stringified map which is used as underlying data type for form handling in the
          * twirl templating library.
          *
          * @return
          *   A stringified map containing the data of the form.
          */
        def toMap: Map[String, String] = {
            val isPrivate =
                if (form.isPrivate)
                    "true"
                else
                    "false"
            val ticketsEnabled =
                if (form.ticketsEnabled)
                    "true"
                else
                    "false"
            val formData = Map(
                EditVcsRepositoryForm.fieldName.toString           -> form.name.toString,
                EditVcsRepositoryForm.fieldIsPrivate.toString      -> isPrivate,
                EditVcsRepositoryForm.fieldTicketsEnabled.toString -> ticketsEnabled
            )
            val description = form.description.fold(Map.empty)(description =>
                Map(EditVcsRepositoryForm.fieldDescription.toString -> description.toString)
            )
            val website =
                form.website.fold(Map.empty)(website =>
                    Map(EditVcsRepositoryForm.fieldWebsite.toString -> website.toString)
                )
            formData ++ description ++ website
        }
    }
}
