/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.nio.charset.StandardCharsets
import java.nio.file.Files
import java.nio.file.attribute.PosixFilePermissions
import java.util.Locale

import cats.arrow.FunctionK
import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import com.typesafe.config.*
import de.smederee.darcs.*
import de.smederee.email.EmailAddress
import de.smederee.email.SimpleJavaMailMiddleware
import de.smederee.html.*
import de.smederee.hub.config.*
import de.smederee.security.*
import de.smederee.ssh.*
import de.smederee.tickets.*
import de.smederee.tickets.config.*
import doobie.*
import org.http4s.*
import org.http4s.ember.server.*
import org.http4s.implicits.*
import org.http4s.server.*
import org.http4s.server.middleware.CSRF
import org.http4s.server.middleware.ErrorAction
import org.http4s.server.middleware.ErrorHandling
import org.http4s.server.staticcontent.resourceServiceBuilder
import org.slf4j.LoggerFactory
import pureconfig.*
import scodec.bits.ByteVector

/** Commands that may be passed to the hub server to trigger specific behaviour.
  *
  * @param help
  *   A string containing basic help information for the command.
  * @param subcommands
  *   A list of subcommands available.
  */
enum HubCommand(val help: String, val subcommands: List[HubSubcommand]) {
    case user extends HubCommand(help = "Manage user accounts.", subcommands = List(HubSubcommand.list))
}

/** Subcommands that are grouped below a [[HubCommand]].
  *
  * @param arguments
  *   A list of possible arguments for the subcommand.
  * @param help
  *   A string containing basic help information for the subcommand.
  */
enum HubSubcommand(val arguments: List[HubArgument], val help: String) {
    case delete
        extends HubSubcommand(
            arguments = List(HubArgument.DryRun, HubArgument.Username),
            help = "Delete the account from the database. This is not reversible!"
        )
    case find
        extends HubSubcommand(
            arguments = List(HubArgument.Email, HubArgument.Username),
            help = "Find a user account in the database."
        )
    case list extends HubSubcommand(arguments = Nil, help = "List all accounts in the database.")
    case lock
        extends HubSubcommand(
            arguments = List(HubArgument.DryRun, HubArgument.Username),
            help = "Lock the account in the database."
        )
    case unlock
        extends HubSubcommand(
            arguments = List(HubArgument.DryRun, HubArgument.Username),
            help = "Unlock the account in the database."
        )
}

/** Possible arguments for commands and subcommands.
  *
  * @param help
  *   A string containing basic help information for the subcommand.
  */
enum HubArgument(val help: String) {
    case DryRun   extends HubArgument(help = "--dry-run")
    case Email    extends HubArgument(help = "<email>")
    case Username extends HubArgument(help = "<username>")
}

/** This is the main entry point for the hub service.
  *
  * It initialises the application (configuration parsing, database migrations) and starts the HTTP service eventually.
  */
object HubServer extends IOApp with AuthenticationMiddleware {
    private val log = LoggerFactory.getLogger(getClass)

    given org.typelevel.log4cats.LoggerFactory[IO] = org.typelevel.log4cats.slf4j.Slf4jFactory.create[IO]

    /** Create a function to perform the origin checks for the CSRF protection middleware using the configuration.
      *
      * @param allowedOrigins
      *   A list of configuration settings describing the external url configuration for a service. These settings are
      *   affecting how a service will communicate several information to the "outside world" e.g. if it runs behind a
      *   reverse proxy. These settings are needed to validate the correct values for expected hostname, port and
      *   transport scheme of the checked requests.
      * @return
      *   A function which will check the correct origin of requests / cookies inside the CSRF middleware.
      */
    private def createCsrfOriginCheck(
        allowedOrigins: NonEmptyList[ExternalUrlConfiguration]
    ): Request[IO] => Boolean = { request =>
        allowedOrigins.exists { linkConfig =>
            CSRF.defaultOriginCheck(request, linkConfig.host.toString, linkConfig.scheme, linkConfig.port.map(_.value))
        }
    }

    /** An error handler that simply logs error messages but nothing more.
      *
      * @param exception
      *   An exception thrown by underlying code.
      * @param message
      *   An error message.
      */
    private def handleError(exception: Throwable, message: => String) = IO(log.error(message, exception))

    /** Try to load the CSRF key from the given path. If it doesn't exist or fails then a new key is generated and
      * stored in the file.
      *
      * @param csrfKeyFile
      *   A file which contains the key used to build the CSRF protection.
      * @return
      *   A byte vector containing the key used for CSRF protection.
      */
    private def loadOrCreateCsrfKey(csrfKeyFile: java.nio.file.Path): IO[ByteVector] =
        for {
            _ <- IO(
                Files.createDirectories(
                    csrfKeyFile.getParent,
                    PosixFilePermissions.asFileAttribute(PosixFilePermissions.fromString("rwxr-xr-x"))
                )
            )
            key <- Files.exists(csrfKeyFile) match {
                case false => CSRF.generateSigningKey[IO]()
                case true  => IO(ByteVector(Files.readAllBytes(csrfKeyFile)))
            }
            _ <- IO(Files.write(csrfKeyFile, key.toArray))
            _ <- IO(Files.setPosixFilePermissions(csrfKeyFile, PosixFilePermissions.fromString("rw-------")))
        } yield key

    /** Run migrations on the databases in the given configurations and return their results.
      *
      * @param hubConfiguration
      *   Configuration for the Smederee hub service.
      * @param ticketsConfiguration
      *   Configuration for the Smederee ticket service.
      * @return
      *   A tuple holding the migration results.
      */
    private def migrateDatabases(
        hubConfiguration: SmedereeHubConfig,
        ticketsConfiguration: SmedereeTicketsConfiguration
    ) = {
        val hubDatabaseMigrator     = new de.smederee.hub.DatabaseMigrator[IO]
        val ticketsDatabaseMigrator = new de.smederee.tickets.config.DatabaseMigrator[IO]
        for {
            hubMigrations <- hubDatabaseMigrator.migrate(
                hubConfiguration.database.url,
                hubConfiguration.database.user,
                hubConfiguration.database.pass
            )
            ticketMigrations <- ticketsDatabaseMigrator.migrate(
                ticketsConfiguration.database.url,
                ticketsConfiguration.database.user,
                ticketsConfiguration.database.pass
            )
        } yield (hubMigrations, ticketMigrations)
    }

    private def generateHelp(): String =
        s"""
       |user delete\t[--dry-run] <name>\t- Delete a locked(!) user account from the database.
       |user find\t<email> || <name>\t- Find a user account with the given email or name.
       |user list\t\t\t\t- List all unlocked(!) user accounts in the database.
       |user lock\t[--dry-run] <name>\t- Lock the user account.
       |user unlock\t[--dry-run] <name>\t- Unlock the user account.
       |""".stripMargin

    def run(args: List[String]): IO[ExitCode] = {
        val _ = Locale.setDefault(Locale.ENGLISH) // TODO: Make this configurable.
        val hubConfiguration = ConfigSource
            .fromConfig(ConfigFactory.load(getClass.getClassLoader))
            .at(SmedereeHubConfig.location)
            .loadOrThrow[SmedereeHubConfig]
        args match {
            case List("help") | List("--help") | List("-h") =>
                // TODO: Add help information!
                for {
                    _ <- IO(println("Welcome to the Smederee Hub Server."))
                    _ <- IO(println("==================================="))
                    _ <- IO(println("To simply start the server just run this command without arguments."))
                    _ <- IO(println(generateHelp()))
                } yield ExitCode.Success
            case commandString :: parameters =>
                val hubTransactor = Transactor.fromDriverManager[IO](
                    driver = hubConfiguration.database.driver,
                    url = hubConfiguration.database.url,
                    user = hubConfiguration.database.user,
                    password = hubConfiguration.database.pass,
                    logHandler = None
                )
                val authenticationRepo = new DoobieAuthenticationRepository[IO](hubTransactor)
                HubCommand.valueOf(commandString.toLowerCase(Locale.ROOT)) match {
                    case HubCommand.user =>
                        parameters match {
                            case subcommandString :: arguments =>
                                HubSubcommand.valueOf(subcommandString.toLowerCase(Locale.ROOT)) match {
                                    case HubSubcommand.delete =>
                                        val accountManagementRepo =
                                            new DoobieAccountManagementRepository[IO](hubTransactor)
                                        val dryRun = arguments.headOption.exists(_ === "--dry-run")
                                        val name =
                                            if (dryRun)
                                                arguments.drop(1).headOption.flatMap(Username.from)
                                            else
                                                arguments.headOption.flatMap(Username.from)
                                        for {
                                            user <- name.traverse(name =>
                                                authenticationRepo.findLockedAccount(name)(None)
                                            )
                                            _ <- user.flatten.fold(IO(println("No such locked user account found!")))(
                                                account =>
                                                    IO(
                                                        println(
                                                            s"Going to delete user ${account.name} (${account.email})!"
                                                        )
                                                    )
                                            )
                                            deleted <-
                                                if (dryRun)
                                                    IO.pure(0.some)
                                                else
                                                    user.flatten.traverse(account =>
                                                        accountManagementRepo.deleteAccount(account.uid)
                                                    )
                                            result = deleted.getOrElse(0) match {
                                                case 0 =>
                                                    ExitCode.Error // If not database rows were updated we assume an error.
                                                case _ => ExitCode.Success
                                            }
                                        } yield result
                                    case HubSubcommand.find =>
                                        val email = arguments.headOption.flatMap(EmailAddress.from)
                                        val name  = arguments.headOption.flatMap(Username.from)
                                        for {
                                            userByEmail <- email.traverse(authenticationRepo.findAccountByEmail)
                                            userByName  <- name.traverse(authenticationRepo.findAccountByName)
                                            users = List(userByEmail.flatten, userByName.flatten).collect {
                                                case Some(user) =>
                                                    user
                                            }
                                            _ <- IO(
                                                users.map(user =>
                                                    println(
                                                        s"${user.name},${user.email},${user.validatedEmail},${user.language}"
                                                    )
                                                )
                                            )
                                        } yield ExitCode.Success
                                    case HubSubcommand.list =>
                                        val users = authenticationRepo.allAccounts()
                                        users
                                            .foreach(user =>
                                                IO(
                                                    println(
                                                        s"${user.name},${user.email},${user.validatedEmail},${user.language}"
                                                    )
                                                )
                                            )
                                            .compile
                                            .drain
                                            .as(ExitCode.Success)
                                    case HubSubcommand.lock =>
                                        val dryRun = arguments.headOption.exists(_ === "--dry-run")
                                        val name =
                                            if (dryRun)
                                                arguments.drop(1).headOption.flatMap(Username.from)
                                            else
                                                arguments.headOption.flatMap(Username.from)
                                        for {
                                            user <- name.traverse(authenticationRepo.findAccountByName)
                                            _ <- user.flatten.fold(IO(println("No such unlocked user account found!")))(
                                                account =>
                                                    IO(
                                                        println(
                                                            s"Going to lock user ${account.name} (${account.email})!"
                                                        )
                                                    )
                                            )
                                            locked <-
                                                if (dryRun)
                                                    IO.pure(0.some)
                                                else
                                                    user.flatten.traverse(account =>
                                                        authenticationRepo.lockAccount(account.uid)(
                                                            UnlockToken.generate.some
                                                        )
                                                    )
                                            result = locked.getOrElse(0) match {
                                                case 0 =>
                                                    ExitCode.Error // If not database rows were updated we assume an error.
                                                case _ => ExitCode.Success
                                            }
                                        } yield result
                                    case HubSubcommand.unlock =>
                                        val dryRun = arguments.headOption.exists(_ === "--dry-run")
                                        val name =
                                            if (dryRun)
                                                arguments.drop(1).headOption.flatMap(Username.from)
                                            else
                                                arguments.headOption.flatMap(Username.from)
                                        for {
                                            user <- name.traverse(name =>
                                                authenticationRepo.findLockedAccount(name)(None)
                                            )
                                            _ <- user.flatten.fold(IO(println("No such locked user account found!")))(
                                                account =>
                                                    IO(
                                                        println(
                                                            s"Going to unlock user ${account.name} (${account.email})!"
                                                        )
                                                    )
                                            )
                                            unlocked <-
                                                if (dryRun)
                                                    IO.pure(0.some)
                                                else
                                                    user.flatten.traverse(account =>
                                                        authenticationRepo.unlockAccount(account.uid)
                                                    )
                                            result = unlocked.getOrElse(0) match {
                                                case 0 =>
                                                    ExitCode.Error // If not database rows were updated we assume an error.
                                                case _ => ExitCode.Success
                                            }
                                        } yield result
                                }
                            case _ => IO.raiseError(new IllegalArgumentException("Missing subcommand/options!"))
                        }
                }
            case _ =>
                for {
                    ticketsConfiguration <- IO(
                        ConfigSource
                            .fromConfig(ConfigFactory.load(getClass.getClassLoader))
                            .at(SmedereeTicketsConfiguration.location)
                            .loadOrThrow[SmedereeTicketsConfiguration]
                    )
                    _ <- migrateDatabases(hubConfiguration, ticketsConfiguration)
                    _ <- IO {
                        val defaultSecret = "CHANGEME".getBytes(StandardCharsets.UTF_8)
                        if (
                            java.util.Arrays.equals(
                                defaultSecret,
                                hubConfiguration.service.authentication.cookieSecret.toArray
                            )
                        )
                            log.warn("SERVICE IS USING DEFAULT COOKIE SECRET! PLEASE CONFIGURE A SECURE ONE!")
                    }
                    repoCheck <- IO {
                        val repositoriesDirectory = hubConfiguration.service.darcs.repositoriesDirectory.toPath
                        if (Files.exists(repositoriesDirectory)) {
                            if (Files.isDirectory(repositoriesDirectory)) {
                                Right(s"Using repositories directory at: $repositoriesDirectory")
                            } else {
                                Left(
                                    s"Path to repositories directory exists but is not a directory: $repositoriesDirectory"
                                )
                            }
                        } else {
                            log.warn(
                                s"Repositories directory does not exist, trying to create it: $repositoriesDirectory"
                            )
                            Files.createDirectories(
                                repositoriesDirectory,
                                PosixFilePermissions.asFileAttribute(PosixFilePermissions.fromString("rwxr-x---"))
                            )
                            Right(s"Using repositories directory at: $repositoriesDirectory")
                        }
                    }
                    _ <- repoCheck match {
                        case Left(error)    => IO.raiseError(new RuntimeException(error))
                        case Right(message) => IO(log.info(message))
                    }
                    hubTransactor = Transactor.fromDriverManager[IO](
                        driver = hubConfiguration.database.driver,
                        url = hubConfiguration.database.url,
                        user = hubConfiguration.database.user,
                        password = hubConfiguration.database.pass,
                        logHandler = None
                    )
                    ticketsTransactor = Transactor.fromDriverManager[IO](
                        driver = ticketsConfiguration.database.driver,
                        url = ticketsConfiguration.database.url,
                        user = ticketsConfiguration.database.user,
                        password = ticketsConfiguration.database.pass,
                        logHandler = None
                    )
                    ticketServiceApi     = new DoobieTicketServiceApi[IO](ticketsTransactor)
                    ticketLabelsRepo     = new DoobieLabelRepository[IO](ticketsTransactor)
                    ticketMilestonesRepo = new DoobieMilestoneRepository[IO](ticketsTransactor)
                    ticketProjectsRepo   = new DoobieProjectRepository[IO](ticketsTransactor)
                    ticketsRepo          = new DoobieTicketRepository[IO](ticketsTransactor)
                    ticketRoutes = new TicketRoutes[IO](
                        ticketsConfiguration,
                        ticketLabelsRepo,
                        ticketMilestonesRepo,
                        ticketProjectsRepo,
                        ticketsRepo
                    )
                    ticketLabelRoutes = new LabelRoutes[IO](ticketsConfiguration, ticketLabelsRepo, ticketProjectsRepo)
                    ticketMilestoneRoutes = new MilestoneRoutes[IO](
                        ticketsConfiguration,
                        ticketMilestonesRepo,
                        ticketProjectsRepo
                    )
                    cryptoClock = java.time.Clock.systemUTC
                    csrfKey <- loadOrCreateCsrfKey(hubConfiguration.service.csrfKeyFile)
                    csrfOriginCheck = createCsrfOriginCheck(
                        NonEmptyList(hubConfiguration.service.external, List(ticketsConfiguration.external))
                    )
                    csrfBuilder = CSRF[IO, IO](csrfKey, csrfOriginCheck)
                    // Present an error page to the user in case of CSRF check failure that recommends deleting the
                    // cookies.
                    csrfMiddleware = csrfBuilder
                        .withCookieDomain(Option(hubConfiguration.service.external.host.toString))
                        .withCookieName(Constants.csrfCookieName.toString)
                        .withCookiePath(Option("/"))
                        .withCSRFCheck(
                            CSRF.checkCSRFinHeaderAndForm[IO, IO](Constants.csrfCookieName.toString, FunctionK.id)
                        )
                        .withOnFailure(
                            Response[IO](
                                Status.Forbidden,
                                entity = Entity.utf8String(de.smederee.hub.views.html.errors.csrfFailed().body)
                            )
                        )
                        .build
                    signAndValidate = SignAndValidate(hubConfiguration.service.authentication.cookieSecret)
                    assetsRoutes <- resourceServiceBuilder[IO](Constants.assetsPath.path.toAbsolute.toString).toRoutes
                    authenticationRepo = new DoobieAuthenticationRepository[IO](hubTransactor)
                    authenticationWithFallThrough = AuthMiddleware.withFallThrough(
                        authenticateUserWithFallThrough(
                            authenticationRepo,
                            signAndValidate,
                            hubConfiguration.service.authentication.timeouts
                        )
                    )
                    darcsWrapper          = new DarcsCommands[IO](hubConfiguration.service.darcs.executable)
                    emailMiddleware       = new SimpleJavaMailMiddleware(hubConfiguration.service.email)
                    orgRepo               = new DoobieOrganisationRepository[IO](hubTransactor)
                    accountManagementRepo = new DoobieAccountManagementRepository[IO](hubTransactor)
                    accountManagementRoutes = new AccountManagementRoutes[IO](
                        accountManagementRepo,
                        hubConfiguration.service,
                        emailMiddleware,
                        orgRepo,
                        signAndValidate,
                        ticketServiceApi
                    )
                    authenticationRoutes = new AuthenticationRoutes[IO](
                        cryptoClock,
                        hubConfiguration.service.authentication,
                        hubConfiguration.service.external,
                        authenticationRepo,
                        signAndValidate
                    )
                    resetPasswordRepo = new DoobieResetPasswordRepository[IO](hubTransactor)
                    resetPasswordRoutes = new ResetPasswordRoutes[IO](
                        hubConfiguration.service.authentication,
                        authenticationRepo,
                        emailMiddleware,
                        hubConfiguration.service.external,
                        resetPasswordRepo
                    )
                    signUpRepo      = new DoobieSignupRepository[IO](hubTransactor)
                    signUpRoutes    = new SignupRoutes[IO](hubConfiguration.service, signUpRepo)
                    landingPages    = new LandingPageRoutes[IO](hubConfiguration.service)
                    orgRoutes       = new OrganisationRoutes[IO](hubConfiguration.service, orgRepo)
                    vcsMetadataRepo = new DoobieVcsMetadataRepository[IO](hubTransactor)
                    vcsRepoRoutes = new VcsRepositoryRoutes[IO](
                        hubConfiguration.service,
                        darcsWrapper,
                        vcsMetadataRepo,
                        ticketProjectsRepo,
                        orgRepo
                    )
                    protectedRoutesWithFallThrough = authenticationWithFallThrough(
                        authenticationRoutes.protectedRoutes <+>
                            accountManagementRoutes.protectedRoutes <+>
                            resetPasswordRoutes.protectedRoutes <+>
                            signUpRoutes.protectedRoutes <+>
                            ticketLabelRoutes.protectedRoutes <+>
                            ticketMilestoneRoutes.protectedRoutes <+>
                            ticketRoutes.protectedRoutes <+>
                            orgRoutes.protectedRoutes <+>
                            vcsRepoRoutes.protectedRoutes <+>
                            landingPages.protectedRoutes
                    )
                    hubWebService = Router(
                        Constants.assetsPath.path.toAbsolute.toString -> assetsRoutes,
                        "/" -> (protectedRoutesWithFallThrough <+>
                            authenticationRoutes.routes <+>
                            resetPasswordRoutes.routes <+>
                            accountManagementRoutes.routes <+>
                            signUpRoutes.routes <+>
                            ticketLabelRoutes.routes <+>
                            ticketMilestoneRoutes.routes <+>
                            ticketRoutes.routes <+>
                            vcsRepoRoutes.routes <+>
                            landingPages.routes)
                    ).orNotFound
                    // Create our ssh server fiber (or a dummy one if disabled).
                    sshServerProvider = hubConfiguration.service.ssh.enabled match {
                        case false => None
                        case true =>
                            Option(
                                new SshServerProvider(
                                    hubConfiguration.service.darcs,
                                    hubConfiguration.database,
                                    hubConfiguration.service.ssh
                                )
                            )
                    }
                    sshServer = sshServerProvider.fold(IO.unit.as(ExitCode.Success))(
                        _.run().use(server =>
                            IO(log.info(s"SSH-Server started at ${server.getHost()}:${server.getPort()}.")) >> IO.never
                                .as(
                                    ExitCode.Success
                                )
                        )
                    )
                    serviceWithErrorHandling = ErrorHandling.Recover.total(
                        ErrorAction.log(
                            hubWebService,
                            messageFailureLogAction = handleError,
                            serviceErrorLogAction = handleError
                        )
                    )
                    // Create our webserver fiber.
                    resource = EmberServerBuilder
                        .default[IO]
                        .withHost(hubConfiguration.service.host)
                        .withPort(hubConfiguration.service.port)
                        .withHttpApp(csrfMiddleware.validate()(serviceWithErrorHandling))
                        .build
                    webServer = resource.use(server =>
                        IO(log.info("Server started at {}", server.address)) >> IO.never.as(ExitCode.Success)
                    )
                    executeFibers <- (webServer, sshServer).parTupled // We run both fibers.
                    (exitCode, _) = executeFibers
                } yield exitCode
        }
    }
}
