/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.util.UUID

import cats.*
import cats.syntax.all.*
import de.smederee.security.*
import org.http4s.Uri

import scala.util.matching.Regex

/** A organisation id is supposed to be globally unique and maps to the [[java.util.UUID]] type beneath.
  */
opaque type OrganisationId = UUID
object OrganisationId {
    val Format: Regex = "^[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}$".r

    given Eq[OrganisationId]       = Eq.fromUniversalEquals
    given Ordering[OrganisationId] = (x: OrganisationId, y: OrganisationId) => x.compareTo(y)
    given Order[OrganisationId]    = Order.fromOrdering

    /** Create an instance of OrganisationId from the given UUID type.
      *
      * @param source
      *   An instance of type UUID which will be returned as a OrganisationId.
      * @return
      *   The appropriate instance of OrganisationId.
      */
    def apply(source: UUID): OrganisationId = source

    /** Try to create an instance of OrganisationId from the given UUID.
      *
      * @param source
      *   A UUID that should fulfil the requirements to be converted into a OrganisationId.
      * @return
      *   An option to the successfully converted OrganisationId.
      */
    def from(source: UUID): Option[OrganisationId] = Option(source)

    /** Try to create an instance of OrganisationId from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a OrganisationId.
      * @return
      *   An option to the successfully converted OrganisationId.
      */
    def fromString(source: String): Either[String, OrganisationId] =
        Option(source)
            .filter(s => Format.matches(s))
            .flatMap { uuidString =>
                Either.catchNonFatal(UUID.fromString(uuidString)).toOption
            }
            .toRight("Illegal value for OrganisationId!")

    /** Generate a new random organisation id.
      *
      * @return
      *   A organisation id which is pseudo randomly generated.
      */
    def randomOrganisationId: OrganisationId = UUID.randomUUID

    extension (uid: OrganisationId) {
        def toUUID: UUID = uid
    }
}

/** Textual description of an [[Organisation]] limited in length to 254 characters.
  */
opaque type OrganisationDescription = String
object OrganisationDescription {
    val MaximumLength: Int = 254

    /** Create an instance of OrganisationDescription from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a OrganisationDescription.
      * @return
      *   The appropriate instance of OrganisationDescription.
      */
    def apply(source: String): OrganisationDescription = source

    /** Try to create an instance of OrganisationDescription from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a OrganisationDescription.
      * @return
      *   An option to the successfully converted OrganisationDescription.
      */
    def from(source: String): Option[OrganisationDescription] = Option(source).map(_.take(MaximumLength))

}

/** An organisation. Because an organisation can own repositories the same restrictions apply to certain attributes of
  * it like to user accounts. Furthermore an organisation can have multiple owners (upon creation only one) that are
  * allowed to make changes to it.
  *
  * @param oid
  *   The globally unique ID of the organisation.
  * @param name
  *   A unique name which will be used in urls.
  * @param owner
  *   The globally unique ID of the user owning the organisation.
  * @param fullName
  *   An optional full name for the organisation.
  * @param description
  *   An optional description of the organisation.
  * @param isPrivate
  *   A flag indicating if this organisation is private i.e. only visible / accessible for accounts with appropriate
  *   permissions.
  * @param website
  *   An optional uri pointing to a website related to the organisation.
  */
final case class Organisation(
    oid: OrganisationId,
    name: Username,
    owner: UserId,
    fullName: Option[FullName],
    description: Option[OrganisationDescription],
    isPrivate: Boolean,
    website: Option[Uri]
)
