/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.data.*
import cats.syntax.all.*
import de.smederee.hub.forms.*
import de.smederee.hub.forms.types.*
import de.smederee.security.*
import org.http4s.Uri

/** Data container for a form to create or edit an [[Organisation]].
  *
  * @param name
  *   A unique name which will be used in urls.
  * @param owner
  *   The globally unique ID of the user owning the organisation.
  * @param fullName
  *   An optional full name for the organisation.
  * @param description
  *   An optional description of the organisation.
  * @param isPrivate
  *   A flag indicating if this organisation is private i.e. only visible / accessible for accounts with appropriate
  *   permissions.
  * @param website
  *   An optional uri pointing to a website related to the organisation.
  */
final case class OrganisationForm(
    name: Username,
    owner: Option[UserId],
    fullName: Option[FullName],
    description: Option[OrganisationDescription],
    isPrivate: Boolean,
    website: Option[Uri]
)

object OrganisationForm extends FormValidator[OrganisationForm] {
    val fieldDescription: FormField = FormField("description")
    val fieldFullName: FormField    = FormField("full_name")
    val fieldIsPrivate: FormField   = FormField("is_private")
    val fieldName: FormField        = FormField("name")
    val fieldOwner: FormField       = FormField("owner")
    val fieldWebsite: FormField     = FormField("website")

    /** Create a form for editing an organisation from the given organisation data.
      *
      * @param organisation
      *   The organisation that provides the source data for the form.
      * @return
      *   A form filled with the data from the given organisation.
      */
    def from(organisation: Organisation): OrganisationForm =
        OrganisationForm(
            name = organisation.name,
            owner = organisation.owner.some,
            fullName = organisation.fullName,
            description = organisation.description,
            isPrivate = organisation.isPrivate,
            website = organisation.website
        )

    override def validate(data: Map[String, String]): ValidatedNec[FormErrors, OrganisationForm] = {
        val name = data
            .get(fieldName)
            .fold(FormFieldError("No name given!").invalidNec)(s =>
                Username.from(s).fold(FormFieldError("Invalid name!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldName -> es.toList)))
        val owner = data
            .get(fieldOwner)
            .fold(None.validNec)(s =>
                UserId.fromString(s) match {
                    case Left(error) => FormFieldError(error).invalidNec
                    case Right(id)   => id.some.validNec
                }
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldName -> es.toList)))
        val fullName = data
            .get(fieldFullName)
            .filter(_.nonEmpty)
            .fold(None.validNec)(s =>
                FullName.from(s).fold(FormFieldError("Invalid full organisation name!").invalidNec)(_.some.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldName -> es.toList)))
        val description = data
            .get(fieldDescription)
            .filter(_.nonEmpty)
            .fold(None.validNec)(s =>
                OrganisationDescription
                    .from(s)
                    .fold(FormFieldError("Invalid organisation description!").invalidNec)(_.some.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldDescription -> es.toList)))
        val privateFlag: ValidatedNec[FormErrors, Boolean] =
            data.get(fieldIsPrivate).fold(false.validNec)(s => s.matches("true").validNec)
        val website = data
            .get(fieldWebsite)
            .fold(Option.empty[Uri].validNec) { s =>
                if (s.trim.isEmpty)
                    Option.empty[Uri].validNec // Sometimes "empty" strings are sent.
                else
                    Uri
                        .fromString(s)
                        .toOption
                        .fold(FormFieldError("Invalid website URI!").invalidNec) { uri =>
                            uri.scheme match {
                                case Some(Uri.Scheme.http) | Some(Uri.Scheme.https) => Option(uri).validNec
                                case _ => FormFieldError("Invalid website URI!").invalidNec
                            }
                        }
            }
            .leftMap(es => NonEmptyChain.of(Map(fieldWebsite -> es.toList)))
        (name, owner, fullName, description, privateFlag, website).mapN {
            case (name, owner, fullName, description, isPrivate, website) =>
                OrganisationForm(
                    name = name,
                    owner = owner,
                    fullName = fullName,
                    description = description,
                    isPrivate = isPrivate,
                    website = website
                )
        }
    }

    extension (form: OrganisationForm) {

        /** Convert the form class into a stringified map which is used as underlying data type for form handling in the
          * twirl templating library.
          *
          * @return
          *   A stringified map containing the data of the form.
          */
        def toMap: Map[String, String] = {
            val name  = Map(OrganisationForm.fieldName.toString -> form.name.toString)
            val owner = form.owner.fold(Map.empty)(owner => Map(OrganisationForm.fieldOwner.toString -> owner.toString))
            val fullName = form.fullName.fold(Map.empty)(fullName =>
                Map(OrganisationForm.fieldFullName.toString -> fullName.toString)
            )
            val description = form.description.fold(Map.empty)(description =>
                Map(OrganisationForm.fieldDescription.toString -> description.toString)
            )
            val isPrivate =
                if (form.isPrivate) {
                    Map(OrganisationForm.fieldIsPrivate.toString -> "true")
                } else {
                    Map(OrganisationForm.fieldIsPrivate.toString -> "false")
                }
            val website =
                form.website.fold(Map.empty)(website => Map(OrganisationForm.fieldWebsite.toString -> website.toString))
            name ++ owner ++ fullName ++ description ++ isPrivate ++ website
        }
    }
}
