/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.io.IOException
import java.nio.file.FileVisitResult
import java.nio.file.FileVisitor
import java.nio.file.Files

import cats.*
import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.html.LinkTools.*
import de.smederee.hub.RequestHelpers.instances.given
import de.smederee.hub.config.*
import de.smederee.hub.forms.types.*
import de.smederee.i18n.LanguageCode
import de.smederee.security.*
import org.http4s.*
import org.http4s.dsl.Http4sDsl
import org.http4s.headers.*
import org.http4s.implicits.*
import org.http4s.twirl.TwirlInstances.*
import org.slf4j.LoggerFactory

/** Routes for handling organisation related functionality.
  *
  * @param configuration
  *   The hub service configuration.
  * @param orgRepo
  *   A repository for handling database functionality for organisations.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class OrganisationRoutes[F[_]: Async](configuration: ServiceConfig, orgRepo: OrganisationRepository[F])
    extends Http4sDsl[F] {
    private val log                               = LoggerFactory.getLogger(getClass)
    given org.typelevel.log4cats.LoggerFactory[F] = org.typelevel.log4cats.slf4j.Slf4jFactory.create[F]

    private val createOrgPath = uri"/user/settings/organisations/create"
    private val linkConfig    = configuration.external

    /** Delete the given directory recursively. It is checked if the given directory is a direct sub directory of the
      * `repositoriesDirectory` and only if this is the case is the directory removed.
      *
      * @param organisationDirectory
      *   The path on the filesystem to the directory that shall be deleted.
      * @return
      *   `true` if the directory was deleted.
      */
    protected def deleteOrganisationDirectory(organisationDirectory: java.nio.file.Path): F[Boolean] =
        for {
            _            <- Sync[F].delay(log.debug(s"Request to delete organisation dir: $organisationDirectory"))
            reposDirPath <- Sync[F].delay(configuration.darcs.repositoriesDirectory.toPath)
            isSubDir     <- Sync[F].delay(reposDirPath.equals(organisationDirectory.getParent()))
            deleted <-
                Sync[F].delay {
                    if (isSubDir) {
                        Files.walkFileTree(
                            organisationDirectory,
                            new FileVisitor[java.nio.file.Path] {
                                override def visitFileFailed(
                                    file: java.nio.file.Path,
                                    exc: IOException
                                ): FileVisitResult = FileVisitResult.CONTINUE

                                override def visitFile(
                                    file: java.nio.file.Path,
                                    attrs: java.nio.file.attribute.BasicFileAttributes
                                ): FileVisitResult = {
                                    Files.delete(file)
                                    FileVisitResult.CONTINUE
                                }

                                override def preVisitDirectory(
                                    dir: java.nio.file.Path,
                                    attrs: java.nio.file.attribute.BasicFileAttributes
                                ): FileVisitResult = FileVisitResult.CONTINUE

                                override def postVisitDirectory(
                                    dir: java.nio.file.Path,
                                    exc: IOException
                                ): FileVisitResult = {
                                    Files.delete(dir)
                                    FileVisitResult.CONTINUE
                                }
                            }
                        )
                        Files.deleteIfExists(organisationDirectory)
                    } else {
                        log.warn(
                            s"Refused requested removal of directory $organisationDirectory which is not a direct sub directory of the configured repositories directory!"
                        )
                        false
                    }
                }
        } yield deleted

    /** Load an organisation and related metadata from the database if the permissions allow it.
      *
      * @param currentUser
      *   The user account that is requesting access to the repository or None for a guest user.
      * @param organisationName
      *   The name of the organisation.
      * @return
      *   An option holding a tuple of the organisation, the list of user accounts that are administrators for it and
      *   the account of the current organisation owner.
      */
    private def loadOrganisation(
        currentUser: Option[Account]
    )(organisationName: Username): F[Option[(Organisation, List[Account], Account)]] =
        for {
            org    <- orgRepo.findByName(organisationName)
            owner  <- org.map(_.oid).traverse(orgRepo.findOwner)
            admins <- org.map(_.oid).traverse(id => orgRepo.getAdministrators(id).compile.toList)
            orgaAndAdmins = (org, admins, owner.flatten).mapN { case (organisation, administrators, owner) =>
                (organisation, administrators, owner)
            }
            // TODO: Replace with whatever we implement as proper permission model. ;-)
            result = currentUser match {
                case None => orgaAndAdmins.filter(_._1.isPrivate === false)
                case Some(user) =>
                    orgaAndAdmins.filter(tuple =>
                        tuple._1.isPrivate === false || tuple._1.owner === user.uid || tuple._2.exists(_ === user)
                    )
            }
        } yield result

    private val createOrganisation: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / "user" / "settings" / "organisations" / "create" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    _ <- Sync[F].raiseUnless(user.validatedEmail)(
                        new Error(
                            "An unvalidated account is not allowed to create an organisation!"
                        ) // FIXME: Proper error handling!
                    )
                    formData <- Sync[F].delay {
                        urlForm.values.map { t =>
                            val (key, values) = t
                            (
                                key,
                                values.headOption.getOrElse("")
                            ) // Pick the first value (a field might get submitted multiple times)!
                        }
                    }
                    form <- Sync[F].delay(OrganisationForm.validate(formData))
                    // TODO: Cleanup the checking into something like `form.andThen(...)` or so.
                    orgExists <- form.traverse(validForm => orgRepo.findByName(validForm.name))
                    checkedForm = orgExists match {
                        case Validated.Invalid(_) => form
                        case Validated.Valid(possibleOrg) =>
                            possibleOrg.fold(form)(_ =>
                                Map(
                                    OrganisationForm.fieldName -> List(FormFieldError("Name already exists!"))
                                ).invalidNec
                            )
                    }
                    resp <- checkedForm match {
                        case Validated.Invalid(es) =>
                            BadRequest(
                                views.html
                                    .createOrganisation(lang = language)(
                                        createOrgPath,
                                        csrf,
                                        Nil,
                                        "Smederee - Create a new organisation".some,
                                        user
                                    )(
                                        formData,
                                        FormErrors.fromNec(es)
                                    )
                            )
                        case Validated.Valid(newOrganisationForm) =>
                            val organisation = Organisation(
                                oid = OrganisationId.randomOrganisationId,
                                name = newOrganisationForm.name,
                                owner = user.uid,
                                fullName = newOrganisationForm.fullName,
                                description = newOrganisationForm.description,
                                isPrivate = newOrganisationForm.isPrivate,
                                website = newOrganisationForm.website
                            )
                            val targetUri = linkConfig.createFullUri(
                                Uri(path = Uri.Path(Vector(Uri.Path.Segment(s"~${organisation.name.toString}"))))
                            )
                            orgRepo.create(organisation) *> SeeOther(Location(targetUri))
                    }
                } yield resp
            }
    }

    private val deleteOrganisation: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / "user" / "settings" / "organisations" / UsernamePathParameter(
                organisationName
            ) / "delete" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    _ <- Sync[F].raiseUnless(user.validatedEmail)(
                        new Error(
                            "An unvalidated account is not allowed to edit an organisation!"
                        ) // FIXME: Proper error handling!
                    )
                    orgAndAdmins <- loadOrganisation(user.some)(organisationName)
                    orgaData = orgAndAdmins.filter(tuple => tuple._1.owner === user.uid || tuple._2.exists(_ === user))
                    response <- orgaData match {
                        case None => NotFound()
                        case Some((organisation, _, owner)) =>
                            for {
                                formData <- Sync[F].delay {
                                    urlForm.values.map { t =>
                                        val (key, values) = t
                                        (
                                            key,
                                            values.headOption.getOrElse("")
                                        ) // Pick the first value (a field might get submitted multiple times)!
                                    }
                                }
                                userIsSure <- Sync[F].delay(formData.get("i-am-sure").exists(_ === "yes"))
                                response <-
                                    if (owner === user && userIsSure) {
                                        for {
                                            _ <- Sync[F].delay(
                                                log.info(
                                                    s"Going to delete organisation ${organisation.name} as requested by the owner ${user.uid}."
                                                )
                                            )
                                            organisationDir <- Sync[F].delay(
                                                java.nio.file.Paths
                                                    .get(
                                                        configuration.darcs.repositoriesDirectory.toPath.toString,
                                                        organisation.name.toString
                                                    )
                                            )
                                            _ <- deleteOrganisationDirectory(organisationDir)
                                            _ <- orgRepo.delete(organisation.oid)
                                            response <- SeeOther(
                                                Location(linkConfig.createFullUri(Uri(path = Uri.Path.Root)))
                                            )
                                        } yield response
                                    } else
                                        SeeOther(Location(linkConfig.createFullUri(uri"user/settings/organisations")))
                            } yield response
                    }
                } yield response
            }
    }

    private val editOrganisation: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / "user" / "settings" / "organisations" / UsernamePathParameter(
                organisationName
            ) / "edit" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    _ <- Sync[F].raiseUnless(user.validatedEmail)(
                        new Error(
                            "An unvalidated account is not allowed to edit an organisation!"
                        ) // FIXME: Proper error handling!
                    )
                    orgAndAdmins <- loadOrganisation(user.some)(organisationName)
                    orgaData = orgAndAdmins.filter(tuple => tuple._1.owner === user.uid || tuple._2.exists(_ === user))
                    resp <- orgaData match {
                        case None => NotFound()
                        case Some((organisation, admins, owner)) =>
                            for {
                                formData <- Sync[F].delay {
                                    urlForm.values.map { t =>
                                        val (key, values) = t
                                        (
                                            key,
                                            values.headOption.getOrElse("")
                                        ) // Pick the first value (a field might get submitted multiple times)!
                                    }
                                }
                                form <- Sync[F].delay(OrganisationForm.validate(formData))
                                resp <- form match {
                                    case Validated.Invalid(es) =>
                                        val actionBaseUri = uri"user/settings/organisations".addSegment(
                                            s"~${organisation.name.toString}"
                                        )
                                        val deleteOrgPath = linkConfig
                                            .createFullUri(actionBaseUri.addSegment("delete"))
                                            .some
                                            .filter(_ => owner === user)
                                        val editOrgPath =
                                            linkConfig.createFullUri(actionBaseUri.addSegment("edit"))
                                        val possibleOwners = (List(owner, user) ::: admins).distinct
                                        BadRequest(
                                            views.html
                                                .editOrganisation(lang = language)(
                                                    editOrgPath,
                                                    deleteOrgPath,
                                                    csrf,
                                                    possibleOwners,
                                                    Option(s"~$organisationName - edit"),
                                                    user
                                                )(
                                                    formData,
                                                    FormErrors.fromNec(es)
                                                )
                                        )
                                    case Validated.Valid(newOrganisationForm) =>
                                        val updatedOrganisation = organisation.copy(
                                            owner = user.uid,
                                            fullName = newOrganisationForm.fullName,
                                            description = newOrganisationForm.description,
                                            isPrivate = newOrganisationForm.isPrivate,
                                            website = newOrganisationForm.website
                                        )
                                        val targetUri = linkConfig.createFullUri(
                                            Uri(path =
                                                Uri.Path(Vector(Uri.Path.Segment(s"~${organisation.name.toString}")))
                                            )
                                        )
                                        orgRepo.update(updatedOrganisation) *> SeeOther(Location(targetUri))
                                }
                            } yield resp
                    }
                } yield resp
            }
    }

    private val showCreateOrganisationForm: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "user" / "settings" / "organisations" / "create" as user =>
            for {
                csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                resp <- user.validatedEmail match {
                    case false =>
                        Forbidden(
                            views.html.errors
                                .unvalidatedAccount(lang = language)(
                                    csrf,
                                    "Smederee - Account not validated!".some,
                                    user
                                )
                        )
                    case true =>
                        Ok(
                            views.html
                                .createOrganisation(lang = language)(
                                    createOrgPath,
                                    csrf,
                                    Nil,
                                    "Smederee - Create a new organisation".some,
                                    user
                                )()
                        )
                }
            } yield resp
    }

    private val showEditOrganisationForm: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / "user" / "settings" / "organisations" / UsernamePathParameter(
                organisationName
            ) / "edit" as user =>
            for {
                csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                orgAndAdmins <- loadOrganisation(user.some)(organisationName)
                orgaData = orgAndAdmins.filter(tuple => tuple._1.owner === user.uid || tuple._2.exists(_ === user))
                resp <- orgaData match {
                    case None => NotFound("Organisation not found!")
                    case Some((organisation, admins, owner)) =>
                        user.validatedEmail match {
                            case false =>
                                Forbidden(
                                    views.html.errors
                                        .unvalidatedAccount(lang = language)(
                                            csrf,
                                            "Smederee - Account not validated!".some,
                                            user
                                        )
                                )
                            case true =>
                                val actionBaseUri =
                                    uri"user/settings/organisations".addSegment(s"~${organisation.name.toString}")
                                val deleteOrgPath = linkConfig
                                    .createFullUri(actionBaseUri.addSegment("delete"))
                                    .some
                                    .filter(_ => owner === user)
                                val editOrgPath =
                                    linkConfig.createFullUri(actionBaseUri.addSegment("edit"))
                                val possibleOwners = (List(owner, user) ::: admins).distinct
                                val formData       = OrganisationForm.from(organisation).toMap
                                Ok(
                                    views.html
                                        .editOrganisation(lang = language)(
                                            editOrgPath,
                                            deleteOrgPath,
                                            csrf,
                                            possibleOwners,
                                            Option(s"~$organisationName - edit"),
                                            user
                                        )(formData)
                                )
                        }
                }
            } yield resp
    }

    val protectedRoutes =
        showCreateOrganisationForm <+> createOrganisation <+> showEditOrganisationForm <+> deleteOrganisation <+> editOrganisation

}
