/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.syntax.all.*
import de.smederee.hub.config.Constants
import de.smederee.security.CsrfToken
import de.smederee.security.SignedToken
import org.http4s.*

trait RequestHelpers[A] {
    extension (a: A) {

        /** Parse the request cookies and extract the authentication token if it can be found.
          *
          * @return
          *   An option to the authentication token.
          */
        def getAuthenticationToken: Option[SignedToken]

        /** Parse the request cookies and extract the CSRF token if it can be found.
          *
          * @return
          *   An option to the CSRF cookie.
          */
        def getCsrfToken: Option[CsrfToken]
    }
}

object RequestHelpers {
    object instances {
        given RequestHelpers[Request[_]] with {
            extension (r: Request[_]) {

                def getAuthenticationToken: Option[SignedToken] =
                    r.cookies
                        .find(_.name === Constants.authenticationCookieName.toString)
                        .flatMap(cookie => SignedToken.from(cookie.content))

                def getCsrfToken: Option[CsrfToken] =
                    r.cookies
                        .find(_.name === Constants.csrfCookieName.toString)
                        .flatMap(cookie => CsrfToken.from(cookie.content))
            }
        }
    }
}
