/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.OffsetDateTime
import java.time.ZoneOffset

import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.email.*
import de.smederee.html.*
import de.smederee.html.LinkTools.*
import de.smederee.hub.*
import de.smederee.hub.RequestHelpers.instances.given_RequestHelpers_Request
import de.smederee.hub.config.*
import de.smederee.hub.forms.types.FormErrors
import de.smederee.hub.forms.types.FormFieldError
import de.smederee.i18n.LanguageCode
import org.http4s.*
import org.http4s.dsl.Http4sDsl
import org.http4s.headers.Location
import org.http4s.implicits.*
import org.http4s.twirl.TwirlInstances.*
import org.slf4j.LoggerFactory

import scala.concurrent.duration.*

final class ResetPasswordRoutes[F[_]: Async](
    authenticationConfig: AuthenticationConfiguration,
    authenticationRepo: AuthenticationRepository[F],
    emailMiddleware: EmailMiddleware[F],
    external: ExternalUrlConfiguration,
    resetPasswordRepo: ResetPasswordRepository[F]
) extends Http4sDsl[F] {
    private val log = LoggerFactory.getLogger(getClass)

    private val loginPath               = uri"/login"
    private val resetPath               = uri"/forgot-password"
    private val resetChangePasswordPath = uri"/forgot-password/change-password"
    private val resetRequestPath        = uri"/forgot-password/request-email"
    private val resetSentPath           = uri"/forgot-password/email-sent"

    private val passwordResetRequestForLoggedInUsers: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case _ @POST -> Root / "forgot-password" / "request-email" as _ =>
            SeeOther(Location(Uri(path = Uri.Path.Root))) // Redirect already logged in users.
    }

    private val passwordResetRequest: HttpRoutes[F] = HttpRoutes.of {
        case req @ POST -> Root / "forgot-password" / "request-email" =>
            req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf <- Sync[F].delay(req.getCsrfToken)
                    formData <- Sync[F].delay {
                        urlForm.values.map { t =>
                            val (key, values) = t
                            (
                                key,
                                values.headOption.getOrElse("")
                            ) // Pick the first value (a field might get submitted multiple times)!
                        }
                    }
                    form <- Sync[F].delay(ResetPasswordForm.validate(formData))
                    response <- form match {
                        case Validated.Invalid(es) =>
                            BadRequest(
                                views.html
                                    .reset()(resetPath, csrf, title = "Smederee - Reset your account password".some)(
                                        formData,
                                        FormErrors.fromNec(es)
                                    )
                            )
                        case Validated.Valid(resetForm) =>
                            authenticationRepo.findAccountByEmail(resetForm.email).flatMap {
                                case None =>
                                    for {
                                        delay <- Sync[F]
                                            .delay(
                                                scala.util.Random.nextInt(2) + 1
                                            ) // Prevent fast response to avoid email guessing.
                                        _        <- Sync[F].sleep(FiniteDuration(delay, SECONDS))
                                        response <- SeeOther(Location(resetSentPath))
                                    } yield response
                                case Some(user) =>
                                    for {
                                        language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                                        token    <- Sync[F].delay(ResetToken.generate)
                                        _ <- resetPasswordRepo
                                            .setResetPasswordToken(user.uid)(
                                                token,
                                                OffsetDateTime.now(ZoneOffset.UTC).plusDays(3L)
                                            )
                                        from <- Sync[F].delay(FromAddress("noreply@smeder.ee"))
                                        to   <- Sync[F].delay(NonEmptyList.of(user.email.toToAddress))
                                        uri <- Sync[F].delay(
                                            external.createFullUri(resetChangePasswordPath.addSegment(token.toString))
                                        )
                                        subject <- Sync[F]
                                            .delay(
                                                SubjectLine(
                                                    "Smederee - Someone has requested a password reset for your email address."
                                                )
                                            )
                                        body <- Sync[F].delay(
                                            TextBody(views.txt.emails.reset(user, uri).toString)
                                        ) // TODO: extension method?
                                        message <- Sync[F]
                                            .delay(EmailMessage(from, to, List.empty, List.empty, subject, body))
                                        result   <- emailMiddleware.send(message)
                                        _        <- Sync[F].delay(result.leftMap(error => log.error(error)))
                                        response <- SeeOther(Location(resetSentPath))
                                    } yield response
                            }
                    }
                } yield response
            }
    }

    private val requestPasswordReset: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root / "forgot-password" =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            response <- Ok(
                views.html.reset()(resetRequestPath, csrf, title = "Smederee - Reset your account password".some)()
            )
        } yield response
    }

    private val requestPasswordResetForLoggedInUsers: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case _ @GET -> Root / "forgot-password" as _ =>
            SeeOther(Location(Uri(path = Uri.Path.Root))) // Redirect already logged in users.
    }

    private val passwordResetEmailSent: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / "forgot-password" / "email-sent" =>
            for {
                csrf     <- Sync[F].delay(req.getCsrfToken)
                response <- Ok(views.html.resetSent()(csrf, title = "Reset password email sent.".some))
            } yield response
    }

    private val changePassword: HttpRoutes[F] = HttpRoutes.of {
        case req @ POST -> Root / "forgot-password" / "change-password" / ResetTokenPathParameter(token) =>
            req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf <- Sync[F].delay(req.getCsrfToken)
                    formData <- Sync[F].delay {
                        urlForm.values.map { t =>
                            val (key, values) = t
                            (
                                key,
                                values.headOption.getOrElse("")
                            ) // Pick the first value (a field might get submitted multiple times)!
                        }
                    }
                    form <- Sync[F].delay(
                        ChangePasswordForm
                            .validate(formData)
                            .andThen(form =>
                                if (form.token === token) form.validNec
                                else
                                    Map(
                                        ChangePasswordForm.fieldGlobal -> List(
                                            FormFieldError("Invalid password reset token!")
                                        )
                                    ).invalidNec
                            )
                    )
                    changePasswordUri = resetChangePasswordPath.addSegment(token.toString)
                    response <- form match {
                        case Validated.Invalid(es) =>
                            BadRequest(
                                views.html
                                    .changePassword()(
                                        changePasswordUri,
                                        csrf,
                                        title = "Smederee - Change your password.".some,
                                        token
                                    )(
                                        formData,
                                        FormErrors.fromNec(es)
                                    )
                            )
                        case Validated.Valid(changePasswordForm) =>
                            for {
                                user <- resetPasswordRepo.findByNameAndResetPasswordToken(
                                    changePasswordForm.name,
                                    token
                                )
                                _ <- user match {
                                    case None =>
                                        Sync[F].delay(
                                            log.info(
                                                s"Password reset form: No user named ${changePasswordForm.name} found!"
                                            )
                                        )
                                    case Some(user) =>
                                        Sync[F].delay(
                                            log.info(s"Password reset form: Changing password for ${user.name}.")
                                        )
                                }
                                _ <- user.traverse(user =>
                                    resetPasswordRepo.setPassword(user.uid)(changePasswordForm.password.encode)
                                )
                                _ <- user.traverse(user => resetPasswordRepo.removeResetPasswordToken(user.uid))
                                _ <- user.traverse(user =>
                                    authenticationRepo.deleteAllUserSessions(user.uid)
                                ) // Close all sessions.
                                response <- SeeOther(Location(loginPath))
                            } yield response
                    }
                } yield response
            }
    }

    private val changePasswordForm: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / "forgot-password" / "change-password" / ResetTokenPathParameter(token) =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                user <- resetPasswordRepo.findByResetPasswordToken(token)
                response <- user match {
                    case None =>
                        Sync[F].delay(log.debug(s"Requested password reset token $token was not found!")) *> NotFound()
                    case Some(user) =>
                        for {
                            _ <- resetPasswordRepo.removeResetPasswordExpirationDate(
                                user.uid
                            ) // The URL shall only work once!
                            _ <- Sync[F].delay(log.debug(s"Password reset uri called for ${user.email}."))
                            changePasswordUri = resetChangePasswordPath.addSegment(token.toString)
                            response <- Ok(
                                views.html.changePassword()(
                                    changePasswordUri,
                                    csrf,
                                    title = "Smederee - Change your password.".some,
                                    token
                                )()
                            )
                        } yield response
                }
            } yield response
    }

    val protectedRoutes =
        if (authenticationConfig.enabled)
            requestPasswordResetForLoggedInUsers <+> passwordResetRequestForLoggedInUsers
        else
            AuthedRoutes.empty[Account, F]

    val routes =
        if (authenticationConfig.enabled)
            changePassword <+> changePasswordForm <+> requestPasswordReset <+> passwordResetRequest <+> passwordResetEmailSent
        else
            HttpRoutes.empty[F]

}
