/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.*

/** Data describing a RSS channel.
  *
  * @see
  *   https://www.rssboard.org/rss-specification
  *
  * @param title
  *   The name of the channel. It's how people refer to your service. If you have an HTML website that contains the same
  *   information as your RSS file, the title of your channel should be the same as the title of your website.
  * @param link
  *   The URL to the HTML website corresponding to the channel.
  * @param description
  *   Phrase or sentence describing the channel.
  * @param copyright
  *   Copyright notice for content in the channel.
  * @param image
  *   Specifies a GIF, JPEG or PNG image that can be displayed with the channel.
  * @param items
  *   A channel may contain any number of <item>s. An item may represent a "story" -- much like a story in a newspaper
  *   or magazine; if so its description is a synopsis of the story, and the link points to the full story.
  */
final case class RssChannel(
    title: String,
    link: String,
    description: String,
    copyright: Option[String],
    image: Option[RssChannelImage],
    items: List[RssItem]
)

/** Description of an image that may be displayed with an RSS channel.
  *
  * @param url
  *   The URL of a GIF, JPEG or PNG image that represents the channel.
  * @param title
  *   Describes the image, it's used in the ALT attribute of the HTML <img> tag when the channel is rendered in HTML.
  * @param link
  *   The URL of the site, when the channel is rendered, the image is a link to the site. (Note, in practice the image
  *   <title> and <link> should have the same value as the channel's <title> and <link>.
  * @param description
  *   Text that is included in the TITLE attribute of the link formed around the image in the HTML rendering.
  * @param height
  *   The height of the image in pixels.
  * @param width
  *   The width of the image in pixels.
  */
final case class RssChannelImage(
    url: String,
    title: String,
    link: String,
    description: Option[String],
    height: Option[Int],
    width: Option[Int]
)

/** Data describing an item (entry) within a RSS channel.
  *
  * @param title
  *   The title of the item.
  * @param link
  *   The url of the item.
  * @param description
  *   The item synopsis.
  * @param author
  *   The author of the item which should per specification contain the email address but we may choose to not include
  *   it due to privacy reasons.
  * @param guid
  *   A string that uniquely identifies the item. In regard of a repository it might be the full url to a patch hash.
  * @param pubDate
  *   Indicates when the item was published.
  */
final case class RssItem(
    title: String,
    link: String,
    description: String,
    author: Option[String],
    guid: Option[String],
    pubDate: Option[OffsetDateTime]
)
