/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.*

import cats.syntax.all.*
import de.smederee.hub.config.AuthenticationTimeouts
import de.smederee.hub.config.Constants
import de.smederee.security.SignedToken
import org.http4s.HttpDate
import org.http4s.ResponseCookie
import org.http4s.SameSite

import scala.concurrent.duration.FiniteDuration

object SessionHelpers {
    object instances {
        extension (session: Session) {

            /** Check if the session has reached the absolute timeout defined by the given timeouts configuration in
              * regard to the given point in time passed as `currentTime`.
              *
              * @param timeouts
              *   Timeouts related to the authentication and session management.
              * @param currentTime
              *   The current time to compare the `createdAt` field of the session to.
              * @return
              *   Return `true` if the session was created at a point in time that is out of the allowed timeout range.
              */
            def hasReachedAbsoluteTimeout(timeouts: AuthenticationTimeouts)(currentTime: OffsetDateTime): Boolean =
                session.createdAt.compareTo(currentTime.minusSeconds(timeouts.absoluteTimeout.toSeconds)) < 0
        }

        extension (signedToken: SignedToken) {

            /** Create an authentication cookie from the signed token. The mentioned cookie will be named after the
              * related constant from [[de.smederee.hub.config.Constants]] and contain the signed token as value.
              * Furthermore the cookie will be using a strict same-site policy, be http-only and respect the given
              * values for expiration and the secure flag.
              *
              * @param domain
              *   An optional domain that shall be set for the cookie. If it is not set then a "host only" cookie will
              *   be created which is in most cases what you want. Setting a specific domain can be necessary if you
              *   want to share cookies between a domain and possible existing sub domains.
              * @param expires
              *   A duration in which the cookie shall expire. It is optional resulting in the cookie being a
              *   non-persistent "session cookie" if not set.
              * @param secure
              *   A flag indication if the secure flag on the cookie shall be set to allow it only to be transmitted via
              *   encrypted channels (i.e. HTTPS).
              * @return
              *   A response cookie ready to be used as an authentication cookie.
              */
            def toAuthenticationCookie(
                domain: Option[String]
            )(expires: Option[FiniteDuration])(secure: Boolean): ResponseCookie =
                ResponseCookie(
                    name = Constants.authenticationCookieName.toString,
                    content = signedToken.toString,
                    expires = expires.flatMap(duration =>
                        HttpDate
                            .fromZonedDateTime(ZonedDateTime.now(ZoneOffset.UTC).plusSeconds(duration.toSeconds))
                            .toOption
                    ),
                    domain = domain,
                    sameSite = Option(SameSite.Strict),
                    secure = secure,
                    httpOnly = true
                )
        }
    }
}
