/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.html.LinkTools.*
import de.smederee.hub.RequestHelpers.instances.given_RequestHelpers_Request
import de.smederee.hub.config.*
import de.smederee.hub.forms.types.FormErrors
import de.smederee.hub.forms.types.FormFieldError
import de.smederee.security.*
import org.http4s.*
import org.http4s.dsl.Http4sDsl
import org.http4s.headers.Location
import org.http4s.implicits.*
import org.http4s.twirl.TwirlInstances.*
import org.slf4j.LoggerFactory

/** The routes for handling the user signup process.
  *
  * @param configuration
  *   The hub service configuration.
  * @param repo
  *   The database repository providing needed functionality for checking and creating accounts.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class SignupRoutes[F[_]: Async](configuration: ServiceConfig, repo: SignupRepository[F]) extends Http4sDsl[F] {
    private val log          = LoggerFactory.getLogger(getClass)
    private val linkConfig   = configuration.external
    private val signupConfig = configuration.signup
    // The base URI for our site which that be passed into some templates which create links themselfes.
    private val baseUri = linkConfig.createFullUri(Uri())
    // The URL path that shall be used in the `action` field of the form.
    private val signupUri = linkConfig.createFullUri(uri"/signup")

    // Parse the signup form and take appropriate actions like returning errors or creating an account.
    private val parseSignUpForm = HttpRoutes.of[F] { case request @ POST -> Root / "signup" =>
        request.decodeStrict[F, UrlForm] { urlForm =>
            for {
                csrf <- Sync[F].delay(request.getCsrfToken)
                formData <- Sync[F].delay {
                    urlForm.values.map { t =>
                        val (key, values) = t
                        (
                            key,
                            values.headOption.getOrElse("")
                        ) // Pick the first value (a field might get submitted multiple times)!
                    }
                }
                form <- Sync[F].delay(SignupForm.validate(formData))
                // Only check for existing email or username if the first validation was successful.
                // TODO: Find a way to remove the nesting of `Validated[Validated[...]]` to also remove the nested pattern match later on.
                checkExisting <- form.traverse { signupForm =>
                    for {
                        duplicateEmail <- repo.findEmail(signupForm.email)
                        duplicateName  <- repo.findUsername(signupForm.name)
                        validatedEmail = duplicateEmail.fold(signupForm.email.validNec[FormErrors])(_ =>
                            Map(
                                SignupForm.fieldEmail -> List(FormFieldError("This email address is already in use!"))
                            ).invalidNec
                        )
                        validatedName = duplicateName.fold(signupForm.name.validNec[FormErrors])(_ =>
                            Map(
                                SignupForm.fieldName -> List(FormFieldError("This username is already in use!"))
                            ).invalidNec
                        )
                        validatedEmailAndName = (validatedEmail, validatedName).mapN { case (_, _) =>
                            signupForm: SignupForm
                        }
                    } yield validatedEmailAndName
                }
                resp <- checkExisting match {
                    case Validated.Invalid(es) =>
                        BadRequest(
                            views.html
                                .signup()(signupUri, csrf, "Smederee - Sign up for an account".some)(
                                    formData,
                                    FormErrors.fromNec(es)
                                )
                        )
                    case Validated.Valid(innerValidation) =>
                        innerValidation match {
                            case Validated.Invalid(es) =>
                                BadRequest(
                                    views.html
                                        .signup()(signupUri, csrf, "Smederee - Sign up for an account".some)(
                                            formData,
                                            FormErrors.fromNec(es)
                                        )
                                )
                            case Validated.Valid(signupForm) =>
                                for {
                                    uid <- Sync[F].delay(UserId.randomUserId)
                                    account <- Sync[F].delay(
                                        Account(
                                            uid = uid,
                                            name = signupForm.name,
                                            email = signupForm.email,
                                            fullName = None,
                                            validatedEmail = false,
                                            language = None
                                        )
                                    )
                                    hash     <- Sync[F].delay(signupForm.password.encode)
                                    _        <- Sync[F].delay(log.info(s"Going to create account for ${account.name}."))
                                    _        <- repo.createAccount(account, hash)
                                    redirect <- SeeOther(Location(signupUri.addPath("welcome")))
                                } yield redirect
                        }
                }
            } yield resp
        }
    }

    private val parseSignUpFormForLoggedInUsers: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case _ @POST -> Root / "signup" as _ =>
            SeeOther(Location(Uri(path = Uri.Path.Root))) // Redirect already logged in users.
    }

    // Render the signup form.
    private val showSignUpForm = HttpRoutes.of[F] { case request @ GET -> Root / "signup" =>
        for {
            csrf <- Sync[F].delay(request.getCsrfToken)
            resp <- Ok(views.html.signup()(signupUri, csrf, "Smederee - Sign up for an account".some)())
        } yield resp
    }

    private val showSignUpFormForLoggedInUsers: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case _ @GET -> Root / "signup" as _ =>
            SeeOther(Location(Uri(path = Uri.Path.Root))) // Redirect already logged in users.
    }

    private val showWelcomePage: HttpRoutes[F] = HttpRoutes.of { case req @ GET -> Root / "signup" / "welcome" =>
        for {
            csrf <- Sync[F].delay(req.getCsrfToken)
            resp <- Ok(views.html.welcome(baseUri)(csrf, "Thank you and welcome!".some))
        } yield resp
    }

    val protectedRoutes =
        if (signupConfig.enabled)
            showSignUpFormForLoggedInUsers <+> parseSignUpFormForLoggedInUsers
        else
            AuthedRoutes.empty[Account, F]

    val routes =
        if (signupConfig.enabled)
            showSignUpForm <+> parseSignUpForm <+> showWelcomePage
        else
            HttpRoutes.empty[F]

}
