/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub.config

import java.nio.charset.StandardCharsets
import java.nio.file.*

import cats.kernel.Eq
import com.comcast.ip4s.Host
import com.comcast.ip4s.Port
import de.smederee.email.*
import de.smederee.html.ExternalUrlConfiguration
import de.smederee.security.*
import de.smederee.ssh.*
import org.http4s.Uri
import pureconfig.*

import scala.concurrent.duration.FiniteDuration
import scala.util.Try

opaque type CookieName = String
object CookieName {

    /** Create an instance of CookieName from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a CookieName.
      * @return
      *   The appropriate instance of CookieName.
      */
    def apply(source: String): CookieName = source

    /** Try to create an instance of CookieName from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a CookieName.
      * @return
      *   An option to the successfully converted CookieName.
      */
    def from(source: String): Option[CookieName] =
        Option(source).map(_.trim.nonEmpty) match {
            case Some(true) => Option(source.trim)
            case _          => None
        }
}

opaque type DirectoryPath = Path
object DirectoryPath {

    /** Create an instance of DirectoryPath from the given Path type.
      *
      * @param source
      *   An instance of type Path which will be returned as a DirectoryPath.
      * @return
      *   The appropriate instance of DirectoryPath.
      */
    def apply(source: Path): DirectoryPath = source

    /** Try to create an instance of DirectoryPath from the given Path.
      *
      * @param source
      *   A Path that should fulfil the requirements to be converted into a DirectoryPath.
      * @return
      *   An option to the successfully converted DirectoryPath.
      */
    def from(source: Path): Option[DirectoryPath] = Option(source)

    /** Try to create an instance of DirectoryPath from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a DirectoryPath.
      * @return
      *   An option to the successfully converted DirectoryPath.
      */
    def fromString(source: String): Option[DirectoryPath] = Try(Paths.get(source)).toOption

    extension (dir: DirectoryPath) {
        def toPath: Path = dir
    }

}

opaque type FailedAttempts = Int
object FailedAttempts {

    /** Create an instance of FailedAttempts from the given Int type.
      *
      * @param source
      *   An instance of type Int which will be returned as a FailedAttempts.
      * @return
      *   The appropriate instance of FailedAttempts.
      */
    def apply(source: Int): FailedAttempts = source

    /** Try to create an instance of FailedAttempts from the given Int.
      *
      * @param source
      *   A Int that should fulfil the requirements to be converted into a FailedAttempts.
      * @return
      *   An option to the successfully converted FailedAttempts.
      */
    def from(source: Int): Option[FailedAttempts] =
        if (source >= 0)
            Option(source)
        else
            None

    extension (attempts: FailedAttempts) {
        def toInt: Int = attempts
    }

}

/** Global constants which are used throughout the code.
  */
object Constants {
    val assetsPath: Uri                      = Uri(path = Uri.Path(Vector(Uri.Path.Segment("assets"))))
    val authenticationCookieName: CookieName = CookieName("sloetel")
    val csrfCookieName: CookieName           = CookieName("csrf-token")
}

/** Configuration for the authentication service.
  *
  * @param database
  *   The configuration needed to access the database.
  * @param service
  *   The general service configuration specifying how the service is run.
  */
final case class SmedereeHubConfig(database: DatabaseConfig, service: ServiceConfig)

object SmedereeHubConfig {
    val location: ConfigurationPath = ConfigurationPath("hub")

    given Eq[SmedereeHubConfig] = Eq.fromUniversalEquals

    given ConfigReader[SmedereeHubConfig] = ConfigReader.forProduct2("database", "service")(SmedereeHubConfig.apply)
}

/** Configuration specifying the database access.
  *
  * @param driver
  *   The class name of the JDBC driver to be used.
  * @param url
  *   The JDBC connection URL **without** username and password.
  * @param user
  *   The username (login) needed to authenticate against the database.
  * @param pass
  *   The password needed to authenticate against the database.
  */
final case class DatabaseConfig(driver: String, url: String, user: String, pass: String)

object DatabaseConfig {
    given Eq[DatabaseConfig]           = Eq.fromUniversalEquals
    given ConfigReader[DatabaseConfig] = ConfigReader.forProduct4("driver", "url", "user", "pass")(DatabaseConfig.apply)
}

/** The general service configuration which describes how the service will be run.
  *
  * @param host
  *   The hostname on which the service shall listen for requests.
  * @param port
  *   The TCP port number on which the service shall listen for requests.
  * @param csrfKeyFile
  *   A file which contains the key used to build the CSRF protection. If it does not exist then it should be created
  *   with sensible permissions.
  * @param downloadDirectory
  *   A directory into which files are written that are supposed to be downloaded by users (e.g. distribution files of
  *   repositories).
  * @param renderMaximumFileSize
  *   The maximum allowed size of a file in bytes to be displayed / rendered via the web interface.
  * @param authentication
  *   The configuration of the authentication feature.
  * @param billing
  *   The configuration of the billing feature.
  * @param darcs
  *   Configuration for the darcs module for vcs related operations via darcs.
  * @param email
  *   The email middleware configuration for sending email messages.
  * @param external
  *   Settings affecting how the service will communicate several information to the "outside world" e.g. if it runs
  *   behind a reverse proxy.
  * @param signup
  *   The configuration for the signup / registration feature.
  * @param ssh
  *   Settings for the embedded SSH server component.
  * @param tickets
  *   Configuration regarding the integration with the hub service.
  */
final case class ServiceConfig(
    host: Host,
    port: Port,
    csrfKeyFile: Path,
    downloadDirectory: DirectoryPath,
    renderMaximumFileSize: Long,
    authentication: AuthenticationConfiguration,
    billing: BillingConfiguration,
    darcs: DarcsConfiguration,
    email: EmailMiddlewareConfiguration,
    external: ExternalUrlConfiguration,
    signup: SignupConfiguration,
    ssh: SshServerConfiguration,
    tickets: TicketIntegrationConfiguration
)

object ServiceConfig {
    given Eq[ServiceConfig] = Eq.fromUniversalEquals

    given ConfigReader[Host]       = ConfigReader.fromStringOpt[Host](Host.fromString)
    given ConfigReader[Port]       = ConfigReader.fromStringOpt[Port](Port.fromString)
    given ConfigReader[Uri]        = ConfigReader.fromStringOpt(s => Uri.fromString(s).toOption)
    given ConfigReader[Uri.Scheme] = ConfigReader.fromStringOpt(s => Uri.Scheme.fromString(s).toOption)

    given ConfigReader[DirectoryPath] = ConfigReader.fromStringOpt(DirectoryPath.fromString)

    given ConfigReader[EmailServerUsername] = ConfigReader.fromStringOpt[EmailServerUsername](EmailServerUsername.from)
    given ConfigReader[EmailServerPassword] = ConfigReader.fromStringOpt[EmailServerPassword](EmailServerPassword.from)
    given ConfigReader[SmtpTransport] =
        ConfigReader.fromStringOpt[SmtpTransport](string => Try(SmtpTransport.valueOf(string)).toOption)

    given ConfigReader[EmailMiddlewareConfiguration] =
        ConfigReader.forProduct5("host", "port", "transport", "username", "password")(
            EmailMiddlewareConfiguration.apply
        )

    given ConfigReader[ExternalUrlConfiguration] =
        ConfigReader.forProduct4("host", "path", "port", "scheme")(ExternalUrlConfiguration.apply)

    given ConfigReader[ServiceConfig] =
        ConfigReader.forProduct13(
            "host",
            "port",
            "csrf-key-file",
            "download-directory",
            "render-maximum-file-size",
            "authentication",
            "billing",
            "darcs",
            "email",
            "external",
            "signup",
            "ssh",
            "ticket-integration"
        )(ServiceConfig.apply)
}

/** Timeouts related to the authentication and session management.
  *
  * @param absoluteTimeout
  *   The maximum allowed age an authentication session. This setting will affect the invalidation of a session on the
  *   server side. This timeout MUST be triggered regardless of session activity.
  * @param idleTimeout
  *   This timeout defines how long after the last activity a session will remain valid.
  * @param renewalTimeout
  *   The time after which a session will be renewed (a new session ID will be generated).
  */
final case class AuthenticationTimeouts(
    absoluteTimeout: FiniteDuration,
    idleTimeout: FiniteDuration,
    renewalTimeout: FiniteDuration
)

object AuthenticationTimeouts {
    given ConfigReader[AuthenticationTimeouts] =
        ConfigReader.forProduct3("absolute-timeout", "idle-timeout", "renewal-timeout")(
            AuthenticationTimeouts.apply
        )
}

/** Configuration for the authentication feature.
  *
  * @param enabled
  *   Determines if the feature is enabled or not.
  * @param cookieSecret
  *   The secret used for the cookie encryption and validation.
  * @param lockAfter
  *   Determines after how many failed login attempts an account gets locked.
  * @param timeouts
  *   Timeouts for the authentication session.
  */
final case class AuthenticationConfiguration(
    enabled: Boolean,
    cookieSecret: PrivateKey,
    lockAfter: FailedAttempts,
    timeouts: AuthenticationTimeouts
)

object AuthenticationConfiguration {
    given Eq[AuthenticationConfiguration] = Eq.fromUniversalEquals

    given ConfigReader[FailedAttempts] =
        ConfigReader.fromStringOpt { s =>
            if (s.forall(_.isDigit))
                FailedAttempts.from(s.toInt)
            else
                None
        }

    given ConfigReader[PrivateKey] =
        ConfigReader.fromStringOpt(s => PrivateKey.from(s.getBytes(StandardCharsets.UTF_8)))

    given ConfigReader[Uri] = ConfigReader.fromStringOpt(s => Uri.fromString(s).toOption)

    given ConfigReader[AuthenticationConfiguration] =
        ConfigReader.forProduct4("enabled", "cookie-secret", "lock-after", "timeouts")(
            AuthenticationConfiguration.apply
        )
}

/** Billing configuration which has a general enabled flag and further more specific settings.
  *
  * @param enabled
  *   Determines if the billing is enabled.
  */
final case class BillingConfiguration(enabled: Boolean)

object BillingConfiguration {
    given Eq[BillingConfiguration] = Eq.fromUniversalEquals

    given ConfigReader[BillingConfiguration] = ConfigReader.forProduct1("enabled")(BillingConfiguration.apply)
}

/** Configuration for the darcs module for vcs related operations via darcs.
  *
  * @param executable
  *   The path to the darcs binary executable. If not a full path (i.e. just `darcs`) it must be present on the `$PATH`
  *   of the environment under which the server is running.
  * @param repositoriesDirectory
  *   The directory used to store the actual repositories structured after owner.
  *   {{{
  * repositories-directory
  *   \_ user1
  *       \_ repo1
  *       \_ repo2
  *   \_ user2
  *       \_ repo1
  *   }}}
  */
final case class DarcsConfiguration(executable: Path, repositoriesDirectory: DirectoryPath)

object DarcsConfiguration {
    given Eq[DarcsConfiguration] = Eq.fromUniversalEquals

    given ConfigReader[DirectoryPath] = ConfigReader.fromStringOpt(DirectoryPath.fromString)

    given ConfigReader[DarcsConfiguration] =
        ConfigReader.forProduct2("executable", "repositories-directory")(DarcsConfiguration.apply)
}

/** Configuration for the signup feature.
  *
  * @param enabled
  *   Determines if the signup (*read* registration) is enabled.
  */
final case class SignupConfiguration(enabled: Boolean)

object SignupConfiguration {
    given Eq[SignupConfiguration] = Eq.fromUniversalEquals

    given ConfigReader[Uri] = ConfigReader.fromStringOpt(s => Uri.fromString(s).toOption)

    given ConfigReader[SignupConfiguration] = ConfigReader.forProduct1("enabled")(SignupConfiguration.apply)

}

/** Configuration regarding the integration with the ticket service.
  *
  * @param baseUri
  *   The base URI used to build links to the ticket service.
  * @param enabled
  *   A flag indicating if the ticket service integration is enabled.
  */
final case class TicketIntegrationConfiguration(baseUri: Uri, enabled: Boolean)

object TicketIntegrationConfiguration {
    given ConfigReader[Uri] = ConfigReader.fromStringOpt(s => Uri.fromString(s).toOption)
    given ConfigReader[TicketIntegrationConfiguration] =
        ConfigReader.forProduct2("base-uri", "enabled")(TicketIntegrationConfiguration.apply)
}
