/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.ssh

import java.util.UUID

import cats.effect.*
import de.smederee.email.EmailAddress
import de.smederee.hub.*
import de.smederee.security.UserId
import de.smederee.security.Username
import doobie.*
import doobie.implicits.*
import doobie.postgres.implicits.*

final class DoobieSshAuthenticationRepository[F[_]: Sync](tx: Transactor[F]) extends SshAuthenticationRepository[F] {
    given Meta[EmailAddress]      = Meta[String].timap(EmailAddress.apply)(_.toString)
    given Meta[EncodedKeyBytes]   = Meta[String].timap(EncodedKeyBytes.unsafeFrom)(_.toString)
    given Meta[KeyComment]        = Meta[String].timap(KeyComment.apply)(_.toString)
    given Meta[KeyFingerprint]    = Meta[String].timap(KeyFingerprint.apply)(_.toString)
    given Meta[SshKeyType]        = Meta[String].timap(id => SshKeyType.Mappings(id))(_.identifier)
    given Meta[UserId]            = Meta[UUID].timap(UserId.apply)(_.toUUID)
    given Meta[Username]          = Meta[String].timap(Username.apply)(_.toString)
    given Meta[VcsRepositoryName] = Meta[String].timap(VcsRepositoryName.apply)(_.toString)

    override def findSshKey(fingerprint: KeyFingerprint): F[Option[PublicSshKey]] =
        sql"""SELECT id, uid, key_type, key, fingerprint, comment, created_at, last_used_at FROM "hub"."ssh_keys" WHERE fingerprint = $fingerprint"""
            .query[PublicSshKey]
            .option
            .transact(tx)

    override def findVcsRepositoryOwner(name: Username): F[Option[VcsRepositoryOwner]] =
        sql"""SELECT uid, name, email FROM "hub"."accounts" WHERE name = $name LIMIT 1"""
            .query[VcsRepositoryOwner]
            .option
            .transact(tx)

    override def updateLastUsed(keyId: UUID): F[Int] =
        sql"""UPDATE "hub"."ssh_keys" SET last_used_at = NOW() WHERE id = $keyId""".update.run.transact(tx)

}
