/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.ssh

import java.io.InputStream
import java.io.OutputStream
import java.nio.charset.StandardCharsets

import cats.syntax.all.*
import com.comcast.ip4s.*
import org.apache.sshd.server.Environment
import org.apache.sshd.server.ExitCallback
import org.apache.sshd.server.channel.ChannelSession
import org.apache.sshd.server.command.Command
import org.apache.sshd.server.shell.ShellFactory

/** A shell factory for our SSH server which prevents interactive logins and shell sessions. This is stripped down to
  * the bare minimum of needed functionality.
  *
  * @param genericUser
  *   A username for generic access to services for darcs clone, pull and push (e.g. `darcs pull
  *   genericUser@smederee-domain:accountName/repository`).
  * @param host
  *   The hostname/address the SSH server is listening on.
  * @param port
  *   The port number on which the SSH server is listening.
  */
final class NoLogin(genericUser: SshUsername, host: Host, port: Port) extends ShellFactory {
    final val DefaultPort = port"22"
    // Generate a possible needed hint about setting the SSH_PORT environment variable.
    val sshPort =
        if (port === DefaultPort)
            ""
        else
            s"SSH_PORT=${port.toString} "
    val banner =
        s"""
       |  █████████                               █████                                     
       | ███░░░░░███                             ░░███                                      
       |░███    ░░░  █████████████    ██████   ███████   ██████  ████████   ██████   ██████ 
       |░░█████████ ░░███░░███░░███  ███░░███ ███░░███  ███░░███░░███░░███ ███░░███ ███░░███
       | ░░░░░░░░███ ░███ ░███ ░███ ░███████ ░███ ░███ ░███████  ░███ ░░░ ░███████ ░███████ 
       | ███    ░███ ░███ ░███ ░███ ░███░░░  ░███ ░███ ░███░░░   ░███     ░███░░░  ░███░░░  
       |░░█████████  █████░███ █████░░██████ ░░████████░░██████  █████    ░░██████ ░░██████ 
       | ░░░░░░░░░  ░░░░░ ░░░ ░░░░░  ░░░░░░   ░░░░░░░░  ░░░░░░  ░░░░░      ░░░░░░   ░░░░░░  
       |
       |Interactive shell sessions have been disabled.
       |
       |Please use:
       |
       |${sshPort}darcs clone ${genericUser.toString}@${host.toString}:username/repository
     """.stripMargin.replaceAll("\n", "\r\n") + "\r\n"

    @SuppressWarnings(Array("scalafix:DisableSyntax.null", "scalafix:DisableSyntax.var"))
    override def createShell(channel: ChannelSession): Command =
        new Command() {
            private var stdin: InputStream     = null
            private var stdout: OutputStream   = null
            private var stderr: OutputStream   = null
            private var callback: ExitCallback = null

            override def destroy(channel: ChannelSession): Unit            = ()
            override def setErrorStream(errorStream: OutputStream): Unit   = this.stderr = errorStream
            override def setExitCallback(callback: ExitCallback): Unit     = this.callback = callback
            override def setInputStream(inputStream: InputStream): Unit    = this.stdin = inputStream
            override def setOutputStream(outputStream: OutputStream): Unit = this.stdout = outputStream

            override def start(channel: ChannelSession, env: Environment): Unit = {
                // Just print out our banner message and close the connection.
                stderr.write(banner.getBytes(StandardCharsets.UTF_8))
                stderr.flush()
                stdin.close()
                stdout.close()
                stderr.close()
                callback.onExit(127)
            }
        }
}
