/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.ssh

import java.security.PublicKey

import cats.*
import cats.effect.*
import cats.effect.std.Dispatcher
import cats.effect.unsafe.implicits.global
import cats.syntax.all.*
import org.apache.sshd.common.config.keys.KeyUtils
import org.apache.sshd.common.config.keys.PublicKeyEntryResolver
import org.apache.sshd.server.auth.pubkey.PublickeyAuthenticator
import org.apache.sshd.server.session.ServerSession
import org.slf4j.LoggerFactory

import scala.jdk.CollectionConverters.*

/** A custom PublickeyAuthenticator implementation for restricting access via ssh.
  *
  * Currently we follow this flow:
  * {{{
  * 1. Find the key and thus the related user account via the key fingerprint.
  * 2. Find the repository via the requested URI (owner/repo).
  * 3. Check if the user is permitted ssh access to the repository.
  * }}}
  *
  * @todo
  *   Currently we only permit ssh access for the repository owner. This should be extended.
  *
  * @param genericUser
  *   A name which represents a ssh user name that can be used for generic access.
  * @param repository
  *   A repository providing the needed functionality like getting ssh keys and user/repo information.
  */
final class SshAuthenticator(
    genericUser: SshUsername,
    repository: SshAuthenticationRepository[IO]
) extends PublickeyAuthenticator {
    private val log = LoggerFactory.getLogger(getClass)

    override def authenticate(providedUsername: String, providedPublicKey: PublicKey, session: ServerSession): Boolean =
        Dispatcher
            .sequential[IO]
            .use { dispatcher =>
                val check = for {
                    _ <- IO.delay(
                        log.debug(s"Authentication request for $providedUsername from ${session.getRemoteAddress()}.")
                    )
                    convertedName   <- IO.delay(SshUsername.from(providedUsername))
                    usernameIsValid <- IO.delay(convertedName.exists(_ === genericUser))
                    keyIsValid <-
                        if (usernameIsValid)
                            for {
                                fingerprint <- IO.delay(
                                    KeyFingerprint.from(KeyUtils.getFingerPrint(providedPublicKey).drop(7))
                                ) // drop the "SHA256:"
                                _ <- IO.delay(
                                    fingerprint.map(fingerprint =>
                                        log.debug(
                                            s"User provided ${providedPublicKey.getAlgorithm()} key with fingerprint: $fingerprint."
                                        )
                                    )
                                )
                                providedKeyBytes <- IO.delay(providedPublicKey.getEncoded())
                                possibleUserKey <- fingerprint match {
                                    case None              => IO.pure(None)
                                    case Some(fingerprint) => repository.findSshKey(fingerprint)
                                }
                                _ <- possibleUserKey match {
                                    case None =>
                                        IO.delay(
                                            log.warn(
                                                s"No ssh key found in the database for given fingerprint: $fingerprint!"
                                            )
                                        )
                                    case Some(userKey) =>
                                        IO.delay(
                                            log.debug(
                                                s"Found matching ${userKey.keyType} key (owner: ${userKey.ownerId})."
                                            )
                                        )
                                }
                                keyIsValid <- possibleUserKey match {
                                    case Some(userKey) =>
                                        // We create a `AuthorizedKeyEntry` from our stored key and delegate the key validation to mina sshd.
                                        // Also we store the owner id of the key in the server session for later usage.
                                        IO.delay {
                                            userKey.toAuthorizedKeyEntry match {
                                                case Left(error) =>
                                                    log.error(
                                                        s"Could not convert user key ${userKey.fingerprint} to AuthorizedKeyEntry: $error"
                                                    )
                                                    false
                                                case Right(authKey) =>
                                                    session.setAttribute(
                                                        SshServerConfiguration.SshKeyOwnerIdAttribute,
                                                        userKey.ownerId
                                                    )
                                                    val authenticator = PublickeyAuthenticator.fromAuthorizedEntries(
                                                        userKey.fingerprint.toString,
                                                        session,
                                                        List(authKey).asJava,
                                                        PublicKeyEntryResolver.IGNORING
                                                    )
                                                    authenticator.authenticate(
                                                        providedUsername,
                                                        providedPublicKey,
                                                        session
                                                    )
                                            }
                                        }
                                    case _ => IO.pure(false)
                                }
                                _ <- IO.delay {
                                    if (keyIsValid)
                                        log.debug("Keys are matching!")
                                    else
                                        log.info("Stored key and provided one are not matching!")
                                }
                            } yield keyIsValid
                        else
                            for {
                                _ <- IO.delay(log.debug(s"Provided username $providedUsername was not accepted!"))
                            } yield false
                } yield usernameIsValid && keyIsValid
                check
            }
            .unsafeRunSync()

}
