/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.ssh

import java.nio.file.*
import java.util.Collections

import cats.*
import cats.effect.*
import com.comcast.ip4s.*
import de.smederee.hub.config.*
import de.smederee.security.UserId
import de.smederee.security.Username
import doobie.*
import org.apache.sshd.common.AttributeRepository.AttributeKey
import org.apache.sshd.common.file.virtualfs.VirtualFileSystemFactory
import org.apache.sshd.server.SshServer
import org.apache.sshd.server.keyprovider.SimpleGeneratorHostKeyProvider
import org.apache.sshd.sftp.server.*
import pureconfig.*

import scala.util.matching.Regex

opaque type SshUsername = String
object SshUsername {
    given Eq[SshUsername] = Eq.fromUniversalEquals

    val Format: Regex = "^[a-z][a-z0-9]{2,31}$".r

    /** Create an instance of SshUsername from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a SshUsername.
      * @return
      *   The appropriate instance of SshUsername.
      */
    def apply(source: String): SshUsername = source

    /** Try to create an instance of SshUsername from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a SshUsername.
      * @return
      *   An option to the successfully converted SshUsername.
      */
    def from(source: String): Option[SshUsername] = Option(source).filter(string => Format.matches(string))

    extension (sshUsername: SshUsername) {

        /** Convert to a [[Username]] instance. The format of both is identical so we just return the wrapped ssh
          * username string.
          *
          * @return
          *   A proper username derived from the given ssh username.
          */
        def toUsername: Username = Username(sshUsername)
    }

}

/** Configuration for the SSH server.
  *
  * @param enabled
  *   Determines if the SSH server will be enabled (started).
  * @param genericUser
  *   A username for generic access to services for darcs clone, pull and push (e.g. `darcs pull
  *   genericUser@smederee-domain:accountName/repository`).
  * @param host
  *   The hostname/address the SSH server will bind to.
  * @param port
  *   The port number on which the SSH server will listen.
  * @param serverKeyFile
  *   A path to the file from which the server key is loaded and also written to if it needs to be generated.
  */
final case class SshServerConfiguration(
    enabled: Boolean,
    genericUser: SshUsername,
    host: Host,
    port: Port,
    serverKeyFile: Path
)

object SshServerConfiguration {
    // A key marker for storing the owner id of an ssh-key in the ssh server session.
    final val SshKeyOwnerIdAttribute: AttributeKey[UserId] = new AttributeKey[UserId]()

    given Eq[SshServerConfiguration] = Eq.fromUniversalEquals

    given ConfigReader[Host]        = ConfigReader.fromStringOpt[Host](Host.fromString)
    given ConfigReader[Port]        = ConfigReader.fromStringOpt[Port](Port.fromString)
    given ConfigReader[SshUsername] = ConfigReader.fromStringOpt[SshUsername](SshUsername.from)

    given ConfigReader[SshServerConfiguration] =
        ConfigReader.forProduct5("enabled", "generic-user", "host", "port", "server-key-file")(
            SshServerConfiguration.apply
        )
}

/** A ssh server using the [Apache MINA SSHD](https://mina.apache.org/sshd-project/) library and the IO monad from cats
  * effect.
  *
  * @param darcsConfiguration
  *   The configuration needed to properly execute underlying darcs commands and access repository data on the
  *   filesystem.
  * @param databaseConfiguration
  *   The configuration needed to access the database.
  * @param sshConfiguration
  *   The configuration with all values needed to properly configure the underlying Apache sshd server.
  */
final class SshServerProvider(
    darcsConfiguration: DarcsConfiguration,
    databaseConfiguration: DatabaseConfig,
    sshConfiguration: SshServerConfiguration
) {

    /** Create an instance of a ssh server with defaults configure it according to the given configuration.
      *
      * @return
      *   An instance of a Apache MINA SSHD server which is not yet started.
      */
    private def createServer(): SshServer = {
        val transactor = Transactor.fromDriverManager[IO](
            driver = databaseConfiguration.driver,
            url = databaseConfiguration.url,
            user = databaseConfiguration.user,
            password = databaseConfiguration.pass,
            logHandler = None
        )
        val repository = new DoobieSshAuthenticationRepository[IO](transactor)
        val server     = SshServer.setUpDefaultServer()
        // We only set the port but not the hostname for the server because we want it to listen on all addresses.
        server.setPort(sshConfiguration.port.toString.toInt)
        val keyProvider = new SimpleGeneratorHostKeyProvider(sshConfiguration.serverKeyFile)
        keyProvider.setAlgorithm("EC")
        keyProvider.setOverwriteAllowed(false)
        server.setKeyPairProvider(keyProvider)
        server.setPublickeyAuthenticator(new SshAuthenticator(sshConfiguration.genericUser, repository))
        server.setShellFactory(
            new NoLogin(sshConfiguration.genericUser, sshConfiguration.host, sshConfiguration.port)
        )
        server.setFileSystemFactory(
            new VirtualFileSystemFactory(darcsConfiguration.repositoriesDirectory.toPath)
        )
        // Add our custom sftp subsystem to provide more performant access for darcs operations.
        val sftpFileSystemAccessor = new DarcsSftpFileSystemAccessor(darcsConfiguration, repository)
        val sftpSubsystem = new SftpSubsystemFactory.Builder().withFileSystemAccessor(sftpFileSystemAccessor).build()
        server.setSubsystemFactories(Collections.singletonList(sftpSubsystem))
        // Add our custom command factory which must provide darcs and scp functionality.
        val darcsCommand = new DarcsSshCommandFactory(darcsConfiguration, repository)
        server.setCommandFactory(darcsCommand)
        server
    }

    def run(): Resource[IO, SshServer] =
        Resource.make {
            for {
                server <- IO.delay(createServer())
                _      <- IO.delay(server.start())
            } yield server
        }(server => IO.delay(server.stop()))

}
