/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.util.UUID

import cats.*
import cats.data.*
import cats.syntax.all.*

/** A submitter id is supposed to be globally unique and maps to the [[java.util.UUID]] type beneath.
  */
opaque type AssigneeId = UUID
object AssigneeId {
    val Format = "^[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}$".r

    given Eq[AssigneeId] = Eq.fromUniversalEquals

    /** Create an instance of AssigneeId from the given UUID type.
      *
      * @param source
      *   An instance of type UUID which will be returned as a AssigneeId.
      * @return
      *   The appropriate instance of AssigneeId.
      */
    def apply(source: UUID): AssigneeId = source

    /** Try to create an instance of AssigneeId from the given UUID.
      *
      * @param source
      *   A UUID that should fulfil the requirements to be converted into a AssigneeId.
      * @return
      *   An option to the successfully converted AssigneeId.
      */
    def from(source: UUID): Option[AssigneeId] = Option(source)

    /** Try to create an instance of AssigneeId from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a AssigneeId.
      * @return
      *   An option to the successfully converted AssigneeId.
      */
    def fromString(source: String): Either[String, AssigneeId] =
        Option(source)
            .filter(s => Format.matches(s))
            .flatMap { uuidString =>
                Either.catchNonFatal(UUID.fromString(uuidString)).toOption
            }
            .toRight("Illegal value for AssigneeId!")

    /** Generate a new random user id.
      *
      * @return
      *   A user id which is pseudo randomly generated.
      */
    def randomAssigneeId: AssigneeId = UUID.randomUUID

    extension (uid: AssigneeId) {
        def toUUID: UUID = uid
    }
}

/** A submitter name has to obey several restrictions which are similiar to the ones found for Unix usernames. It must
  * start with a letter, contain only alphanumeric characters, be at least 2 and at most 31 characters long and be all
  * lowercase.
  */
opaque type AssigneeName = String
object AssigneeName {
    given Eq[AssigneeName]       = Eq.fromUniversalEquals
    given Order[AssigneeName]    = Order.from((x: AssigneeName, y: AssigneeName) => x.compareTo(y))
    given Ordering[AssigneeName] = implicitly[Order[AssigneeName]].toOrdering

    val isAlphanumeric = "^[a-z][a-z0-9]+$".r

    /** Create an instance of AssigneeName from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a AssigneeName.
      * @return
      *   The appropriate instance of AssigneeName.
      */
    def apply(source: String): AssigneeName = source

    /** Try to create an instance of AssigneeName from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a AssigneeName.
      * @return
      *   An option to the successfully converted AssigneeName.
      */
    def from(s: String): Option[AssigneeName] = validate(s).toOption

    /** Validate the given string and return either the validated username or a list of errors.
      *
      * @param s
      *   An arbitrary string which should be a username.
      * @return
      *   Either a list of errors or the validated username.
      */
    def validate(s: String): ValidatedNec[String, AssigneeName] =
        Option(s).map(_.trim.nonEmpty) match {
            case Some(true) =>
                val input = s.trim
                val miniumLength =
                    if (input.length >= 2)
                        input.validNec
                    else
                        "AssigneeName too short (min. 2 characters)!".invalidNec
                val maximumLength =
                    if (input.length < 32)
                        input.validNec
                    else
                        "AssigneeName too long (max. 31 characters)!".invalidNec
                val alphanumeric =
                    if (isAlphanumeric.matches(input))
                        input.validNec
                    else
                        "AssigneeName must be all lowercase alphanumeric characters and start with a character.".invalidNec
                (miniumLength, maximumLength, alphanumeric).mapN { case (_, _, name) =>
                    name
                }
            case _ => "AssigneeName must not be empty!".invalidNec
        }
}

/** Extractor to retrieve an AssigneeName from a path parameter.
  */
object AssigneeNamePathParameter {
    def unapply(str: String): Option[AssigneeName] =
        Option(str).flatMap { string =>
            if (string.startsWith("~"))
                AssigneeName.from(string.drop(1))
            else
                None
        }
}

/** The assignee for a ticket i.e. the person supposed to be working on it.
  *
  * @param id
  *   A globally unique ID identifying the assignee.
  * @param name
  *   The name associated with the assignee which is supposed to be unique.
  */
final case class Assignee(id: AssigneeId, name: AssigneeName)

object Assignee {
    given Eq[Assignee] = Eq.fromUniversalEquals
}
