/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import cats.effect.*
import doobie.*
import doobie.implicits.*
import fs2.Stream
import org.slf4j.LoggerFactory

final class DoobieLabelRepository[F[_]: Sync](tx: Transactor[F]) extends LabelRepository[F] {
    private val log = LoggerFactory.getLogger(getClass)

    given LogHandler[F] = Slf4jLogHandler.createLogHandler[F](log)

    given Meta[ColourCode]       = Meta[String].timap(ColourCode.apply)(_.toString)
    given Meta[LabelDescription] = Meta[String].timap(LabelDescription.apply)(_.toString)
    given Meta[LabelId]          = Meta[Long].timap(LabelId.apply)(_.toLong)
    given Meta[LabelName]        = Meta[String].timap(LabelName.apply)(_.toString)
    given Meta[ProjectId]        = Meta[Long].timap(ProjectId.apply)(_.toLong)

    override def allLabels(projectId: ProjectId): Stream[F, Label] =
        sql"""SELECT id, name, description, colour FROM tickets.labels WHERE project = $projectId ORDER BY name ASC"""
            .query[Label]
            .stream
            .transact(tx)

    override def createLabel(projectId: ProjectId)(label: Label): F[Int] =
        sql"""INSERT INTO tickets.labels
          (
            project,
            name,
            description,
            colour
          )
          VALUES (
            $projectId,
            ${label.name},
            ${label.description},
            ${label.colour}
          )""".update.run.transact(tx)

    override def deleteLabel(label: Label): F[Int] =
        label.id match {
            case None => Sync[F].pure(0)
            case Some(id) =>
                sql"""DELETE FROM tickets.labels WHERE id = $id""".update.run.transact(tx)
        }

    override def findLabel(projectId: ProjectId)(name: LabelName): F[Option[Label]] =
        sql"""SELECT id, name, description, colour FROM tickets.labels WHERE project = $projectId AND name = $name LIMIT 1"""
            .query[Label]
            .option
            .transact(tx)

    override def updateLabel(label: Label): F[Int] =
        label.id match {
            case None => Sync[F].pure(0)
            case Some(id) =>
                sql"""UPDATE tickets.labels 
              SET name = ${label.name},
              description = ${label.description},
              colour = ${label.colour}
              WHERE id = $id""".update.run.transact(tx)
        }
}
