/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.util.UUID

import cats.effect.*
import cats.syntax.all.*
import doobie.*
import doobie.Fragments.*
import doobie.implicits.*
import doobie.postgres.implicits.*
import fs2.Stream
import org.slf4j.LoggerFactory

final class DoobieMilestoneRepository[F[_]: Sync](tx: Transactor[F]) extends MilestoneRepository[F] {
    private val log = LoggerFactory.getLogger(getClass)

    given LogHandler[F] = Slf4jLogHandler.createLogHandler[F](log)

    given Meta[AssigneeId]           = Meta[UUID].timap(AssigneeId.apply)(_.toUUID)
    given Meta[AssigneeName]         = Meta[String].timap(AssigneeName.apply)(_.toString)
    given Meta[ColourCode]           = Meta[String].timap(ColourCode.apply)(_.toString)
    given Meta[LabelDescription]     = Meta[String].timap(LabelDescription.apply)(_.toString)
    given Meta[LabelId]              = Meta[Long].timap(LabelId.apply)(_.toLong)
    given Meta[LabelName]            = Meta[String].timap(LabelName.apply)(_.toString)
    given Meta[MilestoneDescription] = Meta[String].timap(MilestoneDescription.apply)(_.toString)
    given Meta[MilestoneId]          = Meta[Long].timap(MilestoneId.apply)(_.toLong)
    given Meta[MilestoneTitle]       = Meta[String].timap(MilestoneTitle.apply)(_.toString)
    given Meta[ProjectId]            = Meta[Long].timap(ProjectId.apply)(_.toLong)
    given Meta[SubmitterId]          = Meta[UUID].timap(SubmitterId.apply)(_.toUUID)
    given Meta[SubmitterName]        = Meta[String].timap(SubmitterName.apply)(_.toString)
    given Meta[TicketContent]        = Meta[String].timap(TicketContent.apply)(_.toString)
    given Meta[TicketId]             = Meta[Long].timap(TicketId.apply)(_.toLong)
    given Meta[TicketNumber]         = Meta[Int].timap(TicketNumber.apply)(_.toInt)
    given Meta[TicketResolution]     = Meta[String].timap(TicketResolution.valueOf)(_.toString)
    given Meta[TicketStatus]         = Meta[String].timap(TicketStatus.valueOf)(_.toString)
    given Meta[TicketTitle]          = Meta[String].timap(TicketTitle.apply)(_.toString)

    private val selectTicketColumns =
        fr"""SELECT
            "tickets".number AS number,
            "tickets".title AS title,
            "tickets".content AS content,
            "tickets".status AS status,
            "tickets".resolution AS resolution,
            "submitters".uid AS submitter_uid,
            "submitters".name AS submitter_name,
            "tickets".created_at AS created_at,
            "tickets".updated_at AS updated_at
          FROM "tickets"."tickets" AS "tickets"
          LEFT OUTER JOIN "tickets"."users" AS "submitters"
          ON "tickets".submitter = "submitters".uid"""

    override def allMilestones(projectId: ProjectId): Stream[F, Milestone] =
        sql"""SELECT id, title, description, due_date, closed FROM "tickets"."milestones" WHERE project = $projectId ORDER BY due_date ASC, title ASC"""
            .query[Milestone]
            .stream
            .transact(tx)

    override def allTickets(filter: Option[TicketFilter])(milestoneId: MilestoneId): Stream[F, Ticket] = {
        val milestoneFilter =
            fr""""tickets".id IN (SELECT ticket FROM "tickets".milestone_tickets AS "milestone_tickets" WHERE milestone = $milestoneId)"""
        val tickets = filter match {
            case None => selectTicketColumns ++ whereAnd(milestoneFilter)
            case Some(filter) =>
                val numberFilter = filter.number.toNel.map(numbers => Fragments.in(fr""""tickets".number""", numbers))
                val statusFilter = filter.status.toNel.map(status => Fragments.in(fr""""tickets".status""", status))
                val resolutionFilter =
                    filter.resolution.toNel.map(resolutions => Fragments.in(fr""""tickets".resolution""", resolutions))
                val submitterFilter =
                    filter.submitter.toNel.map(submitters => Fragments.in(fr""""submitters".name""", submitters))
                selectTicketColumns ++ whereAndOpt(
                    milestoneFilter.some,
                    numberFilter,
                    statusFilter,
                    resolutionFilter,
                    submitterFilter
                )
        }
        tickets.query[Ticket].stream.transact(tx)
    }

    override def closeMilestone(milestoneId: MilestoneId): F[Int] =
        sql"""UPDATE "tickets"."milestones" SET closed = TRUE WHERE id = $milestoneId""".update.run.transact(tx)

    override def createMilestone(projectId: ProjectId)(milestone: Milestone): F[Int] =
        sql"""INSERT INTO "tickets"."milestones"
          (
            project,
            title,
            due_date,
            description,
            closed
          )
          VALUES (
            $projectId,
            ${milestone.title},
            ${milestone.dueDate},
            ${milestone.description},
            ${milestone.closed}
          )""".update.run.transact(tx)

    override def deleteMilestone(milestone: Milestone): F[Int] =
        milestone.id match {
            case None     => Sync[F].pure(0)
            case Some(id) => sql"""DELETE FROM "tickets"."milestones" WHERE id = $id""".update.run.transact(tx)
        }

    override def findMilestone(projectId: ProjectId)(title: MilestoneTitle): F[Option[Milestone]] =
        sql"""SELECT id, title, description, due_date, closed FROM "tickets"."milestones" WHERE project = $projectId AND title = $title LIMIT 1"""
            .query[Milestone]
            .option
            .transact(tx)

    override def openMilestone(milestoneId: MilestoneId): F[Int] =
        sql"""UPDATE "tickets"."milestones" SET closed = FALSE WHERE id = $milestoneId""".update.run.transact(tx)

    override def updateMilestone(milestone: Milestone): F[Int] =
        milestone.id match {
            case None => Sync[F].pure(0)
            case Some(id) =>
                sql"""UPDATE "tickets"."milestones"
              SET title = ${milestone.title},
                due_date = ${milestone.dueDate},
                description = ${milestone.description}
              WHERE id = $id""".update.run.transact(tx)
        }

}
