/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import cats.*
import cats.syntax.all.*

import scala.util.matching.Regex

opaque type LabelId = Long
object LabelId {
    given Eq[LabelId]       = Eq.fromUniversalEquals
    given Ordering[LabelId] = (x: LabelId, y: LabelId) => x.compareTo(y)
    given Order[LabelId]    = Order.fromOrdering

    val Format: Regex = "^-?\\d+$".r

    /** Create an instance of LabelId from the given Long type.
      *
      * @param source
      *   An instance of type Long which will be returned as a LabelId.
      * @return
      *   The appropriate instance of LabelId.
      */
    def apply(source: Long): LabelId = source

    /** Try to create an instance of LabelId from the given Long.
      *
      * @param source
      *   A Long that should fulfil the requirements to be converted into a LabelId.
      * @return
      *   An option to the successfully converted LabelId.
      */
    def from(source: Long): Option[LabelId] = Option(source)

    /** Try to create an instance of LabelId from the given String.
      *
      * @param source
      *   A string that should fulfil the requirements to be converted into a LabelId.
      * @return
      *   An option to the successfully converted LabelId.
      */
    def fromString(source: String): Option[LabelId] = Option(source).filter(Format.matches).map(_.toLong).flatMap(from)

    extension (id: LabelId) {
        def toLong: Long = id
    }
}

/** Extractor to retrieve an LabelId from a path parameter.
  */
object LabelIdPathParameter {
    def unapply(str: String): Option[LabelId] = Option(str).flatMap(LabelId.fromString)
}

/** A short descriptive name for the label which is supposed to be unique in a project context. It must not be empty and
  * not exceed 40 characters in length.
  */
opaque type LabelName = String
object LabelName {
    given Eq[LabelName]       = Eq.fromUniversalEquals
    given Ordering[LabelName] = (x: LabelName, y: LabelName) => x.compareTo(y)
    given Order[LabelName]    = Order.fromOrdering[LabelName]

    val MaxLength: Int = 40

    /** Create an instance of LabelName from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a LabelName.
      * @return
      *   The appropriate instance of LabelName.
      */
    def apply(source: String): LabelName = source

    /** Try to create an instance of LabelName from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a LabelName.
      * @return
      *   An option to the successfully converted LabelName.
      */
    def from(source: String): Option[LabelName] =
        Option(source).filter(string => string.nonEmpty && string.length <= MaxLength)

}

/** Extractor to retrieve an LabelName from a path parameter.
  */
object LabelNamePathParameter {
    def unapply(str: String): Option[LabelName] = Option(str).flatMap(LabelName.from)
}

/** A maybe needed description of a label which must not be empty and not exceed 254 characters in length.
  */
opaque type LabelDescription = String
object LabelDescription {
    given Eq[LabelDescription] = Eq.fromUniversalEquals

    val MaxLength: Int = 254

    /** Create an instance of LabelDescription from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a LabelDescription.
      * @return
      *   The appropriate instance of LabelDescription.
      */
    def apply(source: String): LabelDescription = source

    /** Try to create an instance of LabelDescription from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a LabelDescription.
      * @return
      *   An option to the successfully converted LabelDescription.
      */
    def from(source: String): Option[LabelDescription] =
        Option(source).filter(string => string.nonEmpty && string.length <= MaxLength)
}

/** An HTML colour code in hexadecimal notation e.g. `#12ab3f`. It must match the format of starting with a hashtag
  * followed by three 2-digit hexadecimal codes (`00-ff`).
  */
opaque type ColourCode = String
object ColourCode {
    given Eq[ColourCode] = Eq.instance((a, b) => a.equalsIgnoreCase(b))

    val Format: Regex = "^#[0-9a-fA-F]{6}$".r

    /** Create an instance of ColourCode from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a ColourCode.
      * @return
      *   The appropriate instance of ColourCode.
      */
    def apply(source: String): ColourCode = source

    /** Try to create an instance of ColourCode from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a ColourCode.
      * @return
      *   An option to the successfully converted ColourCode.
      */
    def from(source: String): Option[ColourCode] = Option(source).filter(string => Format.matches(string))

}

/** A label is intended to mark tickets with keywords and colours to allow filtering on them.
  *
  * @param id
  *   An optional attribute containing the unique internal database ID for the label.
  * @param name
  *   A short descriptive name for the label which is supposed to be unique in a project context.
  * @param description
  *   An optional description if needed.
  * @param colour
  *   A hexadecimal HTML colour code which can be used to mark the label on a rendered website.
  */
final case class Label(id: Option[LabelId], name: LabelName, description: Option[LabelDescription], colour: ColourCode)

object Label {
    given Eq[Label] =
        Eq.instance((thisLabel, thatLabel) =>
            thisLabel.id === thatLabel.id &&
                thisLabel.name === thatLabel.name &&
                thisLabel.description === thatLabel.description &&
                thisLabel.colour === thatLabel.colour
        )
}
