/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import fs2.Stream

/** The base class that defines the needed functionality to handle labels within a database.
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class LabelRepository[F[_]] {

    /** Return all labels associated with the given repository.
      *
      * @param projectId
      *   The unique internal ID of a vcs repository metadata entry for which all labels shall be returned.
      * @return
      *   A stream of labels associated with the vcs repository which may be empty.
      */
    def allLabels(projectId: ProjectId): Stream[F, Label]

    /** Create a database entry for the given label definition.
      *
      * @param projectId
      *   The unique internal ID of a vcs repository metadata entry to which the label belongs.
      * @param label
      *   The label definition that shall be written to the database.
      * @return
      *   The number of affected database rows.
      */
    def createLabel(projectId: ProjectId)(label: Label): F[Int]

    /** Delete the label from the database.
      *
      * @param label
      *   The label definition that shall be deleted from the database.
      * @return
      *   The number of affected database rows.
      */
    def deleteLabel(label: Label): F[Int]

    /** Find the label with the given name for the given vcs repository.
      *
      * @param projectId
      *   The unique internal ID of a vcs repository metadata entry to which the label belongs.
      * @param name
      *   The name of the label which is must be unique in the context of the repository.
      * @return
      *   An option to the found label.
      */
    def findLabel(projectId: ProjectId)(name: LabelName): F[Option[Label]]

    /** Update the database entry for the given label.
      *
      * @param label
      *   The label definition that shall be updated within the database.
      * @return
      *   The number of affected database rows.
      */
    def updateLabel(label: Label): F[Int]

}
