/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import cats.*
import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.html.*
import de.smederee.html.LinkTools.*
import de.smederee.hub.Account
import de.smederee.hub.RequestHelpers.instances.given
import de.smederee.i18n.LanguageCode
import de.smederee.security.CsrfToken
import de.smederee.security.Username
import de.smederee.tickets.config.*
import de.smederee.tickets.forms.types.*
import org.http4s.*
import org.http4s.dsl.Http4sDsl
import org.http4s.headers.Location
import org.http4s.twirl.TwirlInstances.*
import org.slf4j.LoggerFactory

/** Routes for managing labels (basically CRUD functionality).
  *
  * @param configuration
  *   The ticket service configuration.
  * @param labelRepo
  *   A repository for handling database operations for labels.
  * @param projectRepo
  *   A repository for handling database operations regarding our projects and their metadata.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class LabelRoutes[F[_]: Async](
    configuration: SmedereeTicketsConfiguration,
    labelRepo: LabelRepository[F],
    projectRepo: ProjectRepository[F]
) extends Http4sDsl[F] {
    private val log = LoggerFactory.getLogger(getClass)

    private val linkToHubService = configuration.hub.baseUri
    private val linkConfig       = configuration.external

    /** Logic for rendering a list of all labels for a project and optionally management functionality.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of labels shall be rendered.
      * @param projectOwnerName
      *   The username of the account who owns the project.
      * @param projectName
      *   The name of the project.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    private def doShowLabels(
        csrf: Option[CsrfToken]
    )(user: Option[Account])(projectOwnerName: Username)(projectName: ProjectName): F[Response[F]] =
        for {
            _            <- Sync[F].delay(log.debug(s"doShowLabels: $csrf, $user, $projectOwnerName, $projectName"))
            language     <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            projectAndId <- loadProject(user)(projectOwnerName, projectName)
            resp <- projectAndId match {
                case Some((project, projectId)) =>
                    for {
                        labels <- labelRepo.allLabels(projectId).compile.toList
                        projectBaseUri <- Sync[F].delay(
                            linkConfig.createFullUri(
                                Uri(path =
                                    Uri.Path(
                                        Vector(
                                            Uri.Path.Segment(s"~$projectOwnerName"),
                                            Uri.Path.Segment(projectName.toString)
                                        )
                                    )
                                )
                            )
                        )
                        resp <- Ok(
                            views.html.editLabels(lang = language)(
                                projectBaseUri.addSegment("labels"),
                                csrf,
                                linkToHubService,
                                labels,
                                projectBaseUri,
                                "Manage your project labels.".some,
                                user,
                                project
                            )()
                        )
                    } yield resp
                case _ => NotFound("Ticket project not found!")
            }
        } yield resp

    /** Load the project metadata with the given owner and name from the database and return it and its primary key id
      * if the project exists and is readable by the given user account.
      *
      * @param currentUser
      *   The user account that is requesting access to the project or None for a guest user.
      * @param projectOwnerName
      *   The name of the account that owns the project.
      * @param projectName
      *   The name of the project. A project name must start with a letter or number and must contain only alphanumeric
      *   ASCII characters as well as minus or underscore signs. It must be between 2 and 64 characters long.
      * @return
      *   An option to a tuple holding the [[Project]] and its primary key id.
      */
    private def loadProject(
        currentUser: Option[Account]
    )(projectOwnerName: ProjectOwnerName, projectName: ProjectName): F[Option[(Project, ProjectId)]] =
        for {
            owner <- projectRepo.findProjectOwner(projectOwnerName)
            loadedRepo <- owner match {
                case None => Sync[F].pure(None)
                case Some(owner) =>
                    (
                        projectRepo.findProject(owner, projectName),
                        projectRepo.findProjectId(owner, projectName)
                    ).mapN {
                        case (Some(project), Some(projectId)) => (project, projectId).some
                        case _                                => None
                    }
            }
            // TODO: Replace with whatever we implement as proper permission model. ;-)
            projectAndId = currentUser match {
                case None => loadedRepo.filter(tuple => tuple._1.isPrivate === false)
                case Some(user) =>
                    loadedRepo.filter(tuple =>
                        tuple._1.isPrivate === false || tuple._1.owner.uid === ProjectOwnerId.fromUserId(user.uid)
                    )
            }
        } yield projectAndId

    private val addLabel: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "labels" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                val response =
                    for {
                        csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                        language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                        projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                        resp <- projectAndId match {
                            case Some(project, projectId) =>
                                for {
                                    _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                        new Error("Only maintainers may add labels!")
                                    )
                                    formData <- Sync[F].delay(urlForm.values)
                                    form     <- Sync[F].delay(LabelForm.validate(formData))
                                    labels <- projectAndId.traverse(tuple =>
                                        labelRepo.allLabels(tuple._2).compile.toList
                                    )
                                    projectBaseUri <- Sync[F].delay(
                                        linkConfig.createFullUri(
                                            Uri(path =
                                                Uri.Path(
                                                    Vector(
                                                        Uri.Path.Segment(s"~$projectOwnerName"),
                                                        Uri.Path.Segment(projectName.toString)
                                                    )
                                                )
                                            )
                                        )
                                    )
                                    resp <- form match {
                                        case Validated.Invalid(errors) =>
                                            BadRequest(
                                                views.html.editLabels(lang = language)(
                                                    projectBaseUri.addSegment("labels"),
                                                    csrf,
                                                    linkToHubService,
                                                    labels.getOrElse(List.empty),
                                                    projectBaseUri,
                                                    "Manage your project labels.".some,
                                                    user.some,
                                                    project
                                                )(formData.withDefaultValue(Chain.empty), FormErrors.fromNec(errors))
                                            )
                                        case Validated.Valid(labelData) =>
                                            val label =
                                                Label(None, labelData.name, labelData.description, labelData.colour)
                                            for {
                                                checkDuplicate <- labelRepo.findLabel(projectId)(labelData.name)
                                                resp <- checkDuplicate match {
                                                    case None =>
                                                        labelRepo.createLabel(projectId)(label) *> SeeOther(
                                                            Location(projectBaseUri.addSegment("labels"))
                                                        )
                                                    case Some(_) =>
                                                        BadRequest(
                                                            views.html.editLabels(lang = language)(
                                                                projectBaseUri.addSegment("labels"),
                                                                csrf,
                                                                linkToHubService,
                                                                labels.getOrElse(List.empty),
                                                                projectBaseUri,
                                                                "Manage your project labels.".some,
                                                                user.some,
                                                                project
                                                            )(
                                                                formData.withDefaultValue(Chain.empty),
                                                                Map(
                                                                    LabelForm.fieldName -> List(
                                                                        FormFieldError(
                                                                            "A label with that name already exists!"
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                }
                                            } yield resp
                                    }
                                } yield resp
                            case _ => NotFound()
                        }
                    } yield resp
                response.recoverWith { error =>
                    log.error("Internal Server Error", error)
                    for {
                        csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                        language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                        resp <- InternalServerError(
                            views.html.errors.internalServerError(lang = language)(csrf, user.some)
                        )
                    } yield resp
                }
            }
    }

    private val deleteLabel: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "labels" / LabelNamePathParameter(labelName) / "delete" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                    projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                    resp <- projectAndId match {
                        case Some(project, projectId) =>
                            for {
                                _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                    new Error("Only maintainers may add labels!")
                                )
                                label <- labelRepo.findLabel(projectId)(labelName)
                                resp <- label match {
                                    case Some(label) =>
                                        for {
                                            formData <- Sync[F].delay(urlForm.values)
                                            projectBaseUri <- Sync[F].delay(
                                                linkConfig.createFullUri(
                                                    Uri(path =
                                                        Uri.Path(
                                                            Vector(
                                                                Uri.Path.Segment(s"~$projectOwnerName"),
                                                                Uri.Path.Segment(projectName.toString)
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                            userIsSure <- Sync[F].delay(
                                                formData
                                                    .get("i-am-sure")
                                                    .map(_.headOption.exists(_ === "yes"))
                                                    .getOrElse(false)
                                            )
                                            labelIdMatches <- Sync[F].delay(
                                                formData
                                                    .get(LabelForm.fieldId)
                                                    .map(
                                                        _.headOption
                                                            .flatMap(LabelId.fromString)
                                                            .exists(id => label.id.exists(_ === id))
                                                    )
                                                    .getOrElse(false)
                                            )
                                            labelNameMatches <- Sync[F].delay(
                                                formData
                                                    .get(LabelForm.fieldName)
                                                    .map(_.headOption.flatMap(LabelName.from).exists(_ === labelName))
                                                    .getOrElse(false)
                                            )
                                            resp <- (labelIdMatches && labelNameMatches && userIsSure) match {
                                                case false => BadRequest("Invalid form data!")
                                                case true =>
                                                    labelRepo.deleteLabel(label) *> SeeOther(
                                                        Location(projectBaseUri.addSegment("labels"))
                                                    )
                                            }
                                        } yield resp
                                    case _ => NotFound("Label not found!")
                                }
                            } yield resp
                        case _ => NotFound("Repository not found!")
                    }
                } yield resp
            }
    }

    private val editLabel: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "labels" / LabelNamePathParameter(labelName) as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                val response =
                    for {
                        csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                        language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                        projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                        label <- projectAndId match {
                            case Some((_, projectId)) => labelRepo.findLabel(projectId)(labelName)
                            case _                    => Sync[F].delay(None)
                        }
                        resp <- (projectAndId, label) match {
                            case (Some(project, projectId), Some(label)) =>
                                for {
                                    _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                        new Error("Only maintainers may add labels!")
                                    )
                                    projectBaseUri <- Sync[F].delay(
                                        linkConfig.createFullUri(
                                            Uri(path =
                                                Uri.Path(
                                                    Vector(
                                                        Uri.Path.Segment(s"~$projectOwnerName"),
                                                        Uri.Path.Segment(projectName.toString)
                                                    )
                                                )
                                            )
                                        )
                                    )
                                    actionUri <- Sync[F].delay(
                                        projectBaseUri.addSegment("labels").addSegment(label.name.toString)
                                    )
                                    formData <- Sync[F].delay(urlForm.values)
//                labelIdMatches <- Sync[F].delay(
//                  formData
//                    .get(LabelForm.fieldId)
//                    .flatMap(LabelId.fromString)
//                    .exists(id => label.id.exists(_ === id)) match {
//                    case false =>
//                      NonEmptyChain
//                        .of(Map(LabelForm.fieldGlobal -> List(FormFieldError("Label ID does not match!"))))
//                        .invalidNec
//                    case true => label.id.validNec
//                  }
//                )
                                    form <- Sync[F].delay(LabelForm.validate(formData))
                                    resp <- form match {
                                        case Validated.Invalid(errors) =>
                                            BadRequest(
                                                views.html.editLabel(lang = language)(
                                                    actionUri,
                                                    csrf,
                                                    linkToHubService,
                                                    label,
                                                    projectBaseUri,
                                                    s"Edit label ${label.name}".some,
                                                    user,
                                                    project
                                                )(
                                                    formData.toMap.withDefaultValue(Chain.empty),
                                                    FormErrors.fromNec(errors)
                                                )
                                            )
                                        case Validated.Valid(labelData) =>
                                            val updatedLabel =
                                                label.copy(
                                                    name = labelData.name,
                                                    description = labelData.description,
                                                    colour = labelData.colour
                                                )
                                            for {
                                                checkDuplicate <- labelRepo.findLabel(projectId)(updatedLabel.name)
                                                resp <- checkDuplicate.filterNot(_.id === updatedLabel.id) match {
                                                    case None =>
                                                        labelRepo.updateLabel(updatedLabel) *> SeeOther(
                                                            Location(projectBaseUri.addSegment("labels"))
                                                        )
                                                    case Some(_) =>
                                                        BadRequest(
                                                            views.html.editLabel(lang = language)(
                                                                actionUri,
                                                                csrf,
                                                                linkToHubService,
                                                                label,
                                                                projectBaseUri,
                                                                s"Edit label ${label.name}".some,
                                                                user,
                                                                project
                                                            )(
                                                                formData.toMap.withDefaultValue(Chain.empty),
                                                                Map(
                                                                    LabelForm.fieldName -> List(
                                                                        FormFieldError(
                                                                            "A label with that name already exists!"
                                                                        )
                                                                    )
                                                                )
                                                            )
                                                        )
                                                }
                                            } yield resp
                                    }
                                } yield resp
                            case _ => NotFound()
                        }
                    } yield resp
                response.recoverWith { error =>
                    log.error("Internal Server Error", error)
                    for {
                        csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                        language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                        resp <- InternalServerError(
                            views.html.errors.internalServerError(lang = language)(csrf, user.some)
                        )
                    } yield resp
                }
            }
    }

    private val showEditLabelForm: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "labels" / LabelNamePathParameter(labelName) / "edit" as user =>
            for {
                csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                label <- projectAndId match {
                    case Some((_, projectId)) => labelRepo.findLabel(projectId)(labelName)
                    case _                    => Sync[F].delay(None)
                }
                resp <- (projectAndId, label) match {
                    case (Some(project, _), Some(label)) =>
                        for {
                            projectBaseUri <- Sync[F].delay(
                                linkConfig.createFullUri(
                                    Uri(path =
                                        Uri.Path(
                                            Vector(
                                                Uri.Path.Segment(s"~$projectOwnerName"),
                                                Uri.Path.Segment(projectName.toString)
                                            )
                                        )
                                    )
                                )
                            )
                            actionUri <- Sync[F].delay(
                                projectBaseUri.addSegment("labels").addSegment(label.name.toString)
                            )
                            formData <- Sync[F].delay(LabelForm.fromLabel(label))
                            resp <- Ok(
                                views.html
                                    .editLabel()(
                                        actionUri,
                                        csrf,
                                        linkToHubService,
                                        label,
                                        projectBaseUri,
                                        s"Edit label ${label.name}".some,
                                        user,
                                        project
                                    )(
                                        formData.toMap
                                    )
                            )
                        } yield resp
                    case _ => NotFound()
                }
            } yield resp
    }

    private val showEditLabelsPage: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "labels" as user =>
            val response =
                for {
                    csrf <- Sync[F].delay(ar.req.getCsrfToken)
                    resp <- doShowLabels(csrf)(user.some)(projectOwnerName)(projectName)
                } yield resp
            response.recoverWith { error =>
                log.error("Internal Server Error", error)
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    resp <- InternalServerError(views.html.errors.internalServerError(lang = language)(csrf, user.some))
                } yield resp
            }
    }

    private val showLabelsForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "labels" =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowLabels(csrf)(None)(projectOwnerName)(projectName)
            } yield resp
    }

    val protectedRoutes = addLabel <+> deleteLabel <+> editLabel <+> showEditLabelForm <+> showEditLabelsPage

    val routes = showLabelsForGuests

}
