/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.time.LocalDate

import cats.*
import cats.syntax.all.*

import scala.util.matching.Regex

opaque type MilestoneId = Long
object MilestoneId {
    given Eq[MilestoneId] = Eq.fromUniversalEquals

    val Format: Regex = "^-?\\d+$".r

    /** Create an instance of MilestoneId from the given Long type.
      *
      * @param source
      *   An instance of type Long which will be returned as a MilestoneId.
      * @return
      *   The appropriate instance of MilestoneId.
      */
    def apply(source: Long): MilestoneId = source

    /** Try to create an instance of MilestoneId from the given Long.
      *
      * @param source
      *   A Long that should fulfil the requirements to be converted into a MilestoneId.
      * @return
      *   An option to the successfully converted MilestoneId.
      */
    def from(source: Long): Option[MilestoneId] = Option(source)

    /** Try to create an instance of MilestoneId from the given String.
      *
      * @param source
      *   A string that should fulfil the requirements to be converted into a MilestoneId.
      * @return
      *   An option to the successfully converted MilestoneId.
      */
    def fromString(source: String): Option[MilestoneId] =
        Option(source).filter(Format.matches).map(_.toLong).flatMap(from)

    extension (id: MilestoneId) {
        def toLong: Long = id
    }
}

/** Extractor to retrieve an MilestoneId from a path parameter.
  */
object MilestoneIdPathParameter {
    def unapply(str: String): Option[MilestoneId] = Option(str).flatMap(MilestoneId.fromString)
}

/** A title for a milestone, usually a version number, a word or a short phrase that is supposed to be unique within a
  * project context. It must not be empty and not exceed 64 characters in length.
  */
opaque type MilestoneTitle = String
object MilestoneTitle {
    given Eq[MilestoneTitle]       = Eq.fromUniversalEquals
    given Ordering[MilestoneTitle] = (x: MilestoneTitle, y: MilestoneTitle) => x.compareTo(y)
    given Order[MilestoneTitle]    = Order.fromOrdering[MilestoneTitle]

    val MaxLength: Int = 64

    /** Create an instance of MilestoneTitle from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a MilestoneTitle.
      * @return
      *   The appropriate instance of MilestoneTitle.
      */
    def apply(source: String): MilestoneTitle = source

    /** Try to create an instance of MilestoneTitle from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a MilestoneTitle.
      * @return
      *   An option to the successfully converted MilestoneTitle.
      */
    def from(source: String): Option[MilestoneTitle] =
        Option(source).filter(string => string.nonEmpty && string.length <= MaxLength)

}

/** Extractor to retrieve an MilestoneTitle from a path parameter.
  */
object MilestoneTitlePathParameter {
    def unapply(str: String): Option[MilestoneTitle] = Option(str).flatMap(MilestoneTitle.from)
}

/** A longer detailed description of a project milestone which must not be empty.
  */
opaque type MilestoneDescription = String
object MilestoneDescription {
    given Eq[MilestoneDescription] = Eq.fromUniversalEquals

    /** Create an instance of MilestoneDescription from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a MilestoneDescription.
      * @return
      *   The appropriate instance of MilestoneDescription.
      */
    def apply(source: String): MilestoneDescription = source

    /** Try to create an instance of MilestoneDescription from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a MilestoneDescription.
      * @return
      *   An option to the successfully converted MilestoneDescription.
      */
    def from(source: String): Option[MilestoneDescription] = Option(source).map(_.trim).filter(_.nonEmpty)

}

/** A milestone can be used to organise tickets and progress inside a project.
  *
  * @param id
  *   An optional attribute containing the unique internal database ID for the milestone.
  * @param title
  *   A title for the milestone, usually a version number, a word or a short phrase that is supposed to be unique within
  *   a project context.
  * @param description
  *   An optional longer description of the milestone.
  * @param dueDate
  *   An optional date on which the milestone is supposed to be reached.
  * @param closed
  *   This flag indicates if the milestone is closed e.g. considered done or obsolete.
  */
final case class Milestone(
    id: Option[MilestoneId],
    title: MilestoneTitle,
    description: Option[MilestoneDescription],
    dueDate: Option[LocalDate],
    closed: Boolean
)

object Milestone {

    given Eq[LocalDate] = Eq.instance((a, b) => a.compareTo(b) === 0)

    given Eq[Milestone] =
        Eq.instance((a, b) =>
            a.id === b.id && a.title === b.title && a.dueDate === b.dueDate && a.description === b.description
        )

}
