/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import fs2.Stream

/** The base class that defines the needed functionality to handle milestones within a database.
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class MilestoneRepository[F[_]] {

    /** Return all milestones associated with the given repository.
      *
      * @param projectId
      *   The unique internal ID of a vcs repository metadata entry for which all milestones shall be returned.
      * @return
      *   A stream of milestones associated with the vcs repository which may be empty.
      */
    def allMilestones(projectId: ProjectId): Stream[F, Milestone]

    /** Return all tickets associated with the given milestone.
      *
      * @param filter
      *   A ticket filter containing possible values which will be used to filter the list of tickets.
      * @param milestoneId
      *   The unique internal ID of a milestone for which all tickets shall be returned.
      * @return
      *   A stream of tickets associated with the vcs repository which may be empty.
      */
    def allTickets(filter: Option[TicketFilter])(milestoneId: MilestoneId): Stream[F, Ticket]

    /** Change the milestone status with the given id to closed.
      *
      * @param milestoneId
      *   The unique internal ID of a milestone for which all tickets shall be returned.
      * @return
      *   The number of affected database rows.
      */
    def closeMilestone(milestoneId: MilestoneId): F[Int]

    /** Create a database entry for the given milestone definition.
      *
      * @param projectId
      *   The unique internal ID of a vcs repository metadata entry to which the milestone belongs.
      * @param milestone
      *   The milestone definition that shall be written to the database.
      * @return
      *   The number of affected database rows.
      */
    def createMilestone(projectId: ProjectId)(milestone: Milestone): F[Int]

    /** Delete the milestone from the database.
      *
      * @param milestone
      *   The milestone definition that shall be deleted from the database.
      * @return
      *   The number of affected database rows.
      */
    def deleteMilestone(milestone: Milestone): F[Int]

    /** Find the milestone with the given title for the given vcs repository.
      *
      * @param projectId
      *   The unique internal ID of a vcs repository metadata entry to which the milestone belongs.
      * @param title
      *   The title of the milestone which is must be unique in the context of the repository.
      * @return
      *   An option to the found milestone.
      */
    def findMilestone(projectId: ProjectId)(title: MilestoneTitle): F[Option[Milestone]]

    /** Change the milestone status with the given id to open.
      *
      * @param milestoneId
      *   The unique internal ID of a milestone for which all tickets shall be returned.
      * @return
      *   The number of affected database rows.
      */
    def openMilestone(milestoneId: MilestoneId): F[Int]

    /** Update the database entry for the given milestone.
      *
      * @param milestone
      *   The milestone definition that shall be updated within the database.
      * @return
      *   The number of affected database rows.
      */
    def updateMilestone(milestone: Milestone): F[Int]

}
