/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import cats.*
import cats.data.*
import cats.effect.*
import cats.syntax.all.*
import de.smederee.html.*
import de.smederee.html.LinkTools.*
import de.smederee.hub.Account
import de.smederee.hub.RequestHelpers.instances.given
import de.smederee.i18n.LanguageCode
import de.smederee.security.CsrfToken
import de.smederee.security.Username
import de.smederee.tickets.config.*
import de.smederee.tickets.forms.types.*
import org.http4s.*
import org.http4s.dsl.Http4sDsl
import org.http4s.headers.Location
import org.http4s.twirl.TwirlInstances.*
import org.slf4j.LoggerFactory

/** Routes for managing milestones (basically CRUD functionality).
  *
  * @param configuration
  *   The ticket service configuration.
  * @param milestoneRepo
  *   A repository for handling database operations for milestones.
  * @param projectRepo
  *   A repository for handling database operations regarding our vcs repositories and their metadata.
  * @tparam F
  *   A higher kinded type providing needed functionality, which is usually an IO monad like Async or Sync.
  */
final class MilestoneRoutes[F[_]: Async](
    configuration: SmedereeTicketsConfiguration,
    milestoneRepo: MilestoneRepository[F],
    projectRepo: ProjectRepository[F]
) extends Http4sDsl[F] {
    private val log = LoggerFactory.getLogger(getClass)

    private val linkToHubService = configuration.hub.baseUri
    private val linkConfig       = configuration.external

    /** Logic for rendering a detail page for a single milestone.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of tickets shall be rendered.
      * @param projectOwnerName
      *   The username of the account who owns the project.
      * @param projectName
      *   The name of the project.
      * @param milestoneTitle
      *   The title of the milestone that shall be rendered.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    private def doShowMilestone(csrf: Option[CsrfToken])(user: Option[Account])(projectOwnerName: Username)(
        projectName: ProjectName
    )(milestoneTitle: MilestoneTitle): F[Response[F]] =
        for {
            language     <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            projectAndId <- loadProject(user)(projectOwnerName, projectName)
            milestone <- projectAndId match {
                case Some((_, projectId)) => milestoneRepo.findMilestone(projectId)(milestoneTitle)
                case _                    => Sync[F].delay(None)
            }
            resp <- (projectAndId, milestone) match {
                case (Some(project, _), Some(milestone)) =>
                    for {
                        tickets <- milestone.id.traverse(milestoneId =>
                            milestoneRepo.allTickets(None)(milestoneId).compile.toList
                        )
                        projectBaseUri <- Sync[F].delay(
                            linkConfig.createFullUri(
                                Uri(path =
                                    Uri.Path(
                                        Vector(
                                            Uri.Path.Segment(s"~$projectOwnerName"),
                                            Uri.Path.Segment(projectName.toString)
                                        )
                                    )
                                )
                            )
                        )
                        actionUri <- Sync[F].delay(
                            projectBaseUri.addSegment("milestones").addSegment(milestone.title.toString)
                        )
                        renderedDescription <- Sync[F].delay(
                            milestone.description.map(_.toString).map(MarkdownRenderer.render)
                        )
                        resp <- Ok(
                            views.html.showMilestone(lang = language)(
                                actionUri,
                                csrf,
                                linkToHubService,
                                milestone,
                                renderedDescription,
                                projectBaseUri,
                                tickets.getOrElse(Nil),
                                s"Milestone ${milestone.title}".some,
                                user,
                                project
                            )
                        )
                    } yield resp
                case _ => NotFound()
            }
        } yield resp

    /** Logic for rendering a list of all milestones for a project and optionally management functionality.
      *
      * @param csrf
      *   An optional CSRF-Token that shall be used.
      * @param user
      *   An optional user account for whom the list of milestones shall be rendered.
      * @param projectOwnerName
      *   The username of the account who owns the project.
      * @param projectName
      *   The name of the project.
      * @return
      *   An HTTP response containing the rendered HTML.
      */
    private def doShowMilestones(
        csrf: Option[CsrfToken]
    )(user: Option[Account])(projectOwnerName: Username)(projectName: ProjectName): F[Response[F]] =
        for {
            language     <- Sync[F].delay(user.flatMap(_.language).getOrElse(LanguageCode("en")))
            projectAndId <- loadProject(user)(projectOwnerName, projectName)
            resp <- projectAndId match {
                case Some((project, projectId)) =>
                    for {
                        milestones <- milestoneRepo.allMilestones(projectId).compile.toList
                        projectBaseUri <- Sync[F].delay(
                            linkConfig.createFullUri(
                                Uri(path =
                                    Uri.Path(
                                        Vector(
                                            Uri.Path.Segment(s"~$projectOwnerName"),
                                            Uri.Path.Segment(projectName.toString)
                                        )
                                    )
                                )
                            )
                        )
                        resp <- Ok(
                            views.html.editMilestones(lang = language)(
                                projectBaseUri.addSegment("milestones"),
                                csrf,
                                linkToHubService,
                                milestones,
                                projectBaseUri,
                                "Manage your project milestones.".some,
                                user,
                                project
                            )()
                        )
                    } yield resp
                case _ => NotFound()
            }
        } yield resp

    /** Load the project metadata with the given owner and name from the database and return it and its primary key id
      * if the project exists and is readable by the given user account.
      *
      * @param currentUser
      *   The user account that is requesting access to the project or None for a guest user.
      * @param projectOwnerName
      *   The name of the account that owns the project.
      * @param projectName
      *   The name of the project. A project name must start with a letter or number and must contain only alphanumeric
      *   ASCII characters as well as minus or underscore signs. It must be between 2 and 64 characters long.
      * @return
      *   An option to a tuple holding the [[Project]] and its primary key id.
      */
    private def loadProject(
        currentUser: Option[Account]
    )(projectOwnerName: Username, projectName: ProjectName): F[Option[(Project, ProjectId)]] =
        for {
            owner <- projectRepo.findProjectOwner(projectOwnerName)
            loadedRepo <- owner match {
                case None => Sync[F].pure(None)
                case Some(owner) =>
                    (
                        projectRepo.findProject(owner, projectName),
                        projectRepo.findProjectId(owner, projectName)
                    ).mapN {
                        case (Some(project), Some(projectId)) => (project, projectId).some
                        case _                                => None
                    }
            }
            // TODO: Replace with whatever we implement as proper permission model. ;-)
            repoAndId = currentUser match {
                case None => loadedRepo.filter(tuple => tuple._1.isPrivate === false)
                case Some(user) =>
                    loadedRepo.filter(tuple =>
                        tuple._1.isPrivate === false || tuple._1.owner.uid === ProjectOwnerId.fromUserId(user.uid)
                    )
            }
        } yield repoAndId

    private val addMilestone: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                    language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                    resp <- projectAndId match {
                        case Some(project, projectId) =>
                            for {
                                _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                    new Error("Only maintainers may add milestones!")
                                )
                                formData <- Sync[F].delay(urlForm.values)
                                form     <- Sync[F].delay(MilestoneForm.validate(formData))
                                milestones <- projectAndId.traverse(tuple =>
                                    milestoneRepo.allMilestones(tuple._2).compile.toList
                                )
                                projectBaseUri <- Sync[F].delay(
                                    linkConfig.createFullUri(
                                        Uri(path =
                                            Uri.Path(
                                                Vector(
                                                    Uri.Path.Segment(s"~$projectOwnerName"),
                                                    Uri.Path.Segment(projectName.toString)
                                                )
                                            )
                                        )
                                    )
                                )
                                resp <- form match {
                                    case Validated.Invalid(errors) =>
                                        BadRequest(
                                            views.html.editMilestones(lang = language)(
                                                projectBaseUri.addSegment("milestones"),
                                                csrf,
                                                linkToHubService,
                                                milestones.getOrElse(List.empty),
                                                projectBaseUri,
                                                "Manage your project milestones.".some,
                                                user.some,
                                                project
                                            )(formData, FormErrors.fromNec(errors))
                                        )
                                    case Validated.Valid(milestoneData) =>
                                        val milestone =
                                            Milestone(
                                                None,
                                                milestoneData.title,
                                                milestoneData.description,
                                                milestoneData.dueDate,
                                                closed = false
                                            )
                                        for {
                                            checkDuplicate <- milestoneRepo.findMilestone(projectId)(
                                                milestoneData.title
                                            )
                                            resp <- checkDuplicate match {
                                                case None =>
                                                    milestoneRepo.createMilestone(projectId)(milestone) *> SeeOther(
                                                        Location(projectBaseUri.addSegment("milestones"))
                                                    )
                                                case Some(_) =>
                                                    BadRequest(
                                                        views.html.editMilestones(lang = language)(
                                                            projectBaseUri.addSegment("milestones"),
                                                            csrf,
                                                            linkToHubService,
                                                            milestones.getOrElse(List.empty),
                                                            projectBaseUri,
                                                            "Manage your project milestones.".some,
                                                            user.some,
                                                            project
                                                        )(
                                                            formData,
                                                            Map(
                                                                MilestoneForm.fieldTitle -> List(
                                                                    FormFieldError(
                                                                        "A milestone with that name already exists!"
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                            }
                                        } yield resp
                                }
                            } yield resp
                        case _ => NotFound()
                    }
                } yield resp
            }
    }

    private val closeMilestone: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" / MilestoneTitlePathParameter(milestoneTitle) / "close" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                    projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                    resp <- projectAndId match {
                        case Some(project, projectId) =>
                            for {
                                _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                    new Error("Only maintainers may close milestones!")
                                )
                                milestone <- milestoneRepo.findMilestone(projectId)(milestoneTitle)
                                resp <- milestone match {
                                    case Some(milestone) =>
                                        for {
                                            formData <- Sync[F].delay {
                                                urlForm.values.map { t =>
                                                    val (key, values) = t
                                                    (
                                                        key,
                                                        values.headOption.getOrElse("")
                                                    ) // Pick the first value (a field might get submitted multiple times)!
                                                }
                                            }
                                            projectBaseUri <- Sync[F].delay(
                                                linkConfig.createFullUri(
                                                    Uri(path =
                                                        Uri.Path(
                                                            Vector(
                                                                Uri.Path.Segment(s"~$projectOwnerName"),
                                                                Uri.Path.Segment(projectName.toString)
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                            milestoneIdMatches <- Sync[F].delay(
                                                formData
                                                    .get(MilestoneForm.fieldId)
                                                    .flatMap(MilestoneId.fromString)
                                                    .exists(id => milestone.id.exists(_ === id))
                                            )
                                            milestoneTitleMatches <- Sync[F].delay(
                                                formData
                                                    .get(MilestoneForm.fieldTitle)
                                                    .flatMap(MilestoneTitle.from)
                                                    .exists(_ === milestoneTitle)
                                            )
                                            resp <- (milestoneIdMatches && milestoneTitleMatches) match {
                                                case false => BadRequest("Invalid form data!")
                                                case true =>
                                                    milestone.id.traverse(milestoneRepo.closeMilestone) *> SeeOther(
                                                        Location(
                                                            projectBaseUri
                                                                .addSegment("milestones")
                                                                .addSegment(milestone.title.toString)
                                                        )
                                                    )
                                            }
                                        } yield resp
                                    case _ => NotFound("Milestone not found!")
                                }
                            } yield resp
                        case _ => NotFound("Repository not found!")
                    }
                } yield resp
            }
    }

    private val deleteMilestone: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" / MilestoneTitlePathParameter(milestoneTitle) / "delete" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                    projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                    resp <- projectAndId match {
                        case Some(project, projectId) =>
                            for {
                                _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                    new Error("Only maintainers may add milestones!")
                                )
                                milestone <- milestoneRepo.findMilestone(projectId)(milestoneTitle)
                                resp <- milestone match {
                                    case Some(milestone) =>
                                        for {
                                            formData <- Sync[F].delay {
                                                urlForm.values.map { t =>
                                                    val (key, values) = t
                                                    (
                                                        key,
                                                        values.headOption.getOrElse("")
                                                    ) // Pick the first value (a field might get submitted multiple times)!
                                                }
                                            }
                                            projectBaseUri <- Sync[F].delay(
                                                linkConfig.createFullUri(
                                                    Uri(path =
                                                        Uri.Path(
                                                            Vector(
                                                                Uri.Path.Segment(s"~$projectOwnerName"),
                                                                Uri.Path.Segment(projectName.toString)
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                            userIsSure <- Sync[F].delay(formData.get("i-am-sure").exists(_ === "yes"))
                                            milestoneIdMatches <- Sync[F].delay(
                                                formData
                                                    .get(MilestoneForm.fieldId)
                                                    .flatMap(MilestoneId.fromString)
                                                    .exists(id => milestone.id.exists(_ === id))
                                            )
                                            milestoneTitleMatches <- Sync[F].delay(
                                                formData
                                                    .get(MilestoneForm.fieldTitle)
                                                    .flatMap(MilestoneTitle.from)
                                                    .exists(_ === milestoneTitle)
                                            )
                                            resp <- (milestoneIdMatches && milestoneTitleMatches && userIsSure) match {
                                                case false => BadRequest("Invalid form data!")
                                                case true =>
                                                    milestoneRepo.deleteMilestone(milestone) *> SeeOther(
                                                        Location(projectBaseUri.addSegment("milestones"))
                                                    )
                                            }
                                        } yield resp
                                    case _ => NotFound("Milestone not found!")
                                }
                            } yield resp
                        case _ => NotFound("Repository not found!")
                    }
                } yield resp
            }
    }

    private val editMilestone: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" / MilestoneTitlePathParameter(milestoneTitle) as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                    language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                    milestone <- projectAndId match {
                        case Some((_, projectId)) => milestoneRepo.findMilestone(projectId)(milestoneTitle)
                        case _                    => Sync[F].delay(None)
                    }
                    resp <- (projectAndId, milestone) match {
                        case (Some(project, projectId), Some(milestone)) =>
                            for {
                                _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                    new Error("Only maintainers may add milestones!")
                                )
                                projectBaseUri <- Sync[F].delay(
                                    linkConfig.createFullUri(
                                        Uri(path =
                                            Uri.Path(
                                                Vector(
                                                    Uri.Path.Segment(s"~$projectOwnerName"),
                                                    Uri.Path.Segment(projectName.toString)
                                                )
                                            )
                                        )
                                    )
                                )
                                actionUri <- Sync[F].delay(
                                    projectBaseUri.addSegment("milestones").addSegment(milestone.title.toString)
                                )
                                formData <- Sync[F].delay(urlForm.values)
                                milestoneIdMatches <- Sync[F].delay(
                                    formData
                                        .get(MilestoneForm.fieldId)
                                        .flatMap(_.headOption.flatMap(MilestoneId.fromString))
                                        .find(id => milestone.id.exists(_ === id))
                                        .fold(
                                            NonEmptyChain
                                                .of(
                                                    Map(
                                                        MilestoneForm.fieldGlobal -> List(
                                                            FormFieldError("Milestone ID does not match!")
                                                        )
                                                    )
                                                )
                                                .invalidNec
                                        )(_ => milestone.id.validNec)
                                )
                                form <- Sync[F].delay(MilestoneForm.validate(formData))
                                resp <- form match {
                                    case Validated.Invalid(errors) =>
                                        BadRequest(
                                            views.html.editMilestone(lang = language)(
                                                actionUri,
                                                csrf,
                                                linkToHubService,
                                                milestone,
                                                projectBaseUri,
                                                s"Edit milestone ${milestone.title}".some,
                                                user,
                                                project
                                            )(
                                                formData,
                                                FormErrors.fromNec(errors)
                                            )
                                        )
                                    case Validated.Valid(milestoneData) =>
                                        val updatedMilestone =
                                            milestone.copy(
                                                title = milestoneData.title,
                                                description = milestoneData.description,
                                                dueDate = milestoneData.dueDate
                                            )
                                        for {
                                            checkDuplicate <- milestoneRepo.findMilestone(projectId)(
                                                updatedMilestone.title
                                            )
                                            resp <- checkDuplicate.filterNot(_.id === updatedMilestone.id) match {
                                                case None =>
                                                    milestoneRepo.updateMilestone(updatedMilestone) *> SeeOther(
                                                        Location(projectBaseUri.addSegment("milestones"))
                                                    )
                                                case Some(_) =>
                                                    BadRequest(
                                                        views.html.editMilestone(lang = language)(
                                                            actionUri,
                                                            csrf,
                                                            linkToHubService,
                                                            milestone,
                                                            projectBaseUri,
                                                            s"Edit milestone ${milestone.title}".some,
                                                            user,
                                                            project
                                                        )(
                                                            formData,
                                                            Map(
                                                                MilestoneForm.fieldTitle -> List(
                                                                    FormFieldError(
                                                                        "A milestone with that name already exists!"
                                                                    )
                                                                )
                                                            )
                                                        )
                                                    )
                                            }
                                        } yield resp
                                }
                            } yield resp
                        case _ => NotFound()
                    }
                } yield resp
            }
    }

    private val openMilestone: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ POST -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" / MilestoneTitlePathParameter(milestoneTitle) / "open" as user =>
            ar.req.decodeStrict[F, UrlForm] { urlForm =>
                for {
                    csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                    projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                    resp <- projectAndId match {
                        case Some(project, projectId) =>
                            for {
                                _ <- Sync[F].raiseUnless(project.owner.uid === ProjectOwnerId.fromUserId(user.uid))(
                                    new Error("Only maintainers may open milestones!")
                                )
                                milestone <- milestoneRepo.findMilestone(projectId)(milestoneTitle)
                                resp <- milestone match {
                                    case Some(milestone) =>
                                        for {
                                            formData <- Sync[F].delay {
                                                urlForm.values.map { t =>
                                                    val (key, values) = t
                                                    (
                                                        key,
                                                        values.headOption.getOrElse("")
                                                    ) // Pick the first value (a field might get submitted multiple times)!
                                                }
                                            }
                                            projectBaseUri <- Sync[F].delay(
                                                linkConfig.createFullUri(
                                                    Uri(path =
                                                        Uri.Path(
                                                            Vector(
                                                                Uri.Path.Segment(s"~$projectOwnerName"),
                                                                Uri.Path.Segment(projectName.toString)
                                                            )
                                                        )
                                                    )
                                                )
                                            )
                                            milestoneIdMatches <- Sync[F].delay(
                                                formData
                                                    .get(MilestoneForm.fieldId)
                                                    .flatMap(MilestoneId.fromString)
                                                    .exists(id => milestone.id.exists(_ === id))
                                            )
                                            milestoneTitleMatches <- Sync[F].delay(
                                                formData
                                                    .get(MilestoneForm.fieldTitle)
                                                    .flatMap(MilestoneTitle.from)
                                                    .exists(_ === milestoneTitle)
                                            )
                                            resp <- (milestoneIdMatches && milestoneTitleMatches) match {
                                                case false => BadRequest("Invalid form data!")
                                                case true =>
                                                    milestone.id.traverse(milestoneRepo.openMilestone) *> SeeOther(
                                                        Location(
                                                            projectBaseUri
                                                                .addSegment("milestones")
                                                                .addSegment(milestone.title.toString)
                                                        )
                                                    )
                                            }
                                        } yield resp
                                    case _ => NotFound("Milestone not found!")
                                }
                            } yield resp
                        case _ => NotFound("Repository not found!")
                    }
                } yield resp
            }
    }

    private val showEditMilestoneForm: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" / MilestoneTitlePathParameter(milestoneTitle) / "edit" as user =>
            for {
                csrf         <- Sync[F].delay(ar.req.getCsrfToken)
                language     <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                projectAndId <- loadProject(user.some)(projectOwnerName, projectName)
                milestone <- projectAndId match {
                    case Some((_, projectId)) => milestoneRepo.findMilestone(projectId)(milestoneTitle)
                    case _                    => Sync[F].delay(None)
                }
                resp <- (projectAndId, milestone) match {
                    case (Some(project, _), Some(milestone)) =>
                        for {
                            projectBaseUri <- Sync[F].delay(
                                linkConfig.createFullUri(
                                    Uri(path =
                                        Uri.Path(
                                            Vector(
                                                Uri.Path.Segment(s"~$projectOwnerName"),
                                                Uri.Path.Segment(projectName.toString)
                                            )
                                        )
                                    )
                                )
                            )
                            actionUri <- Sync[F].delay(
                                projectBaseUri.addSegment("milestones").addSegment(milestone.title.toString)
                            )
                            formData <- Sync[F].delay(MilestoneForm.fromMilestone(milestone))
                            resp <- Ok(
                                views.html.editMilestone(lang = language)(
                                    actionUri,
                                    csrf,
                                    linkToHubService,
                                    milestone,
                                    projectBaseUri,
                                    s"Edit milestone ${milestone.title}".some,
                                    user,
                                    project
                                )(
                                    formData.toMap
                                )
                            )
                        } yield resp
                    case _ => NotFound()
                }
            } yield resp
    }

    private val showEditMilestonesPage: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" as user =>
            for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                resp <- doShowMilestones(csrf)(user.some)(projectOwnerName)(projectName)
            } yield resp
    }

    private val showMilestone: AuthedRoutes[Account, F] = AuthedRoutes.of {
        case ar @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" / MilestoneTitlePathParameter(milestoneTitle) as user =>
            val response = for {
                csrf <- Sync[F].delay(ar.req.getCsrfToken)
                resp <- doShowMilestone(csrf)(user.some)(projectOwnerName)(projectName)(milestoneTitle)
            } yield resp
            response.recoverWith { error =>
                log.error("Internal Server Error", error)
                for {
                    csrf     <- Sync[F].delay(ar.req.getCsrfToken)
                    language <- Sync[F].delay(user.language.getOrElse(LanguageCode("en")))
                    resp <- InternalServerError(views.html.errors.internalServerError(lang = language)(csrf, user.some))
                } yield resp
            }
    }

    private val showMilestoneForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" / MilestoneTitlePathParameter(milestoneTitle) =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowMilestone(csrf)(None)(projectOwnerName)(projectName)(milestoneTitle)
            } yield resp
    }

    private val showMilestonesForGuests: HttpRoutes[F] = HttpRoutes.of {
        case req @ GET -> Root / UsernamePathParameter(projectOwnerName) / ProjectNamePathParameter(
                projectName
            ) / "milestones" =>
            for {
                csrf <- Sync[F].delay(req.getCsrfToken)
                resp <- doShowMilestones(csrf)(None)(projectOwnerName)(projectName)
            } yield resp
    }

    val protectedRoutes =
        addMilestone <+> closeMilestone <+> deleteMilestone <+> editMilestone <+> openMilestone <+> showEditMilestoneForm <+> showEditMilestonesPage <+> showMilestone

    val routes = showMilestoneForGuests <+> showMilestonesForGuests

}
