/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

/** A base class for a database repository that should handle all functionality regarding projects in the database.
  *
  * @tparam F
  *   A higher kinded type which wraps the actual return values.
  */
abstract class ProjectRepository[F[_]] {

    /** Create the given project within the database.
      *
      * @param project
      *   The project that shall be created.
      * @return
      *   The number of affected database rows.
      */
    def createProject(project: Project): F[Int]

    /** Delete the given project from the database.
      *
      * @param project
      *   The project that shall be deleted.
      * @return
      *   The number of affected database rows.
      */
    def deleteProject(project: Project): F[Int]

    /** Search for the project entry with the given owner and name.
      *
      * @param owner
      *   Data about the owner of the project containing information needed to query the database.
      * @param name
      *   The project name which must be unique in regard to the owner.
      * @return
      *   An option to the successfully found project entry.
      */
    def findProject(owner: ProjectOwner, name: ProjectName): F[Option[Project]]

    /** Search for the internal database specific (auto generated) ID of the given owner / project combination which
      * serves as a primary key for the database table.
      *
      * @param owner
      *   Data about the owner of the project containing information needed to query the database.
      * @param name
      *   The project name which must be unique in regard to the owner.
      * @return
      *   An option to the internal database ID.
      */
    def findProjectId(owner: ProjectOwner, name: ProjectName): F[Option[ProjectId]]

    /** Search for a project owner of whom we only know the name.
      *
      * @param name
      *   The name of the project owner which is the username of the actual owners account.
      * @return
      *   An option to successfully found project owner.
      */
    def findProjectOwner(name: ProjectOwnerName): F[Option[ProjectOwner]]

    /** Increment the counter column for the next ticket number and return the old value (i.e. the value _before_ it was
      * incremented).
      *
      * @param projectId
      *   The internal database id of the project.
      * @return
      *   The ticket number _before_ it was incremented.
      */
    def incrementNextTicketNumber(projectId: ProjectId): F[TicketNumber]

    /** Update the database entry for the given project.
      *
      * @param project
      *   The project that shall be updated within the database.
      * @return
      *   The number of affected database rows.
      */
    def updateProject(project: Project): F[Int]

}
