/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.time.OffsetDateTime

import cats.*
import cats.syntax.all.*
import org.http4s.QueryParamDecoder
import org.http4s.QueryParamEncoder
import org.http4s.dsl.impl.OptionalQueryParamDecoderMatcher

import scala.util.matching.Regex

/** An unlimited text field which must be not empty to describe the ticket in great detail if needed.
  */
opaque type TicketContent = String
object TicketContent {

    /** Create an instance of TicketContent from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a TicketContent.
      * @return
      *   The appropriate instance of TicketContent.
      */
    def apply(source: String): TicketContent = source

    /** Try to create an instance of TicketContent from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a TicketContent.
      * @return
      *   An option to the successfully converted TicketContent.
      */
    def from(source: String): Option[TicketContent] = Option(source).filter(_.nonEmpty)

}

opaque type TicketId = Long
object TicketId {
    given Eq[TicketId] = Eq.fromUniversalEquals

    val Format: Regex = "^-?\\d+$".r

    /** Create an instance of TicketId from the given Long type.
      *
      * @param source
      *   An instance of type Long which will be returned as a TicketId.
      * @return
      *   The appropriate instance of TicketId.
      */
    def apply(source: Long): TicketId = source

    /** Try to create an instance of TicketId from the given Long.
      *
      * @param source
      *   A Long that should fulfil the requirements to be converted into a TicketId.
      * @return
      *   An option to the successfully converted TicketId.
      */
    def from(source: Long): Option[TicketId] = Option(source)

    /** Try to create an instance of TicketId from the given String.
      *
      * @param source
      *   A string that should fulfil the requirements to be converted into a TicketId.
      * @return
      *   An option to the successfully converted TicketId.
      */
    def fromString(source: String): Option[TicketId] = Option(source).filter(Format.matches).map(_.toLong).flatMap(from)

    extension (id: TicketId) {
        def toLong: Long = id
    }
}

/** A ticket number maps to an integer beneath and has the requirement to be greater than zero.
  */
opaque type TicketNumber = Int
object TicketNumber {
    given Eq[TicketNumber]       = Eq.fromUniversalEquals
    given Ordering[TicketNumber] = (x: TicketNumber, y: TicketNumber) => x.compareTo(y)
    given Order[TicketNumber]    = Order.fromOrdering

    val Format: Regex = "^-?\\d+$".r

    /** Create an instance of TicketNumber from the given Int type.
      *
      * @param source
      *   An instance of type Int which will be returned as a TicketNumber.
      * @return
      *   The appropriate instance of TicketNumber.
      */
    def apply(source: Int): TicketNumber = source

    /** Try to create an instance of TicketNumber from the given Int.
      *
      * @param source
      *   A Int that should fulfil the requirements to be converted into a TicketNumber.
      * @return
      *   An option to the successfully converted TicketNumber.
      */
    def from(source: Int): Option[TicketNumber] = Option(source).filter(_ > 0)

    /** Try to create an instance of TicketNumber from the given String.
      *
      * @param source
      *   A string that should fulfil the requirements to be converted into a TicketNumber.
      * @return
      *   An option to the successfully converted TicketNumber.
      */
    def fromString(source: String): Option[TicketNumber] =
        Option(source).filter(Format.matches).map(_.toInt).flatMap(from)

    extension (number: TicketNumber) {
        def toInt: Int = number.toInt
    }
}

/** Extractor to retrieve a TicketNumber from a path parameter.
  */
object TicketNumberPathParameter {
    def unapply(str: String): Option[TicketNumber] = Option(str).flatMap(TicketNumber.fromString)
}

/** Possible states of a ticket which shall model the life cycle from ticket creation until it is closed. To keep things
  * simple we do not model the kind of a resolution (e.g. fixed, won't fix or so) for a ticket.
  */
enum TicketStatus {

    /** The ticket is considered to be a valid ticket / ready to be worked on e.g. a bug has been confirmed to be
      * present.
      */
    case Confirmed

    /** The ticket is being worked on i.e. it is in progress.
      */
    case InProgress

    /** The ticket is pending and cannot be processed right now. It may be moved to another state or closed depending on
      * the circumstances. This could be used to model the "blocked" state of Kanban.
      */
    case Pending

    /** The ticket is resolved (i.e. closed) and considered done.
      */
    case Resolved

    /** The ticket was reported and is open for further investigation. This might map to what is often called "Backlog"
      * nowadays.
      */
    case Submitted
}

object TicketStatus {
    given Eq[TicketStatus] = Eq.fromUniversalEquals

    /** Try to parse a ticket status instance from the given string without throwin an exception like `valueOf`.
      *
      * @param source
      *   A string that should contain the name of a ticket status.
      * @return
      *   An option to the successfully deserialised instance.
      */
    def fromString(source: String): Option[TicketStatus] =
        TicketStatus.values.map(_.toString).find(_ === source).map(TicketStatus.valueOf)
}

/** Possible types of "resolved states" of a ticket.
  */
enum TicketResolution {

    /** The behaviour / scenario described in the ticket is caused by the design of the application and not considered
      * to be a bug.
      */
    case ByDesign

    /** The ticket is finally closed and considered done.
      *
      * This state can be used to model a review process e.g. a developer can move a ticket to `Fixed` and reviewer and
      * tester can later move the ticket to `Closed`.
      */
    case Closed

    /** The ticket is a duplicate of an already existing one.
      */
    case Duplicate

    /** The bug described in the ticket was fixed.
      */
    case Fixed

    /** The feature described in the ticket was implemented.
      */
    case Implemented

    /** The ticket is considered to be invalid.
      */
    case Invalid

    /** The issue described in the ticket will not be fixed.
      */
    case WontFix
}

object TicketResolution {
    given Eq[TicketResolution] = Eq.fromUniversalEquals

    /** Try to parse a ticket resolution instance from the given string without throwin an exception like `valueOf`.
      *
      * @param source
      *   A string that should contain the name of a ticket resolution.
      * @return
      *   An option to the successfully deserialised instance.
      */
    def fromString(source: String): Option[TicketResolution] =
        TicketResolution.values.map(_.toString).find(_ === source).map(TicketResolution.valueOf)
}

/** A concise and short description of the ticket which should not exceed 80 characters.
  */
opaque type TicketTitle = String
object TicketTitle {

    val MaxLength: Int = 72

    /** Create an instance of TicketTitle from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a TicketTitle.
      * @return
      *   The appropriate instance of TicketTitle.
      */
    def apply(source: String): TicketTitle = source

    /** Try to create an instance of TicketTitle from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a TicketTitle.
      * @return
      *   An option to the successfully converted TicketTitle.
      */
    def from(source: String): Option[TicketTitle] =
        Option(source).filter(string => string.nonEmpty && string.length <= MaxLength)
}

/** An ticket used to describe a problem or a task (e.g. implement a concrete feature) within the scope of a project.
  *
  * @param number
  *   The unique identifier of a ticket within the project scope is its number.
  * @param title
  *   A concise and short description of the ticket which should not exceed 72 characters.
  * @param content
  *   An optional field to describe the ticket in great detail if needed.
  * @param status
  *   The current status of the ticket describing its life cycle.
  * @param resolution
  *   An optional resolution state of the ticket that should be set if it is closed.
  * @param submitter
  *   The person who submitted (created) this ticket which is optional because of possible account deletion or other
  *   reasons.
  * @param createdAt
  *   The timestamp when the ticket was created / submitted.
  * @param updatedAt
  *   The timestamp when the ticket was last updated. Upon creation the update time equals the creation time.
  */
final case class Ticket(
    number: TicketNumber,
    title: TicketTitle,
    content: Option[TicketContent],
    status: TicketStatus,
    resolution: Option[TicketResolution],
    submitter: Option[Submitter],
    createdAt: OffsetDateTime,
    updatedAt: OffsetDateTime
)

/** A data container for values that can be used to filter a list of tickets by.
  *
  * @param number
  *   A list of ticket numbers that must be matched.
  * @param status
  *   A list of ticket status flags that must be matched.
  * @param resolution
  *   A list of ticket resolution kinds that must be matched.
  * @param submitter
  *   A list of usernames from whom the ticket must have been submitted.
  */
final case class TicketFilter(
    number: List[TicketNumber],
    status: List[TicketStatus],
    resolution: List[TicketResolution],
    submitter: List[SubmitterName]
)

object TicketFilter {
    given QueryParamDecoder[TicketFilter] = QueryParamDecoder[String].map(TicketFilter.fromQueryParameter)
    given QueryParamEncoder[TicketFilter] = QueryParamEncoder[String].contramap(_.toQueryParameter)

    /** Decode an optional possibly existing query parameter into a `TicketFilter`.
      *
      * Usage: `case GET -> Root / "..." :? OptionalUrlParamter(maybeFilter) => ...`
      */
    object OptionalUrlParameter extends OptionalQueryParamDecoderMatcher[TicketFilter]("q")

    // Only "open" tickets.
    val OpenTicketsOnly = TicketFilter(
        number = Nil,
        status = TicketStatus.values.filterNot(_ === TicketStatus.Resolved).toList,
        resolution = Nil,
        submitter = Nil
    )
    // Only resolved (closed) tickets.
    val ResolvedTicketsOnly =
        TicketFilter(number = Nil, status = List(TicketStatus.Resolved), resolution = Nil, submitter = Nil)

    /** Parse the given query string which must contain a serialised ticket filter instance and return a ticket filter
      * with the successfully parsed filters.
      *
      * @param queryString
      *   A query string parameter passed via an URL.
      * @return
      *   A ticket filter instance which may be empty.
      */
    def fromQueryParameter(queryString: String): TicketFilter = {
        val number =
            if (queryString.contains("numbers: "))
                queryString
                    .drop(queryString.indexOf("numbers: ") + 9)
                    .takeWhile(char => !char.isWhitespace)
                    .split(",")
                    .map(TicketNumber.fromString)
                    .flatten
                    .toList
            else
                Nil
        val status =
            if (queryString.contains("status: "))
                queryString
                    .drop(queryString.indexOf("status: ") + 8)
                    .takeWhile(char => !char.isWhitespace)
                    .split(",")
                    .map(TicketStatus.fromString)
                    .flatten
                    .toList
            else
                Nil
        val resolution =
            if (queryString.contains("resolution: "))
                queryString
                    .drop(queryString.indexOf("resolution: ") + 12)
                    .takeWhile(char => !char.isWhitespace)
                    .split(",")
                    .map(TicketResolution.fromString)
                    .flatten
                    .toList
            else
                Nil
        val submitter =
            if (queryString.contains("by: "))
                queryString
                    .drop(queryString.indexOf("by: ") + 4)
                    .takeWhile(char => !char.isWhitespace)
                    .split(",")
                    .map(SubmitterName.from)
                    .flatten
                    .toList
            else
                Nil
        TicketFilter(number, status, resolution, submitter)
    }

    extension (filter: TicketFilter) {

        /** Convert this ticket filter instance into a query string representation that can be passed as query parameter
          * in a URL and parsed back again.
          *
          * @return
          *   A string containing a serialised form of the ticket filter that can be used as a URL query parameter.
          */
        def toQueryParameter: String = {
            val numbers =
                if (filter.number.isEmpty)
                    None
                else
                    filter.number.map(_.toString).mkString(",").some
            val status =
                if (filter.status.isEmpty)
                    None
                else
                    filter.status.map(_.toString).mkString(",").some
            val resolution =
                if (filter.resolution.isEmpty)
                    None
                else
                    filter.resolution.map(_.toString).mkString(",").some
            val submitter =
                if (filter.submitter.isEmpty)
                    None
                else
                    filter.submitter.map(_.toString).mkString(",").some
            List(
                numbers.map(string => s"numbers: $string"),
                status.map(string => s"status: $string"),
                resolution.map(string => s"resolution: $string"),
                submitter.map(string => s"by: $string")
            ).flatten.mkString(" ")
        }
    }
}
