/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import cats.*
import cats.data.*
import cats.syntax.all.*
import de.smederee.tickets.forms.FormValidator
import de.smederee.tickets.forms.types.*

import scala.util.Try

/** Data container to edit a ticket.
  *
  * @param number
  *   The unique identifier of a ticket within the project scope is its number.
  * @param title
  *   A concise and short description of the ticket which should not exceed 72 characters.
  * @param content
  *   An optional field to describe the ticket in great detail if needed.
  * @param status
  *   The current status of the ticket describing its life cycle.
  * @param resolution
  *   An optional resolution state of the ticket that should be set if it is closed.
  * @param submitter
  *   The unique id of the person who submitted (created) this ticket which is optional because of possible account
  *   deletion or other reasons.
  * @param labels
  *   A list of labels which are associated with the ticket.
  * @param milestones
  *   A list of milestones which are associated with the ticket.
  */
final case class TicketForm(
    number: Option[TicketNumber],
    title: TicketTitle,
    content: Option[TicketContent],
    status: TicketStatus,
    resolution: Option[TicketResolution],
    submitter: Option[SubmitterId],
    labels: List[LabelName],
    milestones: List[MilestoneTitle]
)

object TicketForm extends FormValidator[TicketForm] {
    val fieldContent: FormField    = FormField("content")
    val fieldLabels: FormField     = FormField("labels")
    val fieldMilestones: FormField = FormField("milestones")
    val fieldNumber: FormField     = FormField("number")
    val fieldResolution: FormField = FormField("resolution")
    val fieldStatus: FormField     = FormField("status")
    val fieldSubmitter: FormField  = FormField("submitter")
    val fieldTitle: FormField      = FormField("title")

    /** Create a form for editing a ticket from the given ticket data.
      *
      * @param labels
      *   A list of labels which are associated with the ticket.
      * @param milestones
      *   A list of milestones which are associated with the ticket.
      * @param ticket
      *   The ticket which provides the data for the edit form.
      * @return
      *   A ticket form filled with the data from the given ticket.
      */
    def fromTicket(labels: List[LabelName])(milestones: List[MilestoneTitle])(ticket: Ticket): TicketForm =
        TicketForm(
            number = ticket.number.some,
            title = ticket.title,
            content = ticket.content,
            status = ticket.status,
            resolution = ticket.resolution,
            submitter = ticket.submitter.map(_.id),
            labels = labels,
            milestones = milestones
        )

    override def validate(data: Map[String, Chain[String]]): ValidatedNec[FormErrors, TicketForm] = {
        val number = data
            .get(fieldNumber)
            .fold(Option.empty[TicketNumber].validNec)(
                _.headOption
                    .flatMap(TicketNumber.fromString)
                    .fold(FormFieldError("Invalid ticket number!").invalidNec)(number => Option(number).validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldNumber -> es.toList)))
        val title = data
            .get(fieldTitle)
            .fold(FormFieldError("No ticket title given!").invalidNec)(
                _.headOption
                    .map(_.trim)
                    .flatMap(TicketTitle.from)
                    .fold(FormFieldError("Invalid ticket title!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldTitle -> es.toList)))
        val content = data
            .get(fieldContent)
            .fold(none[TicketContent].validNec)(
                _.headOption
                    .filter(_.nonEmpty)
                    .fold(none[TicketContent].validNec)(s =>
                        TicketContent
                            .from(s)
                            .fold(FormFieldError("Invalid ticket content!").invalidNec)(_.some.validNec)
                    )
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldContent -> es.toList)))
        val status = data
            .get(fieldStatus)
            .fold(FormFieldError("No ticket status given!").invalidNec)(
                _.headOption
                    .flatMap(s => Try(TicketStatus.valueOf(s)).toOption)
                    .fold(FormFieldError("Invalid ticket status!").invalidNec)(_.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldStatus -> es.toList)))
        val resolution = data
            .get(fieldResolution)
            .fold(none[TicketResolution].validNec)(
                _.headOption
                    .flatMap(s => Try(TicketResolution.valueOf(s)).toOption)
                    .fold(FormFieldError("Invalid ticket resolution!").invalidNec)(_.some.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldResolution -> es.toList)))
        val submitterId = data
            .get(fieldSubmitter)
            .fold(none[SubmitterId].validNec)(
                _.headOption
                    .flatMap(s => SubmitterId.fromString(s).toOption)
                    .fold(FormFieldError("Invalid submitter id!").invalidNec)(_.some.validNec)
            )
            .leftMap(es => NonEmptyChain.of(Map(fieldSubmitter -> es.toList)))
        val labels =
            data
                .get(fieldLabels)
                .fold(List.empty[LabelName].validNec[FormErrors])(_.toList.flatMap(LabelName.from).validNec[FormErrors])
        val milestones =
            data
                .get(fieldMilestones)
                .fold(List.empty[MilestoneTitle].validNec[FormErrors])(
                    _.toList.flatMap(MilestoneTitle.from).validNec[FormErrors]
                )
        (number, title, content, status, resolution, submitterId, labels, milestones).mapN {
            case (number, title, content, status, resolution, submitterId, labels, milestones) =>
                TicketForm(number, title, content, status, resolution, submitterId, labels, milestones)
        }
    }

    extension (form: TicketForm) {

        /** Convert the form class into a stringified map which is used as underlying data type for form handling in the
          * twirl templating library.
          *
          * @return
          *   A stringified map containing the data of the form.
          */
        def toMap: Map[String, Chain[String]] =
            Map(
                TicketForm.fieldNumber.toString     -> form.number.map(_.toString).fold(Chain.empty)(nr => Chain(nr)),
                TicketForm.fieldTitle.toString      -> Chain(form.title.toString),
                TicketForm.fieldContent.toString    -> form.content.map(_.toString).fold(Chain.empty)(c => Chain(c)),
                TicketForm.fieldStatus.toString     -> Chain(form.status.toString),
                TicketForm.fieldResolution.toString -> form.resolution.map(_.toString).fold(Chain.empty)(r => Chain(r)),
                TicketForm.fieldSubmitter.toString  -> form.submitter.map(_.toString).fold(Chain.empty)(s => Chain(s)),
                TicketForm.fieldLabels.toString     -> Chain.fromSeq(form.labels.map(_.toString)),
                TicketForm.fieldMilestones.toString -> Chain.fromSeq(form.milestones.map(_.toString))
            )
    }
}
