/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets.forms

import cats.data.*
import de.smederee.tickets.forms.types.*

/** A base class for form validators.
  *
  * <p>It is intended to extend this class if you want to provide a more sophisticated validation for a form which gets
  * submitted as raw stringified map.</p>
  *
  * <p>Please note that you can achieve auto validation if you use proper models (with refined types) in your tapir
  * endpoints.</p>
  *
  * <p>However, sometimes you want to have more fine grained control...</p>
  *
  * @tparam T
  *   The concrete type of the validated form output.
  */
abstract class FormValidator[T] {
    final val fieldGlobal: FormField = FormValidator.fieldGlobal

    /** Validate the submitted form data which is received as stringified map and return either a validated `T` or a
      * list of [[de.smederee.tickets.forms.types.FormErrors]].
      *
      * @param data
      *   The stringified map which was submitted.
      * @return
      *   Either the validated form as concrete type T or a list of form errors.
      */
    def validate(data: Map[String, Chain[String]]): ValidatedNec[FormErrors, T]

}

object FormValidator {
    // A constant for the field name used for global errors.
    val fieldGlobal: FormField = FormField("global")
}
