/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets.forms

import cats.data.*
import cats.syntax.all.*

object types {

    type FormErrors = Map[FormField, List[FormFieldError]]
    object FormErrors {
        val empty: FormErrors = Map.empty[FormField, List[FormFieldError]]

        /** Create a single FormErrors instance from a given non empty chain of FormErrors which is usually returned
          * from validators.
          *
          * @param errors
          *   A non empty chain of FormErrors.
          * @return
          *   A single FormErrors instance containing all the errors.
          */
        def fromNec(errors: NonEmptyChain[FormErrors]): FormErrors = errors.toList.fold(FormErrors.empty)(_ combine _)

        /** Create a single FormErrors instance from a given non empty list of FormErrors which is usually returned from
          * validators.
          *
          * @param errors
          *   A non empty list of FormErrors.
          * @return
          *   A single FormErrors instance containing all the errors.
          */
        def fromNel(errors: NonEmptyList[FormErrors]): FormErrors = errors.toList.fold(FormErrors.empty)(_ combine _)
    }

    opaque type FormField = String
    object FormField {

        /** Create an instance of FormField from the given String type.
          *
          * @param source
          *   An instance of type String which will be returned as a FormField.
          * @return
          *   The appropriate instance of FormField.
          */
        def apply(source: String): FormField = source

        /** Try to create an instance of FormField from the given String.
          *
          * @param source
          *   A String that should fulfil the requirements to be converted into a FormField.
          * @return
          *   An option to the successfully converted FormField.
          */
        def from(source: String): Option[FormField] =
            Option(source).map(_.trim.nonEmpty) match {
                case Some(true) => Option(source.trim)
                case _          => None
            }
    }

    given Conversion[FormField, String] = _.toString

    opaque type FormFieldError = String
    object FormFieldError {

        /** Create an instance of FormFieldError from the given String type.
          *
          * @param source
          *   An instance of type String which will be returned as a FormFieldError.
          * @return
          *   The appropriate instance of FormFieldError.
          */
        def apply(source: String): FormFieldError = source

        /** Try to create an instance of FormFieldError from the given String.
          *
          * @param source
          *   A String that should fulfil the requirements to be converted into a FormFieldError.
          * @return
          *   An option to the successfully converted FormFieldError.
          */
        def from(source: String): Option[FormFieldError] =
            Option(source).map(_.trim.nonEmpty) match {
                case Some(true) => Option(source.trim)
                case _          => None
            }
    }

    given Conversion[FormFieldError, String] = _.toString

}
