/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.io.IOException
import java.net.ServerSocket
import java.nio.file.*
import java.nio.file.attribute.BasicFileAttributes
import java.time.OffsetDateTime

import cats.effect.*
import cats.syntax.all.*
import com.comcast.ip4s.*
import com.typesafe.config.ConfigFactory
import de.smederee.email.EmailAddress
import de.smederee.hub.config.*
import de.smederee.i18n.LanguageCode
import de.smederee.security.*
import org.flywaydb.core.Flyway
import pureconfig.*

import munit.*

import scala.annotation.nowarn

/** Base class for our integration test suites.
  *
  * It loads and provides the configuration as a resource suit fixture (loaded once for all tests within a suite) and
  * does initialise the test database for each suite. The latter means a possibly existing database with the name
  * configured **will be deleted**!
  */
abstract class BaseSpec extends CatsEffectSuite {

    protected final val configuration: SmedereeHubConfig =
        ConfigSource
            .fromConfig(ConfigFactory.load(getClass.getClassLoader))
            .at(SmedereeHubConfig.location)
            .loadOrThrow[SmedereeHubConfig]

    protected final val flyway: Flyway =
        DatabaseMigrator
            .configureFlyway(configuration.database.url, configuration.database.user, configuration.database.pass)
            .cleanDisabled(false)
            .load()

    /** Connect to the DBMS using the generic "template1" database which should always be present.
      *
      * @param dbConfig
      *   The database configuration.
      * @return
      *   The connection to the database ("template1").
      */
    private def connect(dbConfig: DatabaseConfig): IO[java.sql.Connection] =
        for {
            _        <- IO(Class.forName(dbConfig.driver))
            database <- IO(dbConfig.url.split("/").reverse.take(1).mkString)
            connection <- IO(
                java.sql.DriverManager
                    .getConnection(dbConfig.url.replace(database, "template1"), dbConfig.user, dbConfig.pass)
            )
        } yield connection

    @nowarn("msg=discarded non-Unit value.*")
    override def beforeAll(): Unit = {
        // Extract the database name from the URL.
        val database = configuration.database.url.split("/").reverse.take(1).mkString
        val db       = Resource.make(connect(configuration.database))(con => IO(con.close()))
        // Create the test database if it does not already exist.
        db.use { connection =>
            for {
                statement <- IO(connection.createStatement())
                exists <- IO(
                    statement.executeQuery(
                        s"""SELECT datname FROM pg_catalog.pg_database WHERE datname = '$database'"""
                    )
                )
                _ <- IO {
                    if (!exists.next())
                        statement.execute(s"""CREATE DATABASE "$database"""")
                }
                _ <- IO(exists.close)
                _ <- IO(statement.close)
            } yield ()
        }.unsafeRunSync()
    }

    override def beforeEach(context: BeforeEach): Unit = {
        val _ = flyway.migrate()
        val _ = flyway.clean()
        val _ = flyway.migrate()
    }

    override def afterEach(context: AfterEach): Unit = {
        val _ = flyway.migrate()
        val _ = flyway.clean()
    }

    /** Find and return a free port on the local machine by starting a server socket and closing it. The port number
      * used by the socket is marked to allow reuse, considered free and returned.
      *
      * @return
      *   An optional port number if a free one can be found.
      */
    protected def findFreePort(): Option[Port] = {
        val socket = new ServerSocket(0)
        val port   = socket.getLocalPort
        socket.setReuseAddress(true) // Allow instant rebinding of the socket.
        socket.close()               // Free the socket for further use by closing it.
        Port.fromInt(port)
    }

    /** Provide a resource with a database connection to allow db operations and proper resource release later.
      *
      * @param cfg
      *   The application configuration.
      * @return
      *   A cats resource encapsulation a database connection as defined within the given configuration.
      */
    protected def connectToDb(cfg: SmedereeHubConfig): Resource[IO, java.sql.Connection] =
        Resource.make(
            IO.delay(java.sql.DriverManager.getConnection(cfg.database.url, cfg.database.user, cfg.database.pass))
        )(c => IO.delay(c.close()))

    /** Create the given account in the database.
      *
      * @param account
      *   The account to be created.
      * @param hash
      *   A password hash to be stored.
      * @param unlockToken
      *   An optional unlock token to be stored.
      * @param attempts
      *   Optional number of failed login attempts.
      * @param validationToken
      *   An optional validation token to be stored.
      * @return
      *   The number of affected database rows.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    @SuppressWarnings(Array("DisableSyntax.defaultArgs"))
    protected def createAccount(
        account: Account,
        hash: PasswordHash,
        unlockToken: Option[UnlockToken] = None,
        attempts: Option[Int] = None,
        validationToken: Option[ValidationToken] = None
    ): IO[Int] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay {
                    (unlockToken, validationToken) match {
                        case (None, None) =>
                            con.prepareStatement(
                                """INSERT INTO hub.accounts (uid, name, email, full_name, password, failed_attempts, created_at, updated_at, validated_email) VALUES(?, ?, ?, ?, ?, ?, NOW(), NOW(), ?)"""
                            )
                        case (Some(_), None) =>
                            con.prepareStatement(
                                """INSERT INTO hub.accounts (uid, name, email, full_name, password, failed_attempts, validated_email, locked_at, unlock_token, created_at, updated_at) VALUES(?, ?, ?, ?, ?, ?, ?, NOW(), ?, NOW(), NOW())"""
                            )
                        case (None, Some(_)) =>
                            con.prepareStatement(
                                """INSERT INTO hub.accounts (uid, name, email, full_name, password, failed_attempts, validated_email, locked_at, validation_token, created_at, updated_at) VALUES(?, ?, ?, ?, ?, ?, ?, NOW(), ?, NOW(), NOW())"""
                            )
                        case (Some(_), Some(_)) =>
                            con.prepareStatement(
                                """INSERT INTO hub.accounts (uid, name, email, full_name, password, failed_attempts, validated_email, locked_at, unlock_token, validation_token, created_at, updated_at) VALUES(?, ?, ?, ?, ?, ?, ?, NOW(), ?, NOW(), NOW())"""
                            )
                    }
                }
                _ <- IO.delay(statement.setObject(1, account.uid))
                _ <- IO.delay(statement.setString(2, account.name.toString))
                _ <- IO.delay(statement.setString(3, account.email.toString))
                _ <- IO.delay(
                    account.fullName.fold(statement.setNull(4, java.sql.Types.VARCHAR))(name =>
                        statement.setString(4, name.toString)
                    )
                )
                _ <- IO.delay(statement.setString(5, hash.toString))
                _ <- IO.delay(statement.setInt(6, attempts.getOrElse(1)))
                _ <- IO.delay(statement.setBoolean(7, account.validatedEmail))
                _ <- (unlockToken, validationToken) match {
                    case (None, None)     => IO.unit
                    case (Some(ut), None) => IO.delay(statement.setString(8, ut.toString))
                    case (None, Some(vt)) => IO.delay(statement.setString(8, vt.toString))
                    case (Some(ut), Some(vt)) =>
                        IO.delay {
                            statement.setString(8, ut.toString)
                            statement.setString(9, vt.toString)
                        }
                }
                _ <- IO.delay {
                    unlockToken.foreach { token =>
                        statement.setString(8, token.toString)
                    }
                }
                r <- IO.delay(statement.executeUpdate())
                _ <- IO.delay(statement.close())
            } yield r
        }

    /** Create the given user session in the database.
      *
      * @param session
      *   The session that shall be created. The corresponding user account must exist!
      * @return
      *   The number of affected database rows.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def createUserSession(session: Session): IO[Int] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """INSERT INTO hub.sessions (id, uid, created_at, updated_at) VALUES(?, ?, ?, ?)"""
                    )
                )
                _ <- IO.delay(statement.setString(1, session.id.toString))
                _ <- IO.delay(statement.setObject(2, session.uid))
                _ <- IO.delay(
                    statement.setTimestamp(3, java.sql.Timestamp.from(session.createdAt.toInstant))
                )
                _ <- IO.delay(
                    statement.setTimestamp(4, java.sql.Timestamp.from(session.updatedAt.toInstant))
                )
                r <- IO.delay(statement.executeUpdate())
                _ <- IO.delay(statement.close())
            } yield r
        }

    /** Load the account with the given uid from the database.
      *
      * @param uid
      *   The unique identifier for the account.
      * @return
      *   An option to the account if it exists.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    @SuppressWarnings(Array("DisableSyntax.null"))
    protected def loadAccount(uid: UserId): IO[Option[Account]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT uid, name, email, full_name, password, created_at, updated_at, validated_email, language FROM hub.accounts WHERE uid = ? LIMIT 1"""
                    )
                )
                _      <- IO.delay(statement.setObject(1, uid))
                result <- IO.delay(statement.executeQuery)
                account <- IO.delay {
                    if (result.next()) {
                        val language =
                            if (result.getString("language") =!= null)
                                LanguageCode.from(result.getString("language"))
                            else
                                None
                        val fullName = FullName.from(result.getString("full_name"))
                        Option(
                            Account(
                                uid = uid,
                                name = Username(result.getString("name")),
                                email = EmailAddress(result.getString("email")),
                                fullName = fullName,
                                validatedEmail = result.getBoolean("validated_email"),
                                language = language
                            )
                        )
                    } else {
                        None
                    }
                }
                _ <- IO(statement.close())
            } yield account
        }

    /** Load the password hash for the account with the given unique user id.
      *
      * @param uid
      *   The unique identifier for the account.
      * @return
      *   An option to the password hash if the user account exists.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    @SuppressWarnings(Array("DisableSyntax.null"))
    protected def loadPasswordHash(uid: UserId): IO[Option[PasswordHash]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement("""SELECT password FROM hub.accounts WHERE uid = ? LIMIT 1""")
                )
                _      <- IO.delay(statement.setObject(1, uid))
                result <- IO.delay(statement.executeQuery)
                password <- IO.delay {
                    if (result.next()) {
                        Option(result.getString("password")).flatMap(PasswordHash.from)
                    } else {
                        None
                    }
                }
                _ <- IO.delay(statement.close())
            } yield password
        }

    /** Load the password reset related columns for the account with the given unique user id.
      *
      * @param uid
      *   The unique identifier for the account.
      * @return
      *   An option of the columns if it exists.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    @SuppressWarnings(Array("DisableSyntax.null"))
    protected def loadResetColumns(uid: UserId): IO[Option[(Option[OffsetDateTime], Option[ResetToken])]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT reset_expiry, reset_token FROM hub.accounts WHERE uid = ? LIMIT 1"""
                    )
                )
                _      <- IO.delay(statement.setObject(1, uid))
                result <- IO.delay(statement.executeQuery)
                columns <- IO.delay {
                    if (result.next()) {
                        val expiry =
                            if (result.getString("reset_expiry") =!= null)
                                Option(result.getObject("reset_expiry", classOf[OffsetDateTime]))
                            else
                                None
                        val token =
                            if (result.getString("reset_token") =!= null)
                                ResetToken.from(result.getString("reset_token"))
                            else
                                None
                        Option(expiry, token)
                    } else {
                        None
                    }
                }
                _ <- IO.delay(statement.close())
            } yield columns
        }

    /** Load the validation related columns for the account with the given unique user id.
      *
      * @param uid
      *   The unique identifier for the account.
      * @return
      *   An option of the columns if it exists.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def loadValidationColumns(uid: UserId): IO[Option[(Boolean, Option[ValidationToken])]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT validated_email, validation_token FROM hub.accounts WHERE uid = ? LIMIT 1"""
                    )
                )
                _      <- IO.delay(statement.setObject(1, uid))
                result <- IO.delay(statement.executeQuery)
                columns <- IO.delay {
                    if (result.next()) {
                        Option(
                            (
                                result.getBoolean("validated_email"),
                                ValidationToken
                                    .from(result.getString("validation_token"))
                            )
                        )
                    } else {
                        None
                    }
                }
                _ <- IO.delay(statement.close())
            } yield columns
        }

    /** Find the repository ID for the given owner and repository name.
      *
      * @param owner
      *   The unique ID of the user account that owns the repository.
      * @param name
      *   The repository name which must be unique in regard to the owner.
      * @return
      *   An option to the internal database ID.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def loadVcsRepositoryId(owner: UserId, name: VcsRepositoryName): IO[Option[VcsRepositoryId]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT id FROM hub.repositories WHERE owner = ? AND name = ? LIMIT 1"""
                    )
                )
                _      <- IO.delay(statement.setObject(1, owner))
                _      <- IO.delay(statement.setString(2, name.toString))
                result <- IO.delay(statement.executeQuery)
                account <- IO.delay {
                    if (result.next()) {
                        VcsRepositoryId.from(result.getLong("id"))
                    } else {
                        None
                    }
                }
                _ <- IO(statement.close())
            } yield account
        }

    /** Delete the given directory recursively.
      *
      * @param path
      *   The path on the filesystem to the directory that shall be deleted.
      * @return
      *   `true` if the directory was deleted.
      */
    protected def deleteDirectory(path: Path): IO[Boolean] =
        IO.delay {
            if (path.toString.trim =!= "/") {
                Files.walkFileTree(
                    path,
                    new FileVisitor[Path] {
                        override def visitFileFailed(file: Path, exc: IOException): FileVisitResult =
                            FileVisitResult.CONTINUE

                        override def visitFile(file: Path, attrs: BasicFileAttributes): FileVisitResult = {
                            Files.delete(file)
                            FileVisitResult.CONTINUE
                        }

                        override def preVisitDirectory(dir: Path, attrs: BasicFileAttributes): FileVisitResult =
                            FileVisitResult.CONTINUE

                        override def postVisitDirectory(dir: Path, exc: IOException): FileVisitResult = {
                            Files.delete(dir)
                            FileVisitResult.CONTINUE
                        }
                    }
                )
                Files.deleteIfExists(path)
            } else false
        }
}
