/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.time.OffsetDateTime
import java.time.ZoneOffset
import java.util.UUID

import cats.effect.*
import cats.syntax.all.*
import de.smederee.TestTags.*
import de.smederee.hub.Generators.*
import de.smederee.security.*
import de.smederee.ssh.*
import doobie.*

final class DoobieAccountManagementRepositoryTest extends BaseSpec {
    val sshKeyWithComment = ResourceSuiteLocalFixture(
        "ssh-key-with-comment",
        Resource.make(IO {
            val input = scala.io.Source
                .fromInputStream(
                    getClass().getClassLoader().getResourceAsStream("de/smederee/ssh/ssh-key-with-comment.pub"),
                    "UTF-8"
                )
                .getLines()
                .mkString
            val keyString = SshPublicKeyString(input)
            PublicSshKey.from(UUID.randomUUID())(UserId.randomUserId)(OffsetDateTime.now(ZoneOffset.UTC))(keyString)
        })(_ => IO.unit)
    )

    val sshKeyWithoutComment = ResourceSuiteLocalFixture(
        "ssh-key-without-comment",
        Resource.make(IO {
            val input = scala.io.Source
                .fromInputStream(
                    getClass().getClassLoader().getResourceAsStream("de/smederee/ssh/ssh-key-without-comment.pub"),
                    "UTF-8"
                )
                .getLines()
                .mkString
            val keyString = SshPublicKeyString(input)
            PublicSshKey.from(UUID.randomUUID())(UserId.randomUserId)(OffsetDateTime.now(ZoneOffset.UTC))(keyString)
        })(_ => IO.unit)
    )

    override def munitFixtures = List(sshKeyWithComment, sshKeyWithoutComment)

    test("addSshKey must save the key to the database".tag(NeedsDatabase)) {
        (genValidAccount.sample, sshKeyWithComment()) match {
            case (Some(account), Some(sshKey)) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo     = new DoobieAccountManagementRepository[IO](tx)
                val attempts = scala.util.Random.nextInt(128)
                val hash     = PasswordHash("Yet another weak password!")
                val test = for {
                    _    <- createAccount(account, hash, None, Option(attempts))
                    w    <- repo.addSshKey(sshKey.copy(ownerId = account.uid))
                    keys <- repo.listSshKeys(account.uid).compile.toList
                } yield (w, keys)
                test.map { result =>
                    val (written, keys) = result
                    assert(written === 1, "No database rows written!")
                    assert(keys.exists(_.id === sshKey.id), "Key must be in the key list of the user!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("addSshKey must fail if a key with the same fingerprint already exists".tag(NeedsDatabase)) {
        (genValidAccount.sample, sshKeyWithoutComment()) match {
            case (Some(account), Some(sshKey)) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo     = new DoobieAccountManagementRepository[IO](tx)
                val attempts = scala.util.Random.nextInt(128)
                val hash     = PasswordHash("Yet another weak password!")
                val test = for {
                    _ <- createAccount(account, hash, None, Option(attempts))
                    _ <- repo.addSshKey(sshKey.copy(ownerId = account.uid))
                    _ <- repo.addSshKey(sshKey.copy(ownerId = account.uid))
                } yield ()
                test.attempt.map { result =>
                    assert(result.isLeft, "Writing a key with a duplicate fingerprint must fail!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("deleteAccount must remove the account from the database".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo     = new DoobieAccountManagementRepository[IO](tx)
                val attempts = scala.util.Random.nextInt(128)
                val hash     = PasswordHash("Yet another weak password!")
                val test = for {
                    _ <- createAccount(account, hash, None, Option(attempts))
                    _ <- repo.deleteAccount(account.uid)
                    o <- loadAccount(account.uid)
                } yield o
                test.map(result => assert(result === None, "Account not deleted from database!"))
        }
    }

    test("findByValidationToken must return the matching account".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(ac) =>
                val account  = ac.copy(validatedEmail = true)
                val token    = ValidationToken.generate
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieAccountManagementRepository[IO](tx)
                val hash = PasswordHash("Yet another weak password!")
                val test = for {
                    _ <- createAccount(account, hash, validationToken = token.some)
                    o <- repo.findByValidationToken(token)
                } yield o
                test.map(result => assert(result === Some(account)))
        }
    }

    test("findPasswordHash must return correct hash".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo     = new DoobieAccountManagementRepository[IO](tx)
                val attempts = scala.util.Random.nextInt(128)
                val hash     = PasswordHash("Yet another weak password!")
                val test = for {
                    _ <- createAccount(account, hash, None, Option(attempts))
                    o <- repo.findPasswordHash(account.uid)
                } yield o
                test.map(result => assert(result.exists(_ === hash), "Unexpected result from database!"))
        }
    }

    test("listSshKeys must return all keys for the user".tag(NeedsDatabase)) {
        (genValidAccount.sample, sshKeyWithComment(), sshKeyWithoutComment()) match {
            case (Some(account), Some(sshKeyA), Some(sshKeyB)) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo     = new DoobieAccountManagementRepository[IO](tx)
                val attempts = scala.util.Random.nextInt(128)
                val hash     = PasswordHash("Yet another weak password!")
                val test = for {
                    _    <- createAccount(account, hash, None, Option(attempts))
                    _    <- repo.addSshKey(sshKeyA.copy(ownerId = account.uid))
                    _    <- repo.addSshKey(sshKeyB.copy(ownerId = account.uid))
                    keys <- repo.listSshKeys(account.uid).compile.toList
                } yield keys
                test.map { keys =>
                    assertEquals(keys.length, 2, "Expected 2 keys in the key list!")
                    assert(keys.exists(_.id === sshKeyA.id), "Key A must be in the key list of the user!")
                    assert(keys.exists(_.id === sshKeyB.id), "Key B must be in the key list of the user!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("markAsValidated must clear the validation token and set the validated column to true".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(ac) =>
                val account  = ac.copy(validatedEmail = true)
                val token    = ValidationToken.generate
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieAccountManagementRepository[IO](tx)
                val hash = PasswordHash("Yet another weak password!")
                val test = for {
                    _    <- createAccount(account, hash, validationToken = token.some)
                    _    <- repo.markAsValidated(account.uid)
                    cols <- loadValidationColumns(account.uid)
                } yield cols
                test.map { result =>
                    assert(result === Some((true, None)), "Unexpected result from database!")
                }
        }
    }

    test("setLanguage must set the language".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val language = genLanguageCode.sample
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieAccountManagementRepository[IO](tx)
                val hash = PasswordHash("Yet another weak password!")
                val test = for {
                    _               <- createAccount(account, hash)
                    _               <- repo.setLanguage(account.uid, language)
                    modifiedAccount <- loadAccount(account.uid)
                } yield modifiedAccount
                test.map { modifiedAccount =>
                    assert(modifiedAccount.exists(_.language === language), "Written language field does not match!")
                }
        }
    }

    test("setValidationToken must set the validation token".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case None => fail("Could not generate data samples!")
            case Some(account) =>
                val token    = ValidationToken.generate
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieAccountManagementRepository[IO](tx)
                val hash = PasswordHash("Yet another weak password!")
                val test = for {
                    _    <- createAccount(account, hash)
                    _    <- repo.setValidationToken(account.uid, token)
                    cols <- loadValidationColumns(account.uid)
                } yield cols
                test.map { result =>
                    assert(result === Some((account.validatedEmail, Some(token))), "Unexpected result from database!")
                }
        }
    }
}
