/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import cats.effect.*
import cats.syntax.all.*
import de.smederee.TestTags.*
import de.smederee.hub.Generators.*
import de.smederee.security.*
import doobie.*

final class DoobieOrganisationRepositoryTest extends BaseSpec {
    test("addAdministrator must add an administrator to the organisation".tag(NeedsDatabase)) {
        (genValidAccounts.sample, genOrganisation.sample) match {
            case (Some(owner :: accounts), Some(org)) =>
                val organisation = org.copy(owner = owner.uid)
                val dbConfig     = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieOrganisationRepository[IO](tx)
                val test = for {
                    _ <- createAccount(owner, PasswordHash("I am not a password hash!"), None, None)
                    _ <- accounts.traverse(a => createAccount(a, PasswordHash("I am not a password hash!"), None, None))
                    written <- repo.create(organisation)
                    added   <- accounts.headOption.traverse(admin => repo.addAdministrator(organisation.oid)(admin.uid))
                } yield (written, added.getOrElse(0))
                test.map { result =>
                    val (written, added) = result
                    assert(written === 1, "Organisation not written to database!")
                    accounts.headOption.foreach(_ => assert(added === 1, "Administrator not written to database!"))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("create must create an organisation".tag(NeedsDatabase)) {
        (genValidAccount.sample, genOrganisation.sample) match {
            case (Some(account), Some(org)) =>
                val organisation = org.copy(owner = account.uid)
                val dbConfig     = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieOrganisationRepository[IO](tx)
                val test = for {
                    _       <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written <- repo.create(organisation)
                } yield written
                test.map { written =>
                    assert(written === 1, "Creating an organisation must modify one database row!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("allByOwner must return all organisations owned by the user".tag(NeedsDatabase)) {
        (genValidAccount.sample, genOrganisations.sample) match {
            case (Some(account), Some(orgs)) =>
                val organisations = orgs.map(_.copy(owner = account.uid))
                val dbConfig      = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieOrganisationRepository[IO](tx)
                val test = for {
                    _       <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written <- organisations.traverse(repo.create)
                    found   <- repo.allByOwner(account.uid).compile.toList
                } yield (written, found)
                test.map { result =>
                    val (written, found) = result
                    assert(written.sum === organisations.size, "Not all organisations written to database!")
                    assertEquals(found.size, organisations.size, "Not all organisations found!")
                    assertEquals(found.sortBy(_.oid), organisations.sortBy(_.oid))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("delete must delete the organisation".tag(NeedsDatabase)) {
        (genValidAccount.sample, genOrganisation.sample) match {
            case (Some(account), Some(org)) =>
                val organisation = org.copy(owner = account.uid)
                val dbConfig     = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieOrganisationRepository[IO](tx)
                val test = for {
                    _       <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written <- repo.create(organisation)
                    deleted <- repo.delete(organisation.oid)
                    found   <- repo.findByName(organisation.name)
                } yield (written, deleted, found)
                test.map { result =>
                    val (written, deleted, found) = result
                    assert(written === 1, "Creating an organisation must modify one database row!")
                    assert(deleted === 1, "Deleting an organisation must modify one database row!")
                    assertEquals(found, None, "Organisation was not deleted from database!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("find must return the organisation with the given id".tag(NeedsDatabase)) {
        (genValidAccount.sample, genOrganisation.sample) match {
            case (Some(account), Some(org)) =>
                val organisation = org.copy(owner = account.uid)
                val dbConfig     = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieOrganisationRepository[IO](tx)
                val test = for {
                    _       <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written <- repo.create(organisation)
                    found   <- repo.find(organisation.oid)
                } yield (written, found)
                test.map { result =>
                    val (written, found) = result
                    assert(written === 1, "Creating an organisation must modify one database row!")
                    assertEquals(found, organisation.some)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findByName must return a matching organisation".tag(NeedsDatabase)) {
        (genValidAccount.sample, genOrganisation.sample) match {
            case (Some(account), Some(org)) =>
                val organisation = org.copy(owner = account.uid)
                val dbConfig     = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieOrganisationRepository[IO](tx)
                val test = for {
                    _       <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written <- repo.create(organisation)
                    found   <- repo.findByName(organisation.name)
                } yield (written, found)
                test.map { result =>
                    val (written, found) = result
                    assert(written === 1, "Creating an organisation must modify one database row!")
                    assertEquals(found, organisation.some)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findOwner must return the user account of the owner".tag(NeedsDatabase)) {
        (genValidAccount.sample, genOrganisation.sample) match {
            case (Some(account), Some(org)) =>
                val organisation = org.copy(owner = account.uid)
                val dbConfig     = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieOrganisationRepository[IO](tx)
                val test = for {
                    _       <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written <- repo.create(organisation)
                    found   <- repo.findOwner(organisation.oid)
                } yield (written, found)
                test.map { result =>
                    val (written, found) = result
                    assert(written === 1, "Creating an organisation must modify one database row!")
                    assertEquals(found, account.copy(language = None).some)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("getAdministrators must return all administrators".tag(NeedsDatabase)) {
        (genValidAccounts.sample, genOrganisation.sample) match {
            case (Some(owner :: accounts), Some(org)) =>
                val admins = accounts.zipWithLongIndex.foldLeft(List.empty[Account])((acc, tuple) =>
                    if (tuple._2 % 2 === 0) {
                        tuple._1 :: acc
                    } else {
                        acc
                    }
                )
                val organisation = org.copy(owner = owner.uid)
                val dbConfig     = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo     = new DoobieOrganisationRepository[IO](tx)
                val addAdmin = repo.addAdministrator(organisation.oid)
                val test = for {
                    _ <- createAccount(owner, PasswordHash("I am not a password hash!"), None, None)
                    _ <- accounts.traverse(a => createAccount(a, PasswordHash("I am not a password hash!"), None, None))
                    written     <- repo.create(organisation)
                    added       <- admins.map(_.uid).traverse(addAdmin)
                    foundAdmins <- repo.getAdministrators(organisation.oid).compile.toList
                } yield (written, added.sum, foundAdmins)
                test.map { result =>
                    val (written, added, foundAdmins) = result
                    assert(written === 1, "Organisation not written to database!")
                    assert(added === admins.size, "Wrong number of administrators created!")
                    assertEquals(foundAdmins.map(_.uid).sorted, admins.map(_.uid).sorted)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("removeAdministrator must remove an administrator from an organisation".tag(NeedsDatabase)) {
        (genValidAccounts.sample, genOrganisation.sample) match {
            case (Some(owner :: accounts), Some(org)) =>
                val organisation = org.copy(owner = owner.uid)
                val dbConfig     = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieOrganisationRepository[IO](tx)
                val test = for {
                    _ <- createAccount(owner, PasswordHash("I am not a password hash!"), None, None)
                    _ <- accounts.traverse(a => createAccount(a, PasswordHash("I am not a password hash!"), None, None))
                    written <- repo.create(organisation)
                    added   <- accounts.headOption.traverse(admin => repo.addAdministrator(organisation.oid)(admin.uid))
                    removed <- accounts.headOption.traverse(admin =>
                        repo.removeAdministrator(organisation.oid)(admin.uid)
                    )
                } yield (written, added.getOrElse(0), removed.getOrElse(0))
                test.map { result =>
                    val (written, added, removed) = result
                    assert(written === 1, "Organisation not written to database!")
                    accounts.headOption.foreach { _ =>
                        assert(added === 1, "Administrator not written to database!")
                        assert(removed === 1, "Administrator not removed from database!")
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("update must update the organisation data".tag(NeedsDatabase)) {
        (genValidAccount.sample, genOrganisation.sample, genOrganisation.sample) match {
            case (Some(account), Some(org), Some(anotherOrg)) =>
                val organisation        = org.copy(owner = account.uid)
                val updatedOrganisation = anotherOrg.copy(oid = organisation.oid, owner = organisation.owner)
                val dbConfig            = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieOrganisationRepository[IO](tx)
                val test = for {
                    _       <- createAccount(account, PasswordHash("I am not a password hash!"), None, None)
                    written <- repo.create(organisation)
                    updated <- repo.update(updatedOrganisation)
                    found   <- repo.find(organisation.oid)
                } yield (written, updated, found)
                test.map { result =>
                    val (written, updated, found) = result
                    assert(written === 1, "Creating an organisation must modify one database row!")
                    assert(updated === 1, "Updating an organisation must modify one database row!")
                    assertEquals(found, updatedOrganisation.some)
                }
            case _ => fail("Could not generate data samples!")
        }
    }
}
