/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.nio.charset.StandardCharsets
import java.time.OffsetDateTime
import java.time.ZoneOffset

import cats.effect.*
import cats.syntax.all.*
import de.smederee.TestTags.*
import de.smederee.hub.Generators.*
import de.smederee.security.*
import doobie.*
import org.flywaydb.core.Flyway

final class DoobieResetPasswordRepositoryTest extends BaseSpec {
    override def beforeEach(context: BeforeEach): Unit = {
        val dbConfig = configuration.database
        val flyway: Flyway =
            DatabaseMigrator.configureFlyway(dbConfig.url, dbConfig.user, dbConfig.pass).cleanDisabled(false).load()
        val _ = flyway.migrate()
        val _ = flyway.clean()
        val _ = flyway.migrate()
    }

    override def afterEach(context: AfterEach): Unit = {
        val dbConfig = configuration.database
        val flyway: Flyway =
            DatabaseMigrator.configureFlyway(dbConfig.url, dbConfig.user, dbConfig.pass).cleanDisabled(false).load()
        val _ = flyway.migrate()
        val _ = flyway.clean()
    }

    test("findByNameAndResetPasswordToken must find a matching account".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case Some(user) =>
                val token           = ResetToken.generate
                val tokenExpiration = OffsetDateTime.now(ZoneOffset.UTC).plusDays(3L)
                val expected        = user.copy(language = None)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieResetPasswordRepository[IO](tx)
                val test = for {
                    _      <- createAccount(user, PasswordHash("I am not a password hash!"), None, None)
                    _      <- repo.setResetPasswordToken(user.uid)(token, tokenExpiration)
                    _      <- repo.removeResetPasswordExpirationDate(user.uid)
                    result <- repo.findByNameAndResetPasswordToken(user.name, token)
                } yield result
                test.map { result =>
                    assertEquals(result, Option(expected))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findByNameAndResetPasswordToken must not respect tokens with expiration date".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case Some(user) =>
                val token           = ResetToken.generate
                val tokenExpiration = OffsetDateTime.now(ZoneOffset.UTC).plusDays(3L)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieResetPasswordRepository[IO](tx)
                val test = for {
                    _      <- createAccount(user, PasswordHash("I am not a password hash!"), None, None)
                    _      <- repo.setResetPasswordToken(user.uid)(token, tokenExpiration)
                    result <- repo.findByNameAndResetPasswordToken(user.name, token)
                } yield result
                test.map { result =>
                    assertEquals(result, None)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findByResetPasswordToken must find a matching account".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case Some(user) =>
                val token           = ResetToken.generate
                val tokenExpiration = OffsetDateTime.now(ZoneOffset.UTC).plusDays(3L)
                val expected        = user.copy(language = None)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieResetPasswordRepository[IO](tx)
                val test = for {
                    _      <- createAccount(user, PasswordHash("I am not a password hash!"), None, None)
                    _      <- repo.setResetPasswordToken(user.uid)(token, tokenExpiration)
                    result <- repo.findByResetPasswordToken(token)
                } yield result
                test.map { result =>
                    assertEquals(result, Option(expected))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findByResetPasswordToken must not return accounts with expired tokens".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case Some(user) =>
                val token           = ResetToken.generate
                val tokenExpiration = OffsetDateTime.now(ZoneOffset.UTC).minusDays(3L)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieResetPasswordRepository[IO](tx)
                val test = for {
                    _      <- createAccount(user, PasswordHash("I am not a password hash!"), None, None)
                    _      <- repo.setResetPasswordToken(user.uid)(token, tokenExpiration)
                    result <- repo.findByResetPasswordToken(token)
                } yield result
                test.map { result =>
                    assertEquals(result, None)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findByResetPasswordToken must not return accounts without expiration date".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case Some(user) =>
                val token           = ResetToken.generate
                val tokenExpiration = OffsetDateTime.now(ZoneOffset.UTC).minusDays(3L)
                val dbConfig        = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieResetPasswordRepository[IO](tx)
                val test = for {
                    _      <- createAccount(user, PasswordHash("I am not a password hash!"), None, None)
                    _      <- repo.setResetPasswordToken(user.uid)(token, tokenExpiration)
                    _      <- repo.removeResetPasswordExpirationDate(user.uid)
                    result <- repo.findByResetPasswordToken(token)
                } yield result
                test.map { result =>
                    assertEquals(result, None)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("removeResetPasswordExpirationDate must remove the expiration date".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case Some(user) =>
                val token = ResetToken.generate
                val tokenExpiration =
                    OffsetDateTime.now(ZoneOffset.UTC).plusDays(3L).withNano(0) // Nanos are troublesome!
                val expected = ((None, token.some)).some
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieResetPasswordRepository[IO](tx)
                val test = for {
                    _      <- createAccount(user, PasswordHash("I am not a password hash!"), None, None)
                    _      <- repo.setResetPasswordToken(user.uid)(token, tokenExpiration)
                    _      <- repo.removeResetPasswordExpirationDate(user.uid)
                    result <- loadResetColumns(user.uid)
                } yield result
                test.map { result =>
                    assertEquals(result, expected)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("removeResetPasswordToken must remove the token and the expiration date".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case Some(user) =>
                val token = ResetToken.generate
                val tokenExpiration =
                    OffsetDateTime.now(ZoneOffset.UTC).plusDays(3L).withNano(0) // Nanos are troublesome!
                val expected = ((None, None)).some
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieResetPasswordRepository[IO](tx)
                val test = for {
                    _      <- createAccount(user, PasswordHash("I am not a password hash!"), None, None)
                    _      <- repo.setResetPasswordToken(user.uid)(token, tokenExpiration)
                    _      <- repo.removeResetPasswordToken(user.uid)
                    result <- loadResetColumns(user.uid)
                } yield result
                test.map { result =>
                    assertEquals(result, expected)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("setPassword must set the password correctly".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case Some(user) =>
                val expected =
                    Password("This is not the password you're looking for!".getBytes(StandardCharsets.UTF_8)).encode
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieResetPasswordRepository[IO](tx)
                val test = for {
                    _      <- createAccount(user, PasswordHash("I am not a password hash!"), None, None)
                    _      <- repo.setPassword(user.uid)(expected)
                    result <- loadPasswordHash(user.uid)
                } yield result
                test.map { result =>
                    assertEquals(result, expected.some)
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("setResetPasswordToken must set token and expiry date correctly".tag(NeedsDatabase)) {
        genValidAccount.sample match {
            case Some(user) =>
                val token = ResetToken.generate
                val tokenExpiration =
                    OffsetDateTime.now(ZoneOffset.UTC).plusDays(3L).withNano(0) // Nanos are troublesome!
                val expected = ((tokenExpiration.some, token.some)).some
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val repo = new DoobieResetPasswordRepository[IO](tx)
                val test = for {
                    _      <- createAccount(user, PasswordHash("I am not a password hash!"), None, None)
                    _      <- repo.setResetPasswordToken(user.uid)(token, tokenExpiration)
                    result <- loadResetColumns(user.uid)
                } yield result
                test.map { result =>
                    assertEquals(result, expected)
                }
            case _ => fail("Could not generate data samples!")
        }
    }
}
