/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import java.nio.charset.StandardCharsets
import java.time.*
import java.util.Locale
import java.util.UUID

import cats.syntax.all.*
import de.smederee.email.EmailAddress
import de.smederee.i18n.LanguageCode
import de.smederee.security.*

import org.scalacheck.*

import scala.concurrent.duration.*
import scala.jdk.CollectionConverters.*

object Generators {
    val MinimumYear: Int = -4713  // Lowest year supported by PostgreSQL
    val MaximumYear: Int = 294276 // Highest year supported by PostgreSQL

    val genLocale: Gen[Locale] = Gen.oneOf(Locale.getAvailableLocales.toList)
    given Arbitrary[Locale]    = Arbitrary(genLocale)

    val genLanguageCode: Gen[LanguageCode] = genLocale.map(_.getISO3Language).filter(_.nonEmpty).map(LanguageCode.apply)

    val genFiniteDuration: Gen[FiniteDuration] =
        Gen.choose(0, Int.MaxValue).map(seconds => FiniteDuration(seconds, SECONDS))

    given Arbitrary[FiniteDuration]         = Arbitrary(genFiniteDuration)
    given Arbitrary[Option[FiniteDuration]] = Arbitrary(Gen.option(genFiniteDuration))

    val genOffsetDateTime: Gen[OffsetDateTime] =
        for {
            year       <- Gen.choose(MinimumYear, MaximumYear)
            month      <- Gen.choose(1, 12)
            day        <- Gen.choose(1, Month.of(month).length(Year.of(year).isLeap))
            hour       <- Gen.choose(0, 23)
            minute     <- Gen.choose(0, 59)
            second     <- Gen.choose(0, 59)
            nanosecond <- Gen.const(0) // Avoid issues with loosing information during saving and loading.
            offset <- Gen.oneOf(
                ZoneId.getAvailableZoneIds.asScala.toList.map(ZoneId.of).map(ZonedDateTime.now).map(_.getOffset)
            )
        } yield OffsetDateTime.of(year, month, day, hour, minute, second, nanosecond, offset)

    given Arbitrary[OffsetDateTime] = Arbitrary(genOffsetDateTime)

    val genZonedDateTime: Gen[ZonedDateTime] =
        for {
            year       <- Gen.choose(MinimumYear, MaximumYear)
            month      <- Gen.choose(1, 12)
            day        <- Gen.choose(1, Month.of(month).length(Year.of(year).isLeap))
            hour       <- Gen.choose(0, 23)
            minute     <- Gen.choose(0, 59)
            second     <- Gen.choose(0, 59)
            nanosecond <- Gen.const(0) // Avoid issues with loosing information during saving and loading.
            zone       <- Gen.oneOf(ZoneId.getAvailableZoneIds.asScala.toList.map(ZoneId.of))
        } yield ZonedDateTime.of(year, month, day, hour, minute, second, nanosecond, zone)

    val genSessionId: Gen[SessionId] = Gen.delay(SessionId.generate)

    val genSignAndValidate: Gen[SignAndValidate] = Gen
        .nonEmptyListOf(Gen.alphaNumChar)
        .map(cs => PrivateKey(cs.mkString.getBytes(StandardCharsets.UTF_8)))
        .map(key => new SignAndValidate(key))

    given Arbitrary[SignAndValidate] = Arbitrary(genSignAndValidate)

    val genUserId: Gen[UserId] = Gen.delay(UserId.randomUserId)

    val genUUID: Gen[UUID] = Gen.delay(UUID.randomUUID)

    private val validEmailAddressPrefixChars =
        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789.!#$%&’'*+/=?^_`{|}~-".toList
    private val validDomainNameChars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-".toList

    val genValidEmail: Gen[EmailAddress] = for {
        prefix <- Gen.nonEmptyListOf(Gen.oneOf(validEmailAddressPrefixChars)).map(_.take(64).mkString)
        domain <- Gen.nonEmptyListOf(Gen.oneOf(validDomainNameChars)).map(_.take(32).mkString)
        topLevelDomain <- Gen
            .nonEmptyListOf(Gen.oneOf(validDomainNameChars))
            .suchThat(_.length >= 2)
            .map(_.take(24).mkString)
        suffix = s"$domain.$topLevelDomain"
    } yield EmailAddress(s"$prefix@$suffix")

    val genValidFullName: Gen[FullName] =
        Gen.nonEmptyListOf(Gen.alphaNumChar).map(_.take(FullName.MaximumLength).mkString).map(FullName.apply)

    val genValidUsername: Gen[Username] = for {
        length <- Gen.choose(2, 30)
        prefix <- Gen.alphaChar
        chars <- Gen
            .nonEmptyListOf(Gen.alphaNumChar)
            .map(_.take(length).mkString.toLowerCase(Locale.ROOT))
    } yield Username(prefix.toString.toLowerCase(Locale.ROOT) + chars)

    val genValidAccount: Gen[Account] = for {
        id             <- genUserId
        email          <- genValidEmail
        name           <- genValidUsername
        fullName       <- Gen.option(genValidFullName)
        validatedEmail <- Gen.oneOf(List(false, true))
        language       <- Gen.option(genLanguageCode)
    } yield Account(
        uid = id,
        name = name,
        email = email,
        fullName = fullName,
        validatedEmail = validatedEmail,
        language = language
    )

    given Arbitrary[Account] = Arbitrary(genValidAccount)

    val genValidAccounts: Gen[List[Account]] = Gen
        .nonEmptyListOf(genValidAccount)
        .map(_.foldLeft(List.empty[Account]) { (acc, a) =>
            if (acc.exists(_.name === a.name))
                acc
            else
                a :: acc
        }) // Ensure distinct user names.

    val genOrganisationId: Gen[OrganisationId] = Gen.delay(OrganisationId.randomOrganisationId)

    val genOrganisation: Gen[Organisation] =
        for {
            oid         <- genOrganisationId
            name        <- genValidUsername
            owner       <- genUserId
            fullName    <- Gen.option(genValidFullName)
            description <- Gen.alphaNumStr.map(OrganisationDescription.from)
            isPrivate   <- Gen.oneOf(List(false, true))
        } yield Organisation(
            oid = oid,
            name = name,
            owner = owner,
            fullName = fullName,
            description = description,
            isPrivate = isPrivate,
            website = None
        )

    val genOrganisations: Gen[List[Organisation]] = Gen.nonEmptyListOf(genOrganisation)

    val genValidSession: Gen[Session] =
        for {
            id  <- genSessionId
            uid <- genUserId
            cat <- genOffsetDateTime.map(_.withOffsetSameInstant(ZoneOffset.UTC))
            uat <- genOffsetDateTime.map(_.withOffsetSameInstant(ZoneOffset.UTC))
        } yield Session(id, uid, cat, uat)

    given Arbitrary[Session] = Arbitrary(genValidSession)

    val genValidSessions: Gen[List[Session]] = Gen.nonEmptyListOf(genValidSession)

    val genValidVcsRepositoryName: Gen[VcsRepositoryName] = Gen
        .nonEmptyListOf(
            Gen.oneOf(
                List(
                    "a",
                    "b",
                    "c",
                    "d",
                    "e",
                    "f",
                    "g",
                    "h",
                    "i",
                    "j",
                    "k",
                    "l",
                    "m",
                    "n",
                    "o",
                    "p",
                    "q",
                    "r",
                    "s",
                    "t",
                    "u",
                    "v",
                    "w",
                    "x",
                    "y",
                    "z",
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "-",
                    "_"
                )
            )
        )
        .map(cs => VcsRepositoryName(cs.take(64).mkString))

    val genValidVcsRepositoryOwner = for {
        uid   <- genUserId
        name  <- genValidUsername
        email <- genValidEmail
    } yield VcsRepositoryOwner(uid, name, email)

    val genValidVcsType = Gen.oneOf(VcsType.values.toList)

    val genValidVcsRepository: Gen[VcsRepository] = for {
        name           <- genValidVcsRepositoryName
        owner          <- genValidVcsRepositoryOwner
        isPrivate      <- Gen.oneOf(List(false, true))
        description    <- Gen.alphaNumStr.map(VcsRepositoryDescription.from)
        ticketsEnabled <- Gen.oneOf(List(false, true))
        vcsType        <- genValidVcsType
    } yield VcsRepository(name, owner, isPrivate, description, ticketsEnabled, vcsType, None)

    val genValidVcsRepositories: Gen[List[VcsRepository]] = Gen.nonEmptyListOf(genValidVcsRepository)

}
