/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub

import munit.*

import org.scalacheck.*
import org.scalacheck.Prop.*

final class VcsRepositoryPatchMetadataTest extends ScalaCheckSuite {
    private val genVcsPatchFilename: Gen[VcsPatchFilename] =
        Gen.nonEmptyListOf(Gen.alphaNumChar).map(chars => VcsPatchFilename(chars.mkString))
    private val genVcsPatchFilenames: Gen[List[VcsPatchFilename]] = Gen.listOf(genVcsPatchFilename)
    private val genVcsPatchSummaryFileModification: Gen[VcsPatchSummaryFileModification] = for {
        name    <- genVcsPatchFilename
        added   <- Gen.choose(0, Int.MaxValue)
        removed <- Gen.choose(0, Int.MaxValue)
    } yield VcsPatchSummaryFileModification(added, name, removed)
    private val genVcsPatchSummaryFileModifications: Gen[List[VcsPatchSummaryFileModification]] =
        Gen.listOf(genVcsPatchSummaryFileModification)

    /** Create a darcs xml patch element from the given input for testing purposes.
      *
      * @return
      *   An xml node which holds all information required to mimick a darcs log xml-output patch entry.
      */
    private def createPatchXml(
        added: List[VcsPatchFilename],
        modified: List[VcsPatchSummaryFileModification],
        removed: List[VcsPatchFilename]
    ): scala.xml.Node = {
        // We need to preserve the xml parts as they are because line breaks etc. would be added to file names!
        // format: off
        val added_files   = added.map(filename => <add_file>{filename.toString}</add_file>)
        val removed_files = removed.map(filename => <remove_file>{filename.toString}</remove_file>)
        val modified_files = modified.map { summary =>
          <modify_file>{summary.name.toString}<removed_lines num={summary.removed.toString}/><added_lines num={summary.added.toString}/></modify_file>
        }
        val patch =
          <patch>
            <name>TEST</name>
            <comment>A test comment...</comment>
            <summary>
              {added_files}
              {modified_files}
              {removed_files}
            </summary>
          </patch>
        // format: on
        patch
    }

    property("VcsPatchSummary#fromDarcsXmlLog must parse a given XML node correctly") {
        forAll(genVcsPatchFilenames, genVcsPatchSummaryFileModifications, genVcsPatchFilenames) {
            (
                added: List[VcsPatchFilename],
                modified: List[VcsPatchSummaryFileModification],
                removed: List[VcsPatchFilename]
            ) =>
                val xmlInput = createPatchXml(added, modified, removed)
                val expected = VcsPatchSummary(added, modified, removed)
                VcsPatchSummary.fromDarcsXmlLog(xmlInput) match {
                    case None           => fail("VcsPatchSummary could not be generated!")
                    case Some(obtained) => assertEquals(obtained, expected)
                }
        }
    }

}
