/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.hub.config

import cats.syntax.all.*
import com.typesafe.config.*
import org.http4s.Uri
import pureconfig.*

import munit.*

final class ServiceConfigTest extends FunSuite {
    val rawDefaultConfig = new Fixture[Config]("defaultConfig") {
        def apply() = ConfigFactory.load(getClass.getClassLoader)
    }

    override def munitFixtures = List(rawDefaultConfig)

    test("must load from the default configuration successfully") {
        ConfigSource
            .fromConfig(rawDefaultConfig())
            .at(s"${SmedereeHubConfig.location.toString}.service")
            .load[ServiceConfig] match {
                case Left(errors) => fail(errors.toList.mkString(", "))
                case Right(_)     => assert(true)
            }
    }

    test("default configuration must have authentication enabled") {
        ConfigSource
            .fromConfig(rawDefaultConfig())
            .at(s"${SmedereeHubConfig.location.toString}.service")
            .load[ServiceConfig] match {
                case Left(errors) => fail(errors.toList.mkString(", "))
                case Right(cfg)   => assert(cfg.authentication.enabled)
            }
    }

    test("default configuration must have billing disabled") {
        ConfigSource
            .fromConfig(rawDefaultConfig())
            .at(s"${SmedereeHubConfig.location.toString}.service")
            .load[ServiceConfig] match {
                case Left(errors) => fail(errors.toList.mkString(", "))
                case Right(cfg)   => assert(cfg.billing.enabled === false)
            }
    }

    test("default configuration must have sign up enabled") {
        ConfigSource
            .fromConfig(rawDefaultConfig())
            .at(s"${SmedereeHubConfig.location.toString}.service")
            .load[ServiceConfig] match {
                case Left(errors) => fail(errors.toList.mkString(", "))
                case Right(cfg)   => assert(cfg.signup.enabled)
            }
    }

    test("default values for external linking must be setup for local development") {
        ConfigSource
            .fromConfig(rawDefaultConfig())
            .at(s"${SmedereeHubConfig.location.toString}.service")
            .load[ServiceConfig] match {
                case Left(errors) => fail(errors.toList.mkString(", "))
                case Right(cfg) =>
                    val externalCfg = cfg.external
                    assert(externalCfg.host === cfg.host)
                    assert(externalCfg.port === Option(cfg.port))
                    assert(externalCfg.path.isEmpty)
                    assert(externalCfg.scheme === Uri.Scheme.http)
            }
    }
}
