/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.ssh

import java.nio.file.*

import cats.effect.*
import com.comcast.ip4s.*
import de.smederee.hub.BaseSpec
import de.smederee.hub.config.*

final class SshServerProviderTest extends BaseSpec {

    val repositoriesDirectory = ResourceSuiteLocalFixture(
        "repositories-directory",
        Resource.make(IO(Files.createTempDirectory("test-repo-dir")))(path => deleteDirectory(path) *> IO.unit)
    )

    val serverKeyFile = ResourceSuiteLocalFixture(
        "server-key-file",
        Resource.make(
            for {
                path <- IO(Files.createTempFile("test-server-", ".key"))
                _    <- IO(Files.deleteIfExists(path))
            } yield path
        )(_ => IO.unit)
    )

    val freePort = ResourceFunFixture(Resource.make(IO(findFreePort()))(_ => IO.unit))

    override def munitFixtures = List(repositoriesDirectory, serverKeyFile)

    freePort.test("run() must create and start a server with the given configuration") { port =>
        port match {
            case None => fail("Could not find a free port for testing!")
            case Some(portNumber) =>
                val keyfile     = serverKeyFile()
                val darcsConfig = DarcsConfiguration(Paths.get("darcs"), DirectoryPath(repositoriesDirectory()))
                val sshConfig = SshServerConfiguration(
                    enabled = true,
                    genericUser = SshUsername("darcs"),
                    host = host"localhost",
                    port = portNumber,
                    serverKeyFile = keyfile
                )
                val provider = new SshServerProvider(darcsConfig, configuration.database, sshConfig)
                provider.run().use { server =>
                    assert(server.isStarted(), "Server not started!")
                    assertEquals(server.getPort(), portNumber.toString.toInt)
                    assertEquals(server.getHost(), null) // scalafix:ok
                    IO.unit
                }
        }
    }
}
