/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.ssh

import java.util.Locale

import cats.syntax.all.*

import munit.*

import org.scalacheck.*
import org.scalacheck.Prop.*

final class SshUsernameTest extends ScalaCheckSuite {
    val genValidSshUsername: Gen[SshUsername] =
        for {
            firstLetter    <- Gen.alphaChar
            randomAlphaNum <- Gen.nonEmptyListOf(Gen.alphaNumChar).suchThat(_.length > 2)
            name = s"$firstLetter${randomAlphaNum.take(15).mkString}".toLowerCase(Locale.ROOT)
        } yield SshUsername(name)
    given Arbitrary[SshUsername] = Arbitrary(genValidSshUsername)

    property("SshUsername cannot be longer than 16 characters") {
        forAll { (randomSshUsername: SshUsername) =>
            val invalid = s"randomSshUsername${List.fill(16)("a").mkString}"
            assert(SshUsername.from(invalid).isEmpty)
        }
    }

    property("SshUsername must start with a letter") {
        forAll { (randomSshUsername: SshUsername) =>
            val invalid = s"1${randomSshUsername.toString.take(15)}"
            assert(SshUsername.from(invalid).isEmpty)
        }
    }

    property("SshUsername must be at least 3 characters long") {
        forAll { (randomSshUsername: SshUsername) =>
            val invalid = randomSshUsername.toString.take(2)
            assert(SshUsername.from(invalid).isEmpty)
        }
    }

    property("SshUsername.from must fail on invalid input") {
        forAll { (string: String) =>
            assert(SshUsername.from(string).nonEmpty === SshUsername.Format.matches(string))
        }
    }

    property("SshUsername.fromString must succeed on valid input") {
        forAll { (randomSshUsername: SshUsername) =>
            SshUsername.from(randomSshUsername.toString) match {
                case None           => fail(s"Failed to create from valid username: $randomSshUsername")
                case Some(username) => assertEquals(username, randomSshUsername)
            }
        }
    }
}
