/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import cats.effect.*
import cats.syntax.all.*
import de.smederee.TestTags.*
import de.smederee.tickets.Generators.*
import doobie.*

final class DoobieLabelRepositoryTest extends BaseSpec {

    /** Find the label ID for the given project and label name.
      *
      * @param owner
      *   The unique ID of the user account that owns the project.
      * @param vcsRepoName
      *   The project name which must be unique in regard to the owner.
      * @param labelName
      *   The label name which must be unique in the project context.
      * @return
      *   An option to the internal database ID.
      */
    @throws[java.sql.SQLException]("Errors from the underlying SQL procedures will throw exceptions!")
    protected def findLabelId(owner: ProjectOwnerId, vcsRepoName: ProjectName, labelName: LabelName): IO[Option[Long]] =
        connectToDb(configuration).use { con =>
            for {
                statement <- IO.delay(
                    con.prepareStatement(
                        """SELECT "labels".id
              |FROM "tickets"."labels" AS "labels"
              |JOIN "tickets"."projects" AS "projects"
              |ON "labels".project = "projects".id
              |WHERE "projects".owner = ?
              |AND "projects".name = ?
              |AND "labels".name = ?""".stripMargin
                    )
                )
                _      <- IO.delay(statement.setObject(1, owner))
                _      <- IO.delay(statement.setString(2, vcsRepoName.toString))
                _      <- IO.delay(statement.setString(3, labelName.toString))
                result <- IO.delay(statement.executeQuery)
                account <- IO.delay {
                    if (result.next()) {
                        Option(result.getLong("id"))
                    } else {
                        None
                    }
                }
                _ <- IO(statement.close())
            } yield account
        }

    test("allLabels must return all labels".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genLabels.sample) match {
            case (Some(owner), Some(generatedProject), Some(labels)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo = new DoobieLabelRepository[IO](tx)
                val test = for {
                    _         <- createProjectOwner(owner)
                    _         <- createTicketsProject(project)
                    projectId <- loadProjectId(owner.uid, project.name)
                    createdLabels <- projectId match {
                        case None            => IO.pure(List.empty)
                        case Some(projectId) => labels.traverse(label => labelRepo.createLabel(projectId)(label))
                    }
                    foundLabels <- projectId match {
                        case None            => IO.pure(List.empty)
                        case Some(projectId) => labelRepo.allLabels(projectId).compile.toList
                    }
                } yield foundLabels
                test.map { foundLabels =>
                    assert(foundLabels.size === labels.size, "Different number of labels!")
                    foundLabels.sortBy(_.name).zip(labels.sortBy(_.name)).map { (found, expected) =>
                        assertEquals(found.copy(id = expected.id), expected)
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("createLabel must create the label".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genLabel.sample) match {
            case (Some(owner), Some(generatedProject), Some(label)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo = new DoobieLabelRepository[IO](tx)
                val test = for {
                    _               <- createProjectOwner(owner)
                    createdProjects <- createTicketsProject(project)
                    projectId       <- loadProjectId(owner.uid, project.name)
                    createdLabels   <- projectId.traverse(id => labelRepo.createLabel(id)(label))
                    foundLabel      <- projectId.traverse(id => labelRepo.findLabel(id)(label.name))
                } yield (createdProjects, projectId, createdLabels, foundLabel)
                test.map { tuple =>
                    val (createdProjects, projectId, createdLabels, foundLabel) = tuple
                    assert(createdProjects === 1, "Test project was not created!")
                    assert(projectId.nonEmpty, "No project id found!")
                    assert(createdLabels.exists(_ === 1), "Test label was not created!")
                    foundLabel.getOrElse(None) match {
                        case None => fail("Created label not found!")
                        case Some(foundLabel) =>
                            assert(foundLabel.id.nonEmpty, "Label ID must not be empty!")
                            assertEquals(foundLabel.name, label.name)
                            assertEquals(foundLabel.description, label.description)
                            assertEquals(foundLabel.colour, label.colour)
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("createLabel must fail if the label name already exists".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genLabel.sample) match {
            case (Some(owner), Some(generatedProject), Some(label)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo = new DoobieLabelRepository[IO](tx)
                val test = for {
                    _               <- createProjectOwner(owner)
                    createdProjects <- createTicketsProject(project)
                    projectId       <- loadProjectId(owner.uid, project.name)
                    createdLabels   <- projectId.traverse(id => labelRepo.createLabel(id)(label))
                    _               <- projectId.traverse(id => labelRepo.createLabel(id)(label))
                } yield (createdProjects, projectId, createdLabels)
                test.attempt.map(r => assert(r.isLeft, "Creating a label with a duplicate name must fail!"))
            case _ => fail("Could not generate data samples!")
        }
    }

    test("deleteLabel must delete an existing label".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genLabel.sample) match {
            case (Some(owner), Some(generatedProject), Some(label)) =>
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo = new DoobieLabelRepository[IO](tx)
                val test = for {
                    _               <- createProjectOwner(owner)
                    createdProjects <- createTicketsProject(project)
                    projectId       <- loadProjectId(owner.uid, project.name)
                    createdLabels   <- projectId.traverse(id => labelRepo.createLabel(id)(label))
                    labelId         <- findLabelId(owner.uid, project.name, label.name)
                    deletedLabels   <- labelRepo.deleteLabel(label.copy(id = labelId.flatMap(LabelId.from)))
                    foundLabel      <- projectId.traverse(id => labelRepo.findLabel(id)(label.name))
                } yield (createdProjects, projectId, createdLabels, deletedLabels, foundLabel)
                test.map { tuple =>
                    val (createdProjects, projectId, createdLabels, deletedLabels, foundLabel) = tuple
                    assert(createdProjects === 1, "Test vcs project was not created!")
                    assert(projectId.nonEmpty, "No vcs project id found!")
                    assert(createdLabels.exists(_ === 1), "Test label was not created!")
                    assert(deletedLabels === 1, "Test label was not deleted!")
                    assert(foundLabel.getOrElse(None).isEmpty, "Test label was not deleted!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("findLabel must find existing labels".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genLabels.sample) match {
            case (Some(owner), Some(generatedProject), Some(labels)) =>
                val project       = generatedProject.copy(owner = owner)
                val dbConfig      = configuration.database
                val expectedLabel = labels(scala.util.Random.nextInt(labels.size))
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo = new DoobieLabelRepository[IO](tx)
                val test = for {
                    _               <- createProjectOwner(owner)
                    createdProjects <- createTicketsProject(project)
                    projectId       <- loadProjectId(owner.uid, project.name)
                    createdLabels <- projectId match {
                        case None            => IO.pure(List.empty)
                        case Some(projectId) => labels.traverse(label => labelRepo.createLabel(projectId)(label))
                    }
                    foundLabel <- projectId.traverse(id => labelRepo.findLabel(id)(expectedLabel.name))
                } yield foundLabel.flatten
                test.map { foundLabel =>
                    assertEquals(foundLabel.map(_.copy(id = expectedLabel.id)), Option(expectedLabel))
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("updateLabel must update an existing label".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genLabel.sample) match {
            case (Some(owner), Some(generatedProject), Some(label)) =>
                val updatedLabel = label.copy(
                    name = LabelName("updated label"),
                    description = Option(LabelDescription("I am an updated label description...")),
                    colour = ColourCode("#abcdef")
                )
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo = new DoobieLabelRepository[IO](tx)
                val test = for {
                    _               <- createProjectOwner(owner)
                    createdProjects <- createTicketsProject(project)
                    projectId       <- loadProjectId(owner.uid, project.name)
                    createdLabels   <- projectId.traverse(id => labelRepo.createLabel(id)(label))
                    labelId         <- findLabelId(owner.uid, project.name, label.name)
                    updatedLabels   <- labelRepo.updateLabel(updatedLabel.copy(id = labelId.map(LabelId.apply)))
                    foundLabel      <- projectId.traverse(id => labelRepo.findLabel(id)(updatedLabel.name))
                } yield (createdProjects, projectId, createdLabels, updatedLabels, foundLabel.flatten)
                test.map { tuple =>
                    val (createdProjects, projectId, createdLabels, updatedLabels, foundLabel) = tuple
                    assert(createdProjects === 1, "Test vcs project was not created!")
                    assert(projectId.nonEmpty, "No vcs project id found!")
                    assert(createdLabels.exists(_ === 1), "Test label was not created!")
                    assert(updatedLabels === 1, "Test label was not updated!")
                    assert(foundLabel.nonEmpty, "Updated label not found!")
                    foundLabel.map { label =>
                        assertEquals(label, updatedLabel.copy(id = label.id))
                    }
                }
            case _ => fail("Could not generate data samples!")
        }
    }

    test("updateLabel must do nothing if id attribute is empty".tag(NeedsDatabase)) {
        (genProjectOwner.sample, genProject.sample, genLabel.sample) match {
            case (Some(owner), Some(generatedProject), Some(label)) =>
                val updatedLabel = label.copy(
                    id = None,
                    name = LabelName("updated label"),
                    description = Option(LabelDescription("I am an updated label description...")),
                    colour = ColourCode("#abcdef")
                )
                val project  = generatedProject.copy(owner = owner)
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val labelRepo = new DoobieLabelRepository[IO](tx)
                val test = for {
                    _               <- createProjectOwner(owner)
                    createdProjects <- createTicketsProject(project)
                    projectId       <- loadProjectId(owner.uid, project.name)
                    createdLabels   <- projectId.traverse(id => labelRepo.createLabel(id)(label))
                    labelId         <- findLabelId(owner.uid, project.name, label.name)
                    updatedLabels   <- labelRepo.updateLabel(updatedLabel)
                } yield (createdProjects, projectId, createdLabels, updatedLabels)
                test.map { tuple =>
                    val (createdProjects, projectId, createdLabels, updatedLabels) = tuple
                    assert(createdProjects === 1, "Test vcs project was not created!")
                    assert(projectId.nonEmpty, "No vcs project id found!")
                    assert(createdLabels.exists(_ === 1), "Test label was not created!")
                    assert(updatedLabels === 0, "Label with empty id must not be updated!")
                }
            case _ => fail("Could not generate data samples!")
        }
    }
}
