/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import cats.effect.*
import de.smederee.TestTags.*
import de.smederee.tickets.Generators.*
import doobie.*

final class DoobieTicketServiceApiTest extends BaseSpec {
    test("createOrUpdateUser must create new users".tag(NeedsDatabase)) {
        genTicketsUser.sample match {
            case Some(user) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val api = new DoobieTicketServiceApi[IO](tx)
                val test = for {
                    written   <- api.createOrUpdateUser(user)
                    foundUser <- loadTicketsUser(user.uid)
                } yield (written, foundUser)
                test.map { result =>
                    val (written, foundUser) = result
                    assert(written > 0, "No rows written to database!")
                    assertEquals(foundUser, Some(user))
                }

            case _ => fail("Could not generate data samples!")
        }
    }

    test("createOrUpdateUser must update existing users".tag(NeedsDatabase)) {
        (genTicketsUser.sample, genTicketsUser.sample) match {
            case (Some(user), Some(anotherUser)) =>
                val updatedUser = anotherUser.copy(uid = user.uid)
                val dbConfig    = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val api = new DoobieTicketServiceApi[IO](tx)
                val test = for {
                    created   <- api.createOrUpdateUser(user)
                    updated   <- api.createOrUpdateUser(updatedUser)
                    foundUser <- loadTicketsUser(user.uid)
                } yield (created, updated, foundUser)
                test.map { result =>
                    val (created, updated, foundUser) = result
                    assert(created > 0, "No rows written to database!")
                    assert(updated > 0, "No rows updated in database!")
                    assertEquals(foundUser, Some(updatedUser))
                }

            case _ => fail("Could not generate data samples!")
        }
    }

    test("deleteUser must delete existing users".tag(NeedsDatabase)) {
        genTicketsUser.sample match {
            case Some(user) =>
                val dbConfig = configuration.database
                val tx = Transactor.fromDriverManager[IO](
                    driver = dbConfig.driver,
                    url = dbConfig.url,
                    user = dbConfig.user,
                    password = dbConfig.pass,
                    logHandler = None
                )
                val api = new DoobieTicketServiceApi[IO](tx)
                val test = for {
                    _         <- api.createOrUpdateUser(user)
                    deleted   <- api.deleteUser(user.uid)
                    foundUser <- loadTicketsUser(user.uid)
                } yield (deleted, foundUser)
                test.map { result =>
                    val (deleted, foundUser) = result
                    assert(deleted > 0, "No rows deleted from database!")
                    assert(foundUser.isEmpty, "User not deleted from database!")
                }

            case _ => fail("Could not generate data samples!")
        }
    }
}
