/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import java.time.*
import java.util.Locale
import java.util.UUID

import cats.*
import cats.syntax.all.*
import de.smederee.email.EmailAddress
import de.smederee.i18n.LanguageCode
import de.smederee.security.*

import org.scalacheck.Arbitrary
import org.scalacheck.Gen

import scala.jdk.CollectionConverters.*

object Generators {
    val MinimumYear: Int = -4713  // Lowest year supported by PostgreSQL
    val MaximumYear: Int = 294276 // Highest year supported by PostgreSQL

    /** Prepend a zero to a single character hexadecimal code.
      *
      * @param hexCode
      *   A string supposed to contain a hexadecimal code between 0 and ff.
      * @return
      *   Either the given code prepended with a leading zero if it had only a single character or the originally given
      *   code otherwise.
      */
    private def hexPadding(hexCode: String): String =
        if (hexCode.length < 2)
            "0" + hexCode
        else
            hexCode

    val genLocalDate: Gen[LocalDate] =
        for {
            year  <- Gen.choose(MinimumYear, MaximumYear)
            month <- Gen.choose(1, 12)
            day   <- Gen.choose(1, Month.of(month).length(Year.of(year).isLeap))
        } yield LocalDate.of(year, month, day)

    given Arbitrary[LocalDate] = Arbitrary(genLocalDate)

    val genOffsetDateTime: Gen[OffsetDateTime] =
        for {
            year       <- Gen.choose(MinimumYear, MaximumYear)
            month      <- Gen.choose(1, 12)
            day        <- Gen.choose(1, Month.of(month).length(Year.of(year).isLeap))
            hour       <- Gen.choose(0, 23)
            minute     <- Gen.choose(0, 59)
            second     <- Gen.choose(0, 59)
            nanosecond <- Gen.const(0) // Avoid issues with loosing information during saving and loading.
            offset <- Gen.oneOf(
                ZoneId.getAvailableZoneIds.asScala.toList.map(ZoneId.of).map(ZonedDateTime.now).map(_.getOffset)
            )
        } yield OffsetDateTime.of(year, month, day, hour, minute, second, nanosecond, offset)

    given Arbitrary[OffsetDateTime] = Arbitrary(genOffsetDateTime)

    val genLocale: Gen[Locale]             = Gen.oneOf(Locale.getAvailableLocales.toList)
    val genLanguageCode: Gen[LanguageCode] = genLocale.map(_.getISO3Language).filter(_.nonEmpty).map(LanguageCode.apply)

    val genProjectOwnerId: Gen[ProjectOwnerId] = Gen.delay(ProjectOwnerId.randomProjectOwnerId)

    val genSubmitterId: Gen[SubmitterId] = Gen.delay(SubmitterId.randomSubmitterId)

    val genUUID: Gen[UUID] = Gen.delay(UUID.randomUUID)

    val genUserId: Gen[UserId] = Gen.delay(UserId.randomUserId)

    val genUsername: Gen[Username] = for {
        length <- Gen.choose(2, 30)
        prefix <- Gen.alphaChar
        chars <- Gen
            .nonEmptyListOf(Gen.alphaNumChar)
            .map(_.take(length).mkString.toLowerCase(Locale.ROOT))
    } yield Username(prefix.toString.toLowerCase(Locale.ROOT) + chars)

    val genSubmitter: Gen[Submitter] = for {
        id   <- genSubmitterId
        name <- genUsername.map(name => SubmitterName(name.toString))
    } yield Submitter(id, name)

    val genEmailAddress: Gen[EmailAddress] =
        for {
            length <- Gen.choose(4, 64)
            chars  <- Gen.nonEmptyListOf(Gen.alphaNumChar)
            email = chars.take(length).mkString
        } yield EmailAddress(email + "@example.com")

    val genTicketContent: Gen[Option[TicketContent]] = Gen.alphaStr.map(TicketContent.from)

    val genTicketStatus: Gen[TicketStatus]           = Gen.oneOf(TicketStatus.values.toList)
    val genTicketStatusList: Gen[List[TicketStatus]] = Gen.nonEmptyListOf(genTicketStatus).map(_.distinct)

    val genTicketResolution: Gen[TicketResolution]        = Gen.oneOf(TicketResolution.values.toList)
    val genTicketResolutions: Gen[List[TicketResolution]] = Gen.nonEmptyListOf(genTicketResolution).map(_.distinct)

    val genTicketNumber: Gen[TicketNumber]        = Gen.choose(0, Int.MaxValue).map(TicketNumber.apply)
    val genTicketNumbers: Gen[List[TicketNumber]] = Gen.nonEmptyListOf(genTicketNumber).map(_.distinct)

    val genTicketTitle: Gen[TicketTitle] =
        Gen.nonEmptyListOf(Gen.alphaNumChar).map(_.take(TicketTitle.MaxLength).mkString).map(TicketTitle.apply)

    val genTicketsUser: Gen[TicketsUser] = for {
        uid      <- genUserId
        name     <- genUsername
        email    <- genEmailAddress
        language <- Gen.option(genLanguageCode)
    } yield TicketsUser(uid, name, email, language)

    val genTicketsUsers: Gen[List[TicketsUser]] = Gen.nonEmptyListOf(genTicketsUser)

    val genTicket: Gen[Ticket] = for {
        number     <- genTicketNumber
        title      <- genTicketTitle
        content    <- genTicketContent
        status     <- genTicketStatus
        resolution <- Gen.option(genTicketResolution)
        submitter  <- Gen.option(genSubmitter)
        createdAt  <- genOffsetDateTime
        updatedAt  <- genOffsetDateTime
    } yield Ticket(
        number,
        title,
        content,
        status,
        resolution,
        submitter,
        createdAt,
        updatedAt
    )

    val genTickets: Gen[List[Ticket]] =
        Gen.nonEmptyListOf(genTicket)
            .map(_.zipWithIndex.map(tuple => tuple._1.copy(number = TicketNumber(tuple._2 + 1))))

    val genTicketFilter: Gen[TicketFilter] =
        for {
            number     <- Gen.listOf(genTicketNumber)
            status     <- Gen.listOf(genTicketStatus)
            resolution <- Gen.listOf(genTicketResolution)
            submitter  <- Gen.listOf(genSubmitter)
        } yield TicketFilter(number, status, resolution, submitter.map(_.name).distinct)

    val genProjectOwnerName: Gen[ProjectOwnerName] = for {
        length <- Gen.choose(2, 30)
        prefix <- Gen.alphaChar
        chars <- Gen
            .nonEmptyListOf(Gen.alphaNumChar)
            .map(_.take(length).mkString.toLowerCase(Locale.ROOT))
    } yield ProjectOwnerName(prefix.toString.toLowerCase(Locale.ROOT) + chars)

    val genProjectOwner: Gen[ProjectOwner] = for {
        id    <- genProjectOwnerId
        name  <- genProjectOwnerName
        email <- genEmailAddress
    } yield ProjectOwner(uid = id, name = name, email = email)

    given Arbitrary[ProjectOwner] = Arbitrary(genProjectOwner)

    val genProjectOwners: Gen[List[ProjectOwner]] = Gen
        .nonEmptyListOf(genProjectOwner)
        .map(_.foldLeft(List.empty[ProjectOwner]) { (acc, a) =>
            if (acc.exists(_.name === a.name))
                acc
            else
                a :: acc
        }) // Ensure distinct user names.

    val genLabelName: Gen[LabelName] =
        Gen.nonEmptyListOf(Gen.alphaNumChar).map(_.take(LabelName.MaxLength).mkString).map(LabelName.apply)

    val genLabelDescription: Gen[LabelDescription] =
        Gen.nonEmptyListOf(Gen.alphaNumChar)
            .map(_.take(LabelDescription.MaxLength).mkString)
            .map(LabelDescription.apply)

    val genColourCode: Gen[ColourCode] = for {
        red   <- Gen.choose(0, 255).map(_.toHexString).map(hexPadding)
        green <- Gen.choose(0, 255).map(_.toHexString).map(hexPadding)
        blue  <- Gen.choose(0, 255).map(_.toHexString).map(hexPadding)
        hexString = s"#$red$green$blue"
    } yield ColourCode(hexString)

    val genLabel: Gen[Label] = for {
        id          <- Gen.option(Gen.choose(0L, Long.MaxValue).map(LabelId.apply))
        name        <- genLabelName
        description <- Gen.option(genLabelDescription)
        colour      <- genColourCode
    } yield Label(id, name, description, colour)

    val genLabels: Gen[List[Label]] = Gen.nonEmptyListOf(genLabel).map(_.distinct)

    val genMilestoneTitle: Gen[MilestoneTitle] =
        Gen.nonEmptyListOf(Gen.alphaNumChar).map(_.take(MilestoneTitle.MaxLength).mkString).map(MilestoneTitle.apply)

    val genMilestoneDescription: Gen[MilestoneDescription] =
        Gen.nonEmptyListOf(Gen.alphaNumChar).map(_.mkString).map(MilestoneDescription.apply)

    val genMilestone: Gen[Milestone] =
        for {
            id     <- Gen.option(Gen.choose(0L, Long.MaxValue).map(MilestoneId.apply))
            title  <- genMilestoneTitle
            due    <- Gen.option(genLocalDate)
            descr  <- Gen.option(genMilestoneDescription)
            closed <- Gen.oneOf(List(false, true))
        } yield Milestone(id, title, descr, due, closed)

    val genMilestones: Gen[List[Milestone]] = Gen.nonEmptyListOf(genMilestone).map(_.distinct)

    val genProjectName: Gen[ProjectName] = Gen
        .nonEmptyListOf(
            Gen.oneOf(
                List(
                    "a",
                    "b",
                    "c",
                    "d",
                    "e",
                    "f",
                    "g",
                    "h",
                    "i",
                    "j",
                    "k",
                    "l",
                    "m",
                    "n",
                    "o",
                    "p",
                    "q",
                    "r",
                    "s",
                    "t",
                    "u",
                    "v",
                    "w",
                    "x",
                    "y",
                    "z",
                    "0",
                    "1",
                    "2",
                    "3",
                    "4",
                    "5",
                    "6",
                    "7",
                    "8",
                    "9",
                    "-",
                    "_"
                )
            )
        )
        .map(cs => ProjectName(cs.take(64).mkString))

    val genProjectDescription: Gen[Option[ProjectDescription]] =
        Gen.alphaNumStr.map(_.take(ProjectDescription.MaximumLength)).map(ProjectDescription.from)

    val genProject: Gen[Project] =
        for {
            name        <- genProjectName
            description <- genProjectDescription
            owner       <- genProjectOwner
            isPrivate   <- Gen.oneOf(List(false, true))
        } yield Project(owner, name, description, isPrivate)

    val genProjects: Gen[List[Project]] = Gen.nonEmptyListOf(genProject)

}
