/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.tickets

import de.smederee.tickets.Generators.*

import munit.*

import org.scalacheck.*
import org.scalacheck.Prop.*

final class TicketFilterTest extends ScalaCheckSuite {
    given Arbitrary[Submitter]        = Arbitrary(genSubmitter)
    given Arbitrary[TicketFilter]     = Arbitrary(genTicketFilter)
    given Arbitrary[TicketNumber]     = Arbitrary(genTicketNumber)
    given Arbitrary[TicketResolution] = Arbitrary(genTicketResolution)
    given Arbitrary[TicketStatus]     = Arbitrary(genTicketStatus)

    property("fromQueryParameter must produce empty filters for invalid input") {
        forAll { (randomInput: String) =>
            assertEquals(
                TicketFilter.fromQueryParameter(randomInput),
                TicketFilter(number = Nil, status = Nil, resolution = Nil, submitter = Nil)
            )
        }
    }

    property("fromQueryParameter must work for numbers only") {
        forAll { (numbers: List[TicketNumber]) =>
            val filter = TicketFilter(number = numbers, status = Nil, resolution = Nil, submitter = Nil)
            assertEquals(TicketFilter.fromQueryParameter(s"numbers: ${numbers.map(_.toString).mkString(",")}"), filter)
        }
    }

    property("fromQueryParameter must work for status only") {
        forAll { (status: List[TicketStatus]) =>
            val filter = TicketFilter(number = Nil, status = status, resolution = Nil, submitter = Nil)
            assertEquals(TicketFilter.fromQueryParameter(s"status: ${status.map(_.toString).mkString(",")}"), filter)
        }
    }

    property("fromQueryParameter must work for resolution only") {
        forAll { (resolution: List[TicketResolution]) =>
            val filter = TicketFilter(number = Nil, status = Nil, resolution = resolution, submitter = Nil)
            assertEquals(
                TicketFilter.fromQueryParameter(s"resolution: ${resolution.map(_.toString).mkString(",")}"),
                filter
            )
        }
    }

    property("fromQueryParameter must work for submitter only") {
        forAll { (submitters: List[Submitter]) =>
            if (submitters.nonEmpty) {
                val filter =
                    TicketFilter(number = Nil, status = Nil, resolution = Nil, submitter = submitters.map(_.name))
                assertEquals(
                    TicketFilter.fromQueryParameter(s"by: ${submitters.map(_.name.toString).mkString(",")}"),
                    filter
                )
            }
        }
    }

    property("toQueryParameter must include numbers") {
        forAll { (numbers: List[TicketNumber]) =>
            if (numbers.nonEmpty) {
                val filter = TicketFilter(number = numbers, status = Nil, resolution = Nil, submitter = Nil)
                assert(
                    TicketFilter.toQueryParameter(filter).contains(s"numbers: ${numbers.map(_.toString).mkString(",")}")
                )
            }
        }
    }

    property("toQueryParameter must include status") {
        forAll { (status: List[TicketStatus]) =>
            if (status.nonEmpty) {
                val filter = TicketFilter(number = Nil, status = status, resolution = Nil, submitter = Nil)
                assert(
                    TicketFilter.toQueryParameter(filter).contains(s"status: ${status.map(_.toString).mkString(",")}")
                )
            }
        }
    }

    property("toQueryParameter must include resolution") {
        forAll { (resolution: List[TicketResolution]) =>
            if (resolution.nonEmpty) {
                val filter = TicketFilter(number = Nil, status = Nil, resolution = resolution, submitter = Nil)
                assert(
                    TicketFilter
                        .toQueryParameter(filter)
                        .contains(s"resolution: ${resolution.map(_.toString).mkString(",")}"),
                    TicketFilter.toQueryParameter(filter)
                )
            }
        }
    }

    property("toQueryParameter must include submitter") {
        forAll { (submitters: List[Submitter]) =>
            if (submitters.nonEmpty) {
                val filter =
                    TicketFilter(number = Nil, status = Nil, resolution = Nil, submitter = submitters.map(_.name))
                assert(
                    TicketFilter
                        .toQueryParameter(filter)
                        .contains(s"by: ${submitters.map(_.name.toString).mkString(",")}"),
                    TicketFilter.toQueryParameter(filter)
                )
            }
        }
    }

    property("toQueryParameter must be the dual of fromQueryParameter") {
        forAll { (filter: TicketFilter) =>
            assertEquals(TicketFilter.fromQueryParameter(filter.toQueryParameter), filter)
        }
    }
}
