/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.security

import java.nio.charset.StandardCharsets

import cats.*
import cats.data.*
import cats.syntax.all.*

/** A password is stored as an `Array[Byte]` internally and its `validate(source: String)` function will check that the
  * input has a minimum length.
  */
opaque type Password = Array[Byte]
object Password {
    // The minimum length of a password to be considered valid.
    val MinimumLength: Int = 12
    // The maximum length of a password might look contra-productive but it can prevent "long password denial of service attacks".
    val MaximumLength: Int = 128

    /** Create an instance of Password from the given Array[Byte] type.
      *
      * @param source
      *   An instance of type Array[Byte] which will be returned as a Password.
      * @return
      *   The appropriate instance of Password.
      */
    def apply(source: Array[Byte]): Password = source

    /** Try to create an instance of Password from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a Password.
      * @return
      *   An option to the successfully converted Password.
      */
    def from(source: String): Option[Password] =
        Option(source) match {
            case None         => None
            case Some(string) => Option(string.getBytes(StandardCharsets.UTF_8))
        }

    /** Validate the given String against the criteria for a valid password.
      *
      * @param source
      *   A string which should fulfil the password criteria.
      * @return
      *   Either a list of errors or the validated Password.
      */
    def validate(source: String): ValidatedNec[String, Password] =
        Option(source).map(_.trim).filter(_.nonEmpty) match {
            case Some(password) =>
                if (password.length < MinimumLength)
                    s"Password must be at least $MinimumLength characters long!".invalidNec
                else if (password.length > MaximumLength)
                    s"Password must not be longer than $MaximumLength characters!".invalidNec
                else
                    password.getBytes(StandardCharsets.UTF_8).validNec
            case _ => "Password must not be empty!".invalidNec
        }

    extension (p: Password) {
        def toArray: Array[Byte] = p
    }
}
