/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.security

import java.nio.charset.StandardCharsets
import java.security.MessageDigest
import javax.crypto.Mac
import javax.crypto.spec.SecretKeySpec

import cats.syntax.all.*
import org.springframework.security.crypto.codec.Hex

extension (bytes: Array[Byte]) {

    /** Convert the byte array into a string containing the hexadecimal representation of the content.
      *
      * @return
      *   The hexadecimal notation representing the content.
      */
    def toHexString: String = Hex.encode(bytes).mkString
}

extension (string: String) {

    /** Convert the string into an array of bytes, assuming that it follows the notation generated by the corresponding
      * `toHexString` function.
      *
      * @return
      *   An array containing the bytes represented by the given hexadecimal codes.
      */
    def toBytes: Array[Byte] = Hex.decode(string)
}

opaque type SignedToken = String
object SignedToken {
    val Separator: String = "#" // The separator character to be used.

    /** Create an instance of SignedToken from the given String type.
      *
      * @param source
      *   An instance of type String which will be returned as a SignedToken.
      * @return
      *   The appropriate instance of SignedToken.
      */
    def apply(source: String): SignedToken = source

    /** Try to create an instance of SignedToken from the given String. Please note that this function will only do a
      * syntactic check!
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a SignedToken.
      * @return
      *   An option to the successfully converted SignedToken.
      */
    def from(source: String): Option[SignedToken] =
        Option(source).map(_.split(Separator, 3).length) match {
            case Some(3) => Option(source)
            case _       => None
        }
}

/** A class containing a private key and providing functionality for signing and validating messages / tokens.
  *
  * @param privateKey
  *   A private key used to initialise the underlying cryptographic implementations.
  */
final case class SignAndValidate(private val privateKey: PrivateKey) {
    val Algorithm: String = "HmacSHA256"

    /** Sign the given message string and return a string containing the signature in hexadecimal notation.
      *
      * @param message
      *   A string that should be signed.
      * @return
      *   The HMAC-SHA256 signature in hexadecimal notation for the given message string.
      */
    def sign(message: String): String = {
        val hmac = Mac.getInstance(Algorithm)
        hmac.init(new SecretKeySpec(privateKey.toArray, Algorithm))
        val signature = hmac.doFinal(message.getBytes(StandardCharsets.UTF_8))
        signature.toHexString
    }

    /** Sign the given token using the given nonce. The string produced by this function can be validated by the
      * corresponding [[validate]] function.
      *
      * @param token
      *   A string containing a token that shall be signed.
      * @param nonce
      *   A nonce string used to prepend it before the token.
      * @return
      *   A string containing the HMAC-SHA256 signature followed by a hash sign and the raw string used for signing
      *   which contains the nonce a hash sign and the token.
      */
    def signToken(token: String)(nonce: String): SignedToken = {
        val combined = s"$nonce${SignedToken.Separator}$token"
        SignedToken(s"${sign(combined)}${SignedToken.Separator}$combined")
    }

    /** Validate the given string which should contain a signature generated by the corresponding [[signToken]]
      * function. This function uses the [[java.security.MessageDigest#isEqual]] function which should ensure constant
      * time comparison.
      *
      * @param signedToken
      *   A string containing the HMAC-SHA256 signature followed by a hash sign and the raw string used for signing
      *   which contains the nonce a hash sign and the token.
      * @return
      *   Option containing the pure token if the validation was successful.
      */
    def validate(signedToken: SignedToken): Option[String] = {
        val parts     = signedToken.toString.split(SignedToken.Separator, 3)
        val signature = parts.headOption
        val nonce     = parts.drop(1).headOption
        val token     = parts.drop(2).headOption
        val comparison = (signature, nonce, token).mapN { case (signature, nonce, token) =>
            val reSigned = sign(s"$nonce${SignedToken.Separator}$token")
            if (MessageDigest.isEqual(signature.toBytes, reSigned.toBytes))
                Option(token)
            else
                None
        }
        comparison.getOrElse(None) // Could also be flatten...
    }
}
