/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.security

import java.util.UUID

import cats.*
import cats.syntax.all.*

import scala.util.matching.Regex

/** A user id is supposed to be globally unique and maps to the [[java.util.UUID]] type beneath.
  */
opaque type UserId = UUID
object UserId {
    val Format: Regex = "^[0-9a-f]{8}-[0-9a-f]{4}-[0-5][0-9a-f]{3}-[089ab][0-9a-f]{3}-[0-9a-f]{12}$".r

    given Eq[UserId]       = Eq.fromUniversalEquals
    given Ordering[UserId] = (x: UserId, y: UserId) => x.compareTo(y)
    given Order[UserId]    = Order.fromOrdering

    /** Create an instance of UserId from the given UUID type.
      *
      * @param source
      *   An instance of type UUID which will be returned as a UserId.
      * @return
      *   The appropriate instance of UserId.
      */
    def apply(source: UUID): UserId = source

    /** Try to create an instance of UserId from the given UUID.
      *
      * @param source
      *   A UUID that should fulfil the requirements to be converted into a UserId.
      * @return
      *   An option to the successfully converted UserId.
      */
    def from(source: UUID): Option[UserId] = Option(source)

    /** Try to create an instance of UserId from the given String.
      *
      * @param source
      *   A String that should fulfil the requirements to be converted into a UserId.
      * @return
      *   An option to the successfully converted UserId.
      */
    def fromString(source: String): Either[String, UserId] =
        Option(source)
            .filter(s => Format.matches(s))
            .flatMap { uuidString =>
                Either.catchNonFatal(UUID.fromString(uuidString)).toOption
            }
            .toRight("Illegal value for UserId!")

    /** Generate a new random user id.
      *
      * @return
      *   A user id which is pseudo randomly generated.
      */
    def randomUserId: UserId = UUID.randomUUID

    extension (uid: UserId) {
        def toUUID: UUID = uid
    }
}
