/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.security

import java.nio.charset.StandardCharsets

import munit.*

import org.scalacheck.*
import org.scalacheck.Prop.*

final class PasswordTest extends ScalaCheckSuite {
    property("Password.from(null) must always be None") {
        forAll { (_: String) =>
            assertEquals(Password.from(null), None) // scalafix:ok
        }
    }

    property("Password.from(string) must always be UTF8 bytes") {
        forAll { (s: String) =>
            val expected = s.getBytes(StandardCharsets.UTF_8)
            Password.from(s) match {
                case None => fail("Must produce valid byte array from non null string!")
                case Some(p) =>
                    assert(p.toArray.sameElements(expected), "Generated byte array differs from expected one!")
            }
        }
    }

    property(s"Password.validate must fail on trimmed string that is too long") {
        forAll { (input: String) =>
            if (input.trim.length > Password.MaximumLength) {
                assert(
                    Password.validate(input).isInvalid,
                    s"Passwords with more than ${Password.MaximumLength} characters must be invalid!"
                )
            }
        }
    }

    property(s"Password.validate must fail on trimmed string that is too short") {
        forAll { (s: String) =>
            val input = s.take(Password.MinimumLength).drop(1)
            assert(
                Password.validate(input).isInvalid,
                s"Passwords with less than ${Password.MinimumLength} characters must be invalid!"
            )
        }
    }
}
