/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.security

import cats.syntax.all.*

import munit.*

import org.scalacheck.*
import org.scalacheck.Prop.*

final class PermissionTest extends ScalaCheckSuite {
    private val genPermission: Gen[Permission] = Gen.oneOf(Permission.values.toList)
    given Arbitrary[Permission]                = Arbitrary(genPermission)
    given Arbitrary[Set[Permission]]           = Arbitrary(Gen.listOf(genPermission).map(_.toSet))

    property("Eq must hold") {
        forAll { (randomPermission: Permission) =>
            assert(randomPermission === randomPermission, "Identical permissions must be considered equal!")
            Permission.values.toList
                .filterNot(_ === randomPermission)
                .foreach(otherPermission =>
                    assert(randomPermission =!= otherPermission, "Different permissions mut not be considered equal!")
                )
        }
    }

    property("Permission.decode must be the dual of Permission.encode") {
        forAll { (randomPermission: Permission) =>
            assertEquals(Permission.decode(randomPermission.encoded), Set(randomPermission))
        }
    }

    property("Permission.decode must return an empty Set for invalid input") {
        forAll { (randomInteger: Int) =>
            if (randomInteger > 7 || randomInteger < 0) {
                assertEquals(Permission.decode(randomInteger), Set.empty)
            }
        }
    }

    property("Permission.decode must be the dual of Set[Permission].encode") {
        forAll { (permissions: Set[Permission]) =>
            assertEquals(Permission.decode(permissions.encode), permissions)
        }
    }

    property("Permission.from must succeed on valid input") {
        forAll { (randomPermission: Permission) =>
            assertEquals(Permission.from(randomPermission.toString), randomPermission.some)
        }
    }

    property("Permission.fromInt must succeed on valid input") {
        forAll { (randomPermission: Permission) =>
            assertEquals(Permission.fromInt(randomPermission.encoded), randomPermission.some)
        }
    }

    property("Set[Permission].encode must produce correct values") {
        forAll { (permissions: Set[Permission]) =>
            val expected = permissions.foldLeft(0)((acc, p) => acc | p.encoded)
            assertEquals(permissions.encode, expected)
        }
    }
}
