/*
 * Copyright (C) 2022  Contributors as noted in the AUTHORS.md file
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.smederee.security

import java.nio.charset.StandardCharsets

import cats.syntax.all.*

import munit.*

import org.scalacheck.*
import org.scalacheck.Prop.*

final class SignAndValidateTest extends ScalaCheckSuite {
    private val genKey: Gen[PrivateKey] =
        Gen.nonEmptyListOf(Gen.alphaNumChar).map(cs => PrivateKey(cs.mkString.getBytes(StandardCharsets.UTF_8)))
    given Arbitrary[PrivateKey] = Arbitrary(genKey)

    property("SignedToken.from must work for input with correct syntax") {
        forAll { (a: String, b: String, c: String) =>
            val input = s"$a${SignedToken.Separator}$b${SignedToken.Separator}$c"
            assert(SignedToken.from(input).nonEmpty)
        }
    }

    property("SignedToken.from must permit input with incorrect syntax") {
        forAll { (input: String) =>
            if (input.split(SignedToken.Separator, 3).size < 3)
                assert(SignedToken.from(input).isEmpty)
            else
                assert(SignedToken.from(input).nonEmpty)
        }
    }

    property("toHexString and toBytes must be commutative") {
        forAll { (bytes: Array[Byte]) =>
            assert(
                java.util.Arrays.equals(bytes, bytes.toHexString.toBytes),
                "Encoded and decoded value differs from original input!"
            )
        }
    }

    property("signToken and validate must be commutative") {
        forAll { (key: PrivateKey, token: String, nonce: String) =>
            val sav       = SignAndValidate(key)
            val signed    = sav.signToken(token)(nonce)
            val validated = sav.validate(signed)
            assert(validated.map(_ === token).getOrElse(false), "Validated token does not match original input!")
        }
    }
}
